#ifndef lint
static const char rcsid[] = "$Header: /net/hopkins/home/tristan/mpeg/dec/RCS/mtv.c,v 1.328 1999/03/10 02:25:13 tristan Exp tristan $";
static const char copyright[] = "Copyright (c) 1997, MpegTV LLC";
#endif


/*
 * Disclaimer of Warranty
 *
 * These software programs are available to the user without any license fee or
 * royalty on an "as is" basis.  Copyright-holder disclaims
 * any and all warranties, whether express, implied, or statuary, including any
 * implied warranties or merchantability or of fitness for a particular
 * purpose.  In no event shall the copyright-holder be liable for any
 * incidental, punitive, or consequential damages of any kind whatsoever
 * arising from the use of these programs.
 *
 * This disclaimer of warranty extends to the user of these programs and user's
 * customers, employees, agents, transferees, successors, and assigns.
 *
 */

#include "mtvp/mtvp.h"

#ifdef SOLARIS
#include <volmgt.h>
#ifdef USE_SMC
#include "MsmClient.h"
#endif
#endif

#include "constrained.h"
#include "forms.h"
#include "controls.h"
#include "version.h"
#include "get_time_now.h"
#include "pref.h"

#ifndef DEBUG
#	define GET_TIME_NOW_INLINE	/* use in-line version if possible - not implemented in public version of get_time_now.h */
#endif
#include "get_time_now.h"

#include "mtv.h"
#include "mozilla.h"

/* check XForms version - error if unknow/untested versions of XForms */
#if (FL_VERSION != 0) || ((FL_REVISION != 88) && (FL_REVISION != 87))
#	error unsupported XForms version - must be 0.87 or 0.88
#endif

/* challenge.h not available in public distribution */
#ifdef HANDLE_CHALLENGE
#	include "challenge.h"
#	include "keyformat.h"
#define KEY_PRODUCT_MTV_PLAYER	4
#define KEY_CATEGORY_COMMERCIAL	3
#endif

/*--------------------------------------------------------------------------+
| Configuration																|
+--------------------------------------------------------------------------*/
#if defined(LINUX) || (defined(SOLARIS) && !defined(__i386__))
#define VCD	 /* play Video-CD [experimental] */
#endif

/*
 * Initial PCM volume setting at startup.
 */
#define INITIAL_PCM_VOLUME	(-1)
#define INITIAL_PCM_BALANCE	(-1)
#define INITIAL_MASTER_VOLUME	(-1)
#define INITIAL_MASTER_BALANCE	(-1)

/* #define MTV_EASTER_EGG test the current implementation of the "reserved" MTVP_DEMO command - not for public use: reserved commands are not part of the published API and may change at any time */

/*--------------------------------------------------------------------------+
| Macros																	|
+--------------------------------------------------------------------------*/

/* fast forward plays FAST_FORWARD_COEF times faster */
#define FAST_FORWARD_COEF	8

/* frequence of slider position update when playing (in sec) */
#define SLIDER_UPDATE_FREQUENCY	1

#define G2A_SEND_VOID(com, flags)		g2a_send_void(&Player, com, flags)
#define G2A_SEND_INT(com, flags, n)		g2a_send_int(&Player, com, flags, n)
#define G2A_SEND_DOUBLE(com, flags, n)	g2a_send_double(&Player, com, flags, n)
#define G2A_SEND_STRING(com, flags, n)	g2a_send_string(&Player, com, flags, n)
#define G2A_SEND_CUSTOM(com, flags, p, n)	g2a_send_custom(&Player, com, flags, p, n)

/* values for slider controls */
#ifdef SGI
#define GAMMA_DEFAULT	0.8
#elif defined(SOLARIS)
#define GAMMA_DEFAULT	1.4
#else
#define GAMMA_DEFAULT	1.0
#endif

#define CONTRAST_DEFAULT	/*(36.0/219.0)*/ (-LUMA_FLOOR_NORMAL)
#define CONTRAST_MIN	(-LUMA_FLOOR_MAX)
#define CONTRAST_MAX	(-LUMA_FLOOR_MIN)

#define BRIGHTNESS_DEFAULT	/*(20.0/219.0)*/ LUMA_SHIFT_NORMAL
#define BRIGHTNESS_MIN	LUMA_SHIFT_MIN
#define BRIGHTNESS_MAX	LUMA_SHIFT_MAX

#define CHROMA_SATURATION_DEFAULT	CHROMA_SATURATION_NORMAL

#define FRAMERATE_MIN	SYNC_FRAMERATE_MIN
#define FRAMERATE_MAX	240.0

#define SET_ONELINER(ob, str) STMT( \
		(ob)->u_vdata = (str); \
		fl_set_object_posthandler(ob, post_handler); \
	)

/* max audio play time for unregistered player (in sec) */
#define DEMO_AUDIO_PLAY_TIME 60

/* for the msg.h package */
#define MSG_PTR (&Msg)

/* Those URL's are hardcoded in mtv */
#define URL_RELEASE_NOTES	"http://www.mpegtv.com/mtv/release-notes"
#define URL_FAQ				"http://www.mpegtv.com/mtv/faq"
#define URL_MANUAL			"http://www.mpegtv.com/mtv/manual"
#define URL_SUPPORT			"http://www.mpegtv.com/mtv/support"
#define URL_REGISTER		"http://www.mpegtv.com/mtv/register"
#define URL_ABOUT_MPEG		"http://www.mpegtv.com/mtv/about-mpeg"
#define URL_BITSTREAMS		"http://www.mpegtv.com/mtv/bitstreams"

/*--------------------------------------------------------------------------+
| Data types																|
+--------------------------------------------------------------------------*/
/*
 * User options
 *
 * The xxx_defined variables, if set to True, indicate that a value for xxx
 * was specified by the user using command-line options. In this case,
 * the specified value will overwrite any default from the application,
 * including initial settings read from the preferences file.
 */
typedef struct {
	Bool	usage;
	Bool	loop;
	Bool	loop_defined;
	ZOOM_TYPE	zoom;
	Bool	zoom_defined;
	Bool	iconify;
	Bool	auto_play;
	Bool	auto_play_defined;
	DITHER_TYPE	dither_type;
	SYNC_TYPE	sync_type;
	Bool	sync_type_defined;
	double	sync_framerate;
	Bool	sync_framerate_defined;
	const char *url;
	u_int		msg_select;
	double	gamma;
	Bool	gamma_defined;
	double	contrast;
	Bool	contrast_defined;
	double	brightness;
	Bool	brightness_defined;
	double	chroma_saturation;
	Bool	chroma_saturation_defined;
	u_int		audio_msg_select;
	AUDIO_MODE	audio_mode;
	AUDIO_QUALITY	audio_quality;
	Bool		audio_quality_defined;
	AUDIO_CHANNEL	audio_channel;
	Bool		audio_channel_defined;
	AUDIO_CONTROLS	audio_controls;
	double		audio_pcm_volume;
	double		audio_pcm_balance;
	double		audio_master_volume;
	double		audio_master_balance;
	int			audio_ports;
	const char **xforms_argv;
	int		pmp_base_fd;
	int		existing_window;	/* -1 means not set by option -W<n> */
	VIDEO_LAYOUT	video_layout;
	off_t	loop_start_pos;
	off_t	loop_end_pos;
	int		mtvp_geometry_x;
	int		mtvp_geometry_y;
	Bool	mtvp_geometry_defined;
	int		geometry_x;
	int		geometry_y;
	u_int	geometry_w;
	u_int	geometry_h;
	const char *pref_filename;
	Bool	private_colormap;
} USER_OPTIONS;

/* keys used in the preference file */
#define PREF_LOOP "LOOP"
#define PREF_ZOOM "ZOOM"
#define PREF_AUTO_PLAY "AUTO_PLAY"
#if 0 /* Tristan Savatier - Fri Jan 16 02:36:27 PST 1998 */
//#define PREF_AUTO_EXIT "AUTO_EXIT"
#endif /* 0 - Tristan Savatier - Fri Jan 16 02:36:27 PST 1998 */
#define PREF_VIDEO_GAMMA "VIDEO_GAMMA"
#define PREF_VIDEO_LUMA_CONTRAST "VIDEO_LUMA_CONTRAST"
#define PREF_VIDEO_LUMA_BRIGHTNESS "VIDEO_LUMA_BRIGHTNESS"
#define PREF_VIDEO_CHROMA_SATURATION "VIDEO_CHROMA_SATURATION"
#define PREF_MTVP_WINDOW_POSITION_X "MTVP_WINDOW_POSITION_X"
#define PREF_MTVP_WINDOW_POSITION_Y "MTVP_WINDOW_POSITION_Y"
#define PREF_AUDIO_CHANNEL "AUDIO_CHANNEL"
#define PREF_AUDIO_QUALITY "AUDIO_QUALITY"
#define PREF_SYNC_TYPE "SYNC_TYPE"
#define PREF_SYNC_FRAMERATE "SYNC_FRAMERATE"
#define PREF_VIDEO_ASPECT_RATIO_MODE "VIDEO_ASPECT_RATIO_MODE"
#define PREF_VIDEO_PRIVATE_COLORMAP "VIDEO_PRIVATE_COLORMAP"

#define PREF_REGISTRATION_NAME "REGISTRATION_NAME"
#define PREF_REGISTRATION_ORDER_NUMBER "REGISTRATION_ORDER_NUMBER"
#define PREF_REGISTRATION_KEY "REGISTRATION_KEY"

/* keys and their default values for the preferences */
static const ENTRY Pref_keys[] = {
	{PREF_LOOP, "False"},
	{PREF_ZOOM, "0"},
	{PREF_AUTO_PLAY, "True"},
#ifdef SGI
	{PREF_VIDEO_GAMMA, "0.8"},
#elif defined(SOLARIS)
	{PREF_VIDEO_GAMMA, "1.4"},
#else
	{PREF_VIDEO_GAMMA, "1.0"},
#endif
	{PREF_VIDEO_LUMA_CONTRAST, "0.0"},
	{PREF_VIDEO_LUMA_BRIGHTNESS, "0.0"},
	{PREF_VIDEO_CHROMA_SATURATION, "1.0"},
	{PREF_MTVP_WINDOW_POSITION_X, "200"},
	{PREF_MTVP_WINDOW_POSITION_Y, "300"},
	{PREF_AUDIO_CHANNEL, "1"},
	{PREF_AUDIO_QUALITY, "1"},
	{PREF_SYNC_TYPE, "0"},
	{PREF_SYNC_FRAMERATE, "0.0"},
	{PREF_VIDEO_ASPECT_RATIO_MODE, "0"},
	{PREF_VIDEO_PRIVATE_COLORMAP, "False"},
	{PREF_REGISTRATION_NAME, ""},
	{PREF_REGISTRATION_ORDER_NUMBER, ""},
	{PREF_REGISTRATION_KEY, ""},
};


/* for the file selector state */
typedef struct {
	char directory[PATH_MAX];
	char pattern[PATH_MAX];
	char filename[PATH_MAX];
} SEL_STATE;

/* File menu items */
typedef enum {
	FILE_MENU_NONE = -1,
	FILE_MENU_HANGING = 0,	/* undocumented feature of Xforms menus */
	FILE_MENU_OPEN_FILE = 1,
	FILE_MENU_OPEN_URL,
	FILE_MENU_STDIN,
#ifdef _REENTRANT
	FILE_MENU_UDP,
#endif
#ifdef VCD
	FILE_MENU_VCD,
#endif
#ifdef USE_SMC
	FILE_MENU_SMC,
#endif
	FILE_MENU_CLOSE,
	FILE_MENU_SAVE_AS,
	FILE_MENU_INFOS,
	FILE_MENU_QUIT
} FILE_MENU;

/* Options menu items */
typedef enum {
	OPTIONS_MENU_NONE = -1,
	OPTIONS_MENU_HANGING = 0,	/* undocumented feature of Xforms menus */
	OPTIONS_MENU_AUDIO,
	OPTIONS_MENU_VIDEO,
	OPTIONS_MENU_SYNC,
	OPTIONS_MENU_AUTO_PLAY,
	OPTIONS_MENU_SAVE
} OPTIONS_MENU;

/* Window menu items */
typedef enum {
	WINDOW_MENU_NONE = -1,
	WINDOW_MENU_HANGING = 0,	/* undocumented feature of Xforms menus */
#ifdef USE_SDL
	WINDOW_MENU_FULLSCREEN,
#endif
	WINDOW_MENU_RAISE,
	WINDOW_MENU_ROOT,
	WINDOW_MENU_FRAME_INFOS,
	WINDOW_MENU_SAVE_PPM,
	WINDOW_MENU_SAVE_TGA
} WINDOW_MENU;

/* Help menu items */
typedef enum {
	HELP_MENU_NONE = -1,
	HELP_MENU_HANGING = 0,	/* undocumented feature of Xforms menus */
	HELP_MENU_ABOUT_MTV,
	HELP_MENU_RELEASE_NOTES,
	HELP_MENU_MANUAL,
	HELP_MENU_FAQ,
	HELP_MENU_SUPPORT,
	HELP_MENU_LICENSE,
	HELP_MENU_REGISTER,
	HELP_MENU_KEY,
	HELP_MENU_ABOUT_MPEG,
	HELP_MENU_BITSTREAMS
} HELP_MENU;

/* Slider control */
typedef struct {
	/*
	 * not_ready == True means message return receipt not received yet,
	 * any new value is put in pending (instead of sending it)
	 * and is_pending is set to True
	 */
	Bool	not_ready;
	Bool	is_pending;	/* is there a new value to send ? */
	double	pending;	/* if is_pending == True, new value to send */
} SLIDER_CONTROL;

/*--------------------------------------------------------------------------+
| Globals																	|
+--------------------------------------------------------------------------*/

/* The form pointers */
FD_controls *Controls;
FD_about *About;
FD_license *License;
FD_registration *Registration;
FD_splash *Splash;
FD_audio *Audio;
FD_sync *Sync;
FD_infos *Infos;
FD_frame_infos *Frame_infos;
FD_video *Video;
#ifdef _REENTRANT
FD_udp *Udp;
#endif

/* to communicate with player */
PMP_STRUCT	Player;

/* for the msg.h package */
MSG_STRUCT Msg;

/*--------------------------------------------------------------------------+
| Statics																	|
+--------------------------------------------------------------------------*/
static struct tm Expiration_time = {
	0, /* sec */
	0, /* min */
	0, /* hour */
	1, /* day of month */
	0,	/* month [0..11] */
	101	/* year since 1900 */
};

#if 0 /* Tristan Savatier - Mon Jan 25 11:54:45 PST 1999 */
//static Bool Expired;	/* expired unregistered version */
#endif /* 0 - Tristan Savatier - Mon Jan 25 11:54:45 PST 1999 */

/* name of run-time Player. Default is "mtvp", but can be changed. */
static const char *Player_name = "mtvp";

static const char *Argv0;	/* argv[0] */

/* default for the file selectors */
static SEL_STATE Open_file = {"", "*.mpg", ""};
static SEL_STATE Save_file = {"", "*.mpg", ""};
static SEL_STATE Save_image_ppm = {"", "*.ppm", ""};
static SEL_STATE Save_image_tga = {"", "*.tga", ""};

/* for the play/pause button */
static STATE Gui_play_pause_set_state = STATE_PLAYING;

/* player's state, used by the my_idle() routine */
static STATE Gui_state = STATE_STOPPED;

/* user options */
static USER_OPTIONS User_options;

/* Will be set the first time the program is run */
static Bool Must_save_preferences = False;

/* The pid of the control process */
static pid_t Pid;

/*
 * set Not_ready when sending a seek command. Reset it when a A2G_COM_DISPLAY_READY
 * message is received.
 * Not_ready is used to avoid piling up seek commands faster than the
 * player can process them.
 */
static Bool Not_ready;
static int Pending_seek = -1;	/* pending seek or -1 */

static double Nominal_framerate;		/* nominal framerate */
static Bool Framerate_is_nominal;	/* True means mtvp is playing in sync at Nominal_framerate */
static Bool Specified_framerate_is_nominal;
static int Bits_filesize = -1;	/* stream size (in bytes) or -1 if n/a (not seekable) */
static u_int Mux_bitrate;			/* mux bitrate or 0 if n/a */
static u_int Video_bitrate;		/* video bitrate or 0 if n/a */

static SLIDER_CONTROL Gamma;		/* for gamma slider */
static SLIDER_CONTROL Contrast;	/* for contrast slider */
static SLIDER_CONTROL Brightness;	/* for brightness slider */
static SLIDER_CONTROL Chroma_saturation;	/* for chroma slider */

static Pixmap Icon_pixmap;			/* MpegTV Icon for closed windows */

static FL_OBJECT *Oneliner_object;	/* NULL or ptr to current object for oneliner */

static off_t Stream_pos;		/* last position received from player */

/* Fast_forward is necessary to work around a bug of the IN-OUT buttons in Xforms! */
static Bool Fast_forward;		/* if True we are playing in FFW mode */
static STATE State_before_fast_forward;/* restore this state after FFW */
static Bool Seeking;			/* if True we are playing in seeking mode */
static STATE State_before_seeking;		/* restore this state after seeking */

static int Audio_ports;			/* Audio ports available */

static AUDIO_QUALITY	Audio_quality;
static AUDIO_CHANNEL	Audio_channel;
static SYNC_TYPE Sync_type;

static const char *License_lines[] = {
#include "player-license.h"
};

static const char *MpegTV = "MpegTV";

static ASPECT_RATIO_TYPE Aspect_ratio_mode = ASPECT_RATIO_SQUARE_PELS;

static Bool Registered;		/* valid registration key found */
static Bool Use_mtvp_key;	/* tell mtvp to look for key in mtvp.key file */
static const char *Order_number;	/* registration order number */
static const char *Key;		/* registration key */
static const char *Name;	/* registered user rame */

static double Play_start_time;	/* time play started */
static Bool Audio_no_more;	/* no more audio! */
static Bool Audio_is_available;	/* audio available */

static Bool Quitting;

#ifdef VCD
static const char *Vcd_device;	/* VCD device name */
static u_int Vcd_tracks;	/* 0 if VCD device not opened, >=1 if opened */
static u_int Vcd_track;		/* current track played (>= 1) */
#endif

#ifdef USE_SMC
static MsmServer Smc_ms;
static MsmPlayer Smc_mp;
#endif

static int Timeout_id = -1;

#ifdef HANDLE_CHALLENGE
static LicenseKey License_key;
#endif

/*--------------------------------------------------------------------------+
| clean_exit																|
+--------------------------------------------------------------------------*/
static void
clean_exit(
	int		n
)
{
	MESSAGE(("(pid=%u): exiting, status = %d", (u_int)Pid, n));

#if 0 /* Tristan Savatier - Thu Dec 10 20:04:12 PST 1998 */
//	G2A_SEND_VOID(COM_QUIT, 0);
#endif /* 0 - Tristan Savatier - Thu Dec 10 20:04:12 PST 1998 */

	/* make sure we ignore SIGCHLD */
	(void)signal(SIGCHLD, SIG_IGN);
	
	/* send SIGTERM signal to Player */
	if (Player.pid != 0) {
		kill(Player.pid, SIGTERM);
	}
	
	/* terminate (real exit this time) */
	exit(n);
}


/*--------------------------------------------------------------------------+
| get_elapsed_seconds														|
+--------------------------------------------------------------------------*/
static INLINE int
get_elapsed_seconds(
	off_t		pos
)
{
	if (Mux_bitrate / BITS(char) != 0) {
		return (int)((u_int)pos / (Mux_bitrate / BITS(char)));
	} else if (Video_bitrate / BITS(char) != 0) {
		return (int)((u_int)pos / (Video_bitrate / BITS(char)));
	} else {
		return -1;
	}
}

/*--------------------------------------------------------------------------+
| get_time_string															|
+--------------------------------------------------------------------------*/
static const char *
get_time_string(
	int		sec
)
{
	static char buffer[9];

#if 0 /* Tristan Savatier - Sat Aug 23 02:57:21 PDT 1997 */
//	printf("sec = %d\n", sec);
#endif /* 0 - Tristan Savatier - Sat Aug 23 02:57:21 PDT 1997 */
	
	if (sec >= 0) {
		int		hh, mm, ss;

		hh = sec / 3600;
		mm = (sec / 60) % 60;
		ss = sec % 60;
		
		if (hh > 99) {
			/*
			 * This can only happen if an absurd value of Video_bitrate
			 * or Mux_bitrate is received.  But if we don't test that,
			 * some bad streams can (and do) cause mtv to crash.
			 */
			sprintf(buffer, "??:??:??");
		} else {
			sprintf(buffer, "%02d:%02d:%02d", hh, mm, ss);
		}
		if (buffer[sizeof buffer - 1] != 0) abort_now();
		
		return buffer;
	} else {
		return "n/a";
	}
}


/*--------------------------------------------------------------------------+
| copy_str																	|
+--------------------------------------------------------------------------*/
static void
copy_str(
	char *dst,
	const char *src,
	u_int	size
)
{
	strncpy(dst, src, size - 1);
	dst[size - 1] = 0;
}


/*--------------------------------------------------------------------------+
| get_position_parameters													|
+--------------------------------------------------------------------------*/
static void
get_position_parameters()
{
	G2A_SEND_VOID(COM_GET_POS, 0);
	if (!Seeking) {
		G2A_SEND_VOID(COM_GET_NUMBER_OF_FRAMES_PLAYED, 0);
		if (Gui_state == STATE_PLAYING) {
			G2A_SEND_VOID(COM_GET_ACTUAL_FRAMERATE, 0);
		}
	}
}

/*--------------------------------------------------------------------------+
| show_and_raise_form														|
+--------------------------------------------------------------------------*/
static void
show_and_raise_form(
	FL_FORM *form,
	int		place,
	int		border,
	const char *title
)
{
	fl_show_form(form, place, border, title);
	fl_raise_form(form);
}

/*--------------------------------------------------------------------------+
| at_close_free																|
+--------------------------------------------------------------------------*/
static int 
at_close_free(
	FL_FORM *fl_form,
	void	*data
)
{
	fl_hide_form(fl_form);
	fl_free_form(fl_form);
	return FL_IGNORE;
}

/*--------------------------------------------------------------------------+
| show_alert																|
+--------------------------------------------------------------------------*/
void
show_alert(
	const char *s1,
	const char *s2,
	const char *s3,
	int		n	/* not used in this replacement routine... */
)
{
	FD_alert *alert;
	
	/*
	 * This is a replacement for Xform's fl_show_alert() routine.
	 * There is a bug in Xforms-0.81 that causes text in fl_show_alert() windows
	 * to get garbled when several alerts windows are opened at the same time.
	 * Note: fl_show_alert() does not work well in Xforms-0.86 either:
	 * Alert windows disapear when fl_show_alert() is called several times
	 * successively (without dismissing the previous alert window), which can
	 * happen. We like better our custom non-modal alerts.
	 * This replacement routine prevents all the problems...
	 */

	if ((alert = create_form_alert()) == NULL) {
		FATAL_ERROR(("cannot create Alert form"));
	}
	fl_set_object_label(alert->s1, s1);
	fl_set_object_label(alert->s2, s2);
	fl_set_object_label(alert->s3, s3);
	
	fl_set_form_atclose(alert->alert, at_close_free, NULL);
	fl_set_form_icon(alert->alert, Icon_pixmap, 0);

	show_and_raise_form(alert->alert, FL_PLACE_CENTER, FL_FULLBORDER, "MpegTV Alert");
}


/*--------------------------------------------------------------------------+
| disable_audio																|
+--------------------------------------------------------------------------*/
static void
disable_audio()
{
	fl_set_button(Controls->audio_mute, 1);
	G2A_SEND_INT(COM_AUDIO_MUTE, 0, 1);
	Audio_no_more = True;
#if 0 /* Tristan Savatier - Mon Jan 25 12:02:20 PST 1999 */
//	show_alert("Thanks for trying MpegTV Player 1.0", "To get unlimited audio playback, please REGISTER your copy", "at www.mpegtv.com", 1);
#endif /* 0 - Tristan Savatier - Mon Jan 25 12:02:20 PST 1999 */
	fl_show_message("Thanks for trying mtv. To get unlimited audio playback,", "please REGISTER your copy at www.mpegtv.com", "Registered copies do not display this notice");
}

/*--------------------------------------------------------------------------+
| update_slider																|
+--------------------------------------------------------------------------*/
static void
update_slider()
{		
	if (Seeking) {
		return;
	}
	if (Gui_state != STATE_PLAYING && !Fast_forward) {
		return;
	}
		
	get_position_parameters();

	if (!Registered && Audio_is_available && !Audio_no_more && get_time_now() > Play_start_time + DEMO_AUDIO_PLAY_TIME) {
		disable_audio();
	}
}

#if 0 /* Tristan Savatier - Mon Mar 16 00:15:28 PST 1998 */
///*--------------------------------------------------------------------------+
//| my_idle																	|
//+--------------------------------------------------------------------------*/
//static int
//my_idle(
//	const XEvent *xev,
//	const void *data
//)
//{
//	double time_now;
//	static double Last_time;
//	
//	printf("."); fflush(stdout);
//		
//	time_now = get_time_now();
//
//	if (time_now - Last_time < SLIDER_UPDATE_FREQUENCY) {
//		return 0;
//	}
//	
//	Last_time = time_now;
//
//	update_slider();
//
//	return 0;
//}
#endif /* 0 - Tristan Savatier - Mon Mar 16 00:15:28 PST 1998 */


/*--------------------------------------------------------------------------+
| at_close_hide																|
+--------------------------------------------------------------------------*/
static int 
at_close_hide(
	FL_FORM *fl_form,
	void	*data
)
{
	fl_hide_form(fl_form);
	return FL_IGNORE;
}

/*--------------------------------------------------------------------------+
| at_close_ignore															|
+--------------------------------------------------------------------------*/
static int 
at_close_ignore(
	FL_FORM *fl_form,
	void	*data
)
{
	return FL_IGNORE;
}

/*--------------------------------------------------------------------------+
| file_browser_at_close_hide												|
+--------------------------------------------------------------------------*/
static int 
file_browser_at_close_hide(
	FL_FORM *fl_form,
	void	*data
)
{
	fl_hide_form(fl_form);
	
#if FL_REVISION == 81
	/*
	 * A bug of some sort. File selector disables all
	 * other forms, so when it goes away this way, 
	 * all forms remain disabled. This will be fixed.
	 * For the time being,  you can add fl_activate_all_forms
	 * after fl_hide_form.
	 */
	fl_activate_all_forms();
#endif

	return FL_IGNORE;
}


/*--------------------------------------------------------------------------+
| quit																		|
+--------------------------------------------------------------------------*/
static void
quit(int status)
{ 
	G2A_SEND_VOID(COM_QUIT, 0);
	
	Quitting = True;
#if 0 /* Tristan Savatier - Mon Feb 9 00:06:44 PST 1998 */
//	clean_exit(status);
#endif /* 0 - Tristan Savatier - Mon Feb 9 00:06:44 PST 1998 */
}

/*--------------------------------------------------------------------------+
| at_close_controls															|
+--------------------------------------------------------------------------*/
static int 
at_close_controls(
	FL_FORM *fl_form,
	void	*data
)
{
	quit(EXIT_SUCCESS);
	return FL_OK;
}

/*--------------------------------------------------------------------------+
| at_close_quit																|
+--------------------------------------------------------------------------*/
static int 
at_close_quit(
	FL_FORM *fl_form,
	void	*data
)
{
	clean_exit(EXIT_FAILURE);
	return 0;	/* to avoid a compiler warning */
}

#if 0 /* Tristan Savatier - Sun Jan 11 20:29:14 PST 1998 */
///*--------------------------------------------------------------------------+
//| at_activate_controls														|
//+--------------------------------------------------------------------------*/
//static void 
//at_activate_controls(
//	FL_FORM *fl_form,
//	void	*data
//)
//{
//	printf("at activate controls\n");
//}
//
///*--------------------------------------------------------------------------+
//| at_deactivate_controls													|
//+--------------------------------------------------------------------------*/
//static void 
//at_deactivate_controls(
//	FL_FORM *fl_form,
//	void	*data
//)
//{
//	printf("at deactivate controls\n");
//}
#endif /* 0 - Tristan Savatier - Sun Jan 11 20:29:14 PST 1998 */

/*--------------------------------------------------------------------------+
| show_alert_message														|
+--------------------------------------------------------------------------*/
void
show_alert_message(
	const char *msg1,
	const char *msg2
)
{
	const char *s;
	
	if (errno != 0) {
		s = strerror(errno);
	} else {
		s = NULL;
	}

	show_alert(msg1, msg2, s == NULL ? "" : s, 1);
}

/*--------------------------------------------------------------------------+
| reset_loop																|
+--------------------------------------------------------------------------*/
static void
reset_loop()
{
	/* reset Auto_seek to 0 */
	if (User_options.loop_start_pos != 0) {
		G2A_SEND_INT(COM_LOOP_START_POS, 0, 0);
		User_options.loop_start_pos = 0;
	}
	if (User_options.loop_end_pos != 0) {
		G2A_SEND_INT(COM_LOOP_END_POS, 0, 0);
		User_options.loop_end_pos = 0;
	}
}

/*--------------------------------------------------------------------------+
| reset_auto_seek															|
+--------------------------------------------------------------------------*/
static void
reset_auto_seek()
{
	/* reset Auto_seek to 0 */
	G2A_SEND_INT(COM_AUTO_SEEK, 0, 0);
}

/*--------------------------------------------------------------------------+
| reset_loop_if_pos_out_of_loop												|
+--------------------------------------------------------------------------*/
static void
reset_loop_if_pos_out_of_loop(
	off_t	n
)
{
	/* reset Auto_seek and loop_end_pos 0 is n is out of the loop */
	
	if ((User_options.loop_start_pos != 0 && n < User_options.loop_start_pos) || (User_options.loop_end_pos != 0 && n > User_options.loop_end_pos)) {
		G2A_SEND_INT(COM_LOOP_START_POS, 0, 0);
		G2A_SEND_INT(COM_LOOP_END_POS, 0, 0);
		User_options.loop_start_pos = 0;
		User_options.loop_end_pos = 0;
	}
}

/*--------------------------------------------------------------------------+
| file_command																|
+--------------------------------------------------------------------------*/
static void
file_command(
	SEL_STATE	*p,
	int		sel_number,
	const char *message,
	COMMAND	com
)
{
	const char *answer;

	/*
	 * This code makes the file selector smarter than
	 * the default fl_show_fselector() routine.
	 */
	
	fl_use_fselector(sel_number);
	
	fl_set_form_atclose(fl_get_fselector_form(), file_browser_at_close_hide, NULL);
	fl_set_form_icon(fl_get_fselector_form(), Icon_pixmap, 0);
	
	fl_set_fselector_title(message);
	fl_invalidate_fselector_cache();

	answer = fl_show_fselector (message,
									p->directory,
									p->pattern,
									p->filename);

	/* remember state for next time! */
	copy_str(p->pattern, (const char *)fl_get_pattern(), sizeof p->pattern);
	copy_str(p->directory, (const char *)fl_get_directory(), sizeof p->directory);
	copy_str(p->filename, (const char *)fl_get_filename(), sizeof p->filename);
	
	if (answer == NULL) {
		return;
	}
	if (answer[0] == 0) {
		errno = 0;
		show_alert_message("No file selected!", "");
		return;
	}

	if (com == COM_OPEN_STREAM_URL) {
		PMP_DATA_GENERIC data_generic;
		char *url = data_generic.data_string;

		url[sizeof data_generic.data_string - 1] = 0;
		sprintf(url, "file:%s", answer);
		if (url[sizeof data_generic.data_string - 1] != 0) abort_now();
	
		G2A_SEND_STRING(COM_OPEN_STREAM_URL, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, url);
	} else {

		G2A_SEND_STRING(com, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, answer);
	}
}

/*--------------------------------------------------------------------------+
| select_file_in_directory													|
+--------------------------------------------------------------------------*/
static void
select_file_in_directory(
	const char *dirname
)
{
	SEL_STATE	*p = &Open_file;
	
	if (dirname[0] != 0) {
		if (chdir(dirname) != 0) {
			FATAL_ERROR(("cannot chdir to %s", dirname));
		}
		copy_str(p->directory, (const char *)dirname, sizeof p->directory);
	}

	file_command(&Open_file, 0, "Open MPEG file:", COM_OPEN_STREAM_URL);
}

/*--------------------------------------------------------------------------+
| url_open																	|
+--------------------------------------------------------------------------*/
static void
url_open()
{
	static const char *default_url = "http://";
	const char *url;

	reset_auto_seek();
	reset_loop();


	/* get file name with standard Xforms file selector */
	url = fl_show_input ("Enter the location (URL) of an MPEG file:", default_url);
	
	if (url != NULL) {
	
		G2A_SEND_STRING(COM_OPEN_STREAM_URL, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, url);
		
		default_url = url;
	}
}


/*--------------------------------------------------------------------------+
| file_open																	|
+--------------------------------------------------------------------------*/
static void
file_open()
{
	reset_auto_seek();
	reset_loop();
	file_command(&Open_file, 0, "Open MPEG file:", COM_OPEN_STREAM_URL);
}


/*--------------------------------------------------------------------------+
| file_save																	|
+--------------------------------------------------------------------------*/
static void
file_save()
{
	file_command(&Save_file, 1, "Save MPEG file as:", COM_SAVE_STREAM);
}

/*--------------------------------------------------------------------------+
| image_save_ppm															|
+--------------------------------------------------------------------------*/
static void
image_save_ppm()
{
	file_command(&Save_image_ppm, 2, "Save Image in PPM file:", COM_SAVE_IMAGE_PPM);
}


/*--------------------------------------------------------------------------+
| image_save_tga															|
+--------------------------------------------------------------------------*/
static void
image_save_tga()
{
	file_command(&Save_image_tga, 3, "Save Image in TGA file:", COM_SAVE_IMAGE_TGA);
}


/*--------------------------------------------------------------------------+
| hide_seek_controls														|
+--------------------------------------------------------------------------*/
static void 
hide_seek_controls()
{
	fl_hide_object(Controls->frame_slider);
	fl_hide_object(Controls->time_total);
	fl_hide_object(Controls->rewind);
	fl_hide_object(Controls->end);
	fl_hide_object(Controls->loop);

	fl_set_slider_value(Controls->frame_slider, 0.0);
	
	fl_set_menu_item_mode(Controls->file_menu, FILE_MENU_SAVE_AS, FL_PUP_GREY);
}

/*--------------------------------------------------------------------------+
| show_seek_controls														|
+--------------------------------------------------------------------------*/
static void 
show_seek_controls()
{
	fl_show_object(Controls->frame_slider);
	fl_show_object(Controls->time_elapsed);
	fl_show_object(Controls->time_total);
	fl_show_object(Controls->rewind);
	fl_show_object(Controls->end);
	fl_show_object(Controls->loop);

	/*
	 * The player currently resets the Loop mode when it plays a non-seekable
	 * stream.  Since it is possible that this happened, we must
	 * send the Loop mode again.
	 */
	G2A_SEND_INT(COM_LOOP, 0, fl_get_button(Controls->loop));

	fl_set_menu_item_mode(Controls->file_menu, FILE_MENU_SAVE_AS, FL_PUP_NONE);
}

/*--------------------------------------------------------------------------+
| disable_audio_infos														|
+--------------------------------------------------------------------------*/
static void 
disable_audio_infos()
{
	fl_set_object_label(Infos->mux_bitrate, "n/a");
	fl_set_object_label(Infos->audio_bitrate, "n/a");
	fl_set_object_label(Infos->audio_mode, "n/a");
	fl_set_object_label(Infos->audio_layer, "n/a");
	fl_set_object_label(Infos->audio_frequency, "n/a");
	fl_set_object_label(Infos->audio_copyright, "n/a");
	fl_set_object_label(Infos->audio_original, "n/a");
	fl_set_object_label(Infos->audio_protected, "n/a");
}

/*--------------------------------------------------------------------------+
| disable_stream_infos														|
+--------------------------------------------------------------------------*/
static void 
disable_stream_infos()
{
#if 0 /* Tristan Savatier - Wed Feb 25 03:03:59 PST 1998 */
//	printf("disable_stream_infos\n");
#endif /* 0 - Tristan Savatier - Wed Feb 25 03:03:59 PST 1998 */
	
	Bits_filesize = -1;
	Video_bitrate = 0;
	Mux_bitrate = 0;
	
	fl_set_object_label(Infos->file_name, "n/a");
	fl_set_object_label(Infos->file_size, "n/a");
	fl_set_object_label(Infos->video_width, "n/a");
	fl_set_object_label(Infos->video_height, "n/a");
	fl_set_object_label(Infos->video_bitrate, "n/a");
	fl_set_object_label(Infos->video_framerate, "n/a");
	fl_set_object_label(Infos->video_vbv_buffer_size, "n/a");
	fl_set_object_label(Infos->video_constrained, "n/a");
	fl_set_object_label(Infos->video_pel_aspect_ratio, "n/a");
	fl_set_object_label(Infos->video_compliant, "n/a");
	disable_audio_infos();
}

/*--------------------------------------------------------------------------+
| hide_audio_controls														|
+--------------------------------------------------------------------------*/
static void 
hide_audio_controls()
{
	fl_hide_object(Controls->audio_mute);

	fl_hide_object(Controls->audio_volume_slider);
	fl_hide_object(Controls->audio_balance_slider);
	
	Audio_is_available = False;
}

/*--------------------------------------------------------------------------+
| show_audio_controls														|
+--------------------------------------------------------------------------*/
static void 
show_audio_controls()
{
	fl_show_object(Controls->audio_mute);

#if !(defined(LINUX) && defined(__powerpc__))
	fl_show_object(Controls->audio_volume_slider);
	fl_show_object(Controls->audio_balance_slider);
#endif

	Audio_is_available = True;
}



/*--------------------------------------------------------------------------+
| hide_stream_controls														|
+--------------------------------------------------------------------------*/
static void 
hide_stream_controls()
{
	fl_hide_object(Controls->actual_framerate);
	fl_hide_object(Controls->actual_framerate_label);
	fl_hide_object(Controls->play_pause);
	fl_hide_object(Controls->stop);
	fl_hide_object(Controls->step);
	fl_hide_object(Controls->fast_forward);
	fl_hide_object(Controls->zoom);
	fl_hide_object(Controls->number_of_frames_played);
	fl_hide_object(Controls->number_of_frames_played_label);

	fl_hide_object(Controls->time_elapsed);
	
	fl_hide_object(Controls->vcd_track);
	
	fl_set_menu_item_mode(Controls->file_menu, FILE_MENU_INFOS, FL_PUP_GREY);
	fl_set_menu_item_mode(Controls->file_menu, FILE_MENU_CLOSE, FL_PUP_GREY);
#ifdef USE_SDL
	fl_set_menu_item_mode(Controls->window_menu, WINDOW_MENU_FULLSCREEN, FL_PUP_GREY);
#endif
	fl_set_menu_item_mode(Controls->window_menu, WINDOW_MENU_RAISE, FL_PUP_GREY);
	fl_set_menu_item_mode(Controls->window_menu, WINDOW_MENU_FRAME_INFOS, FL_PUP_GREY);
	fl_set_menu_item_mode(Controls->window_menu, WINDOW_MENU_SAVE_PPM, FL_PUP_GREY);
	fl_set_menu_item_mode(Controls->window_menu, WINDOW_MENU_SAVE_TGA, FL_PUP_GREY);

	hide_seek_controls();
	hide_audio_controls();
	disable_stream_infos();
}

/*--------------------------------------------------------------------------+
| show_stream_controls														|
+--------------------------------------------------------------------------*/
static void 
show_stream_controls()
{
	fl_show_object(Controls->actual_framerate);
	fl_show_object(Controls->actual_framerate_label);
	fl_show_object(Controls->play_pause);
	fl_show_object(Controls->stop);
	fl_show_object(Controls->step);
	if (fl_get_button(Sync->sync_sync) || fl_get_button(Sync->sync_nominal)) {
		fl_show_object(Controls->fast_forward);
	}
	fl_show_object(Controls->zoom);
	fl_show_object(Controls->number_of_frames_played);
	fl_show_object(Controls->number_of_frames_played_label);

	fl_show_object(Controls->time_elapsed);
	
	fl_set_menu_item_mode(Controls->file_menu, FILE_MENU_INFOS, FL_PUP_NONE);
	fl_set_menu_item_mode(Controls->file_menu, FILE_MENU_CLOSE, FL_PUP_NONE);
	fl_set_menu_item_mode(Controls->window_menu, WINDOW_MENU_FRAME_INFOS, FL_PUP_NONE);
	fl_set_menu_item_mode(Controls->window_menu, WINDOW_MENU_SAVE_PPM, FL_PUP_NONE);
	fl_set_menu_item_mode(Controls->window_menu, WINDOW_MENU_SAVE_TGA, FL_PUP_NONE);
}

/*--------------------------------------------------------------------------+
| u2str																		|
+--------------------------------------------------------------------------*/
static const char *
u2str(
	u_int	n
)
{
	static char buffer[11];
	
	/* buffer[sizeof buffer - 1] = 0; */
	sprintf(buffer, "%u", n);
	if (buffer[sizeof buffer - 1] != 0) abort_now();
	
	return buffer;
}

/*--------------------------------------------------------------------------+
| d2str																		|
+--------------------------------------------------------------------------*/
static const char *
d2str(
	double	d,
	int		decimals
)
{
	static char buffer[16];

	/* buffer[sizeof buffer - 1] = 0; */
	sprintf(buffer, "%1.*f", decimals, d);
	if (buffer[sizeof buffer - 1] != 0) abort_now();

	return buffer;
}

/*--------------------------------------------------------------------------+
| clear_dither_buttons														|
+--------------------------------------------------------------------------*/
static void
clear_dither_buttons()
{
	fl_set_button(Video->video_dither_none, 0);
	fl_set_button(Video->video_dither_gray, 0);
	fl_set_button(Video->video_dither_ordered, 0);
	fl_set_button(Video->video_dither_color, 0);
}

/*--------------------------------------------------------------------------+
| set_dither_type_controls													|
+--------------------------------------------------------------------------*/
static void
set_dither_type_controls(
	DITHER_TYPE		dither_type
)
{
	clear_dither_buttons();

	switch (dither_type) {
		case DITHER_NONE:
			fl_set_button(Video->video_dither_none, 1);
#ifdef USE_SDL
			fl_set_menu_item_mode(Controls->window_menu, WINDOW_MENU_FULLSCREEN, FL_PUP_GREY);
#endif
			break;
		case DITHER_GRAY:
			fl_set_button(Video->video_dither_gray, 1);
#ifdef USE_SDL
			fl_set_menu_item_mode(Controls->window_menu, WINDOW_MENU_FULLSCREEN, FL_PUP_GREY);
#endif
			break;
		case DITHER_ORDERED:
			fl_set_button(Video->video_dither_ordered, 1);
#ifdef USE_SDL
			fl_set_menu_item_mode(Controls->window_menu, WINDOW_MENU_FULLSCREEN, FL_PUP_GREY);
#endif
			break;
		case DITHER_COLOR:
			fl_set_button(Video->video_dither_color, 1);
#ifdef USE_SDL
			fl_set_menu_item_mode(Controls->window_menu, WINDOW_MENU_FULLSCREEN, FL_PUP_NONE);
#endif
			break;
		default:
			printf("mtv: bad dither type: %d\n", dither_type);
			abort_now();
	}
}

/*--------------------------------------------------------------------------+
| set_sync_type_controls													|
+--------------------------------------------------------------------------*/
static void
set_sync_type_controls(
	SYNC_TYPE sync_type
)
{
	FL_OBJECT	*obj;
		
	Sync_type = sync_type;
	switch (sync_type) {
		case SYNC_FRAMERATE:	if (Framerate_is_nominal) {
									obj = Sync->sync_nominal;
								} else {
									obj = Sync->sync_sync;
								}
								break;
		case SYNC_I:			obj = Sync->sync_i; break;
		case SYNC_IP:			obj = Sync->sync_ip; break;
		case SYNC_IPB:			obj = Sync->sync_ipb; break;
		default:	abort_now();
	}
	fl_set_button(obj, 1);
}

/*--------------------------------------------------------------------------+
| set_file_selector_pattern													|
+--------------------------------------------------------------------------*/
static void
set_file_selector_pattern(
	const char *filename
)
{
	const char *name;	/* last part of file pathname */
	const char *extension;	/* file extension, if any */
	const char *p;
	char    pattern[PATH_MAX];

	assert(filename != NULL);
	
	if ((p = rindex(filename, (int)'/')) == NULL) {
		name = filename;
	} else {
		name = p + 1;
	}
	
	pattern[sizeof pattern - 1] = '\0';
	
	if ((extension = rindex(name, (int)'.')) != NULL) {
		sprintf (pattern, "*%s", extension);

		if (pattern[sizeof pattern - 1] != '\0') {
			/* buffer overflow! */
			abort_now();
		}
	
		copy_str(Open_file.pattern, pattern, sizeof Open_file.pattern);
	}
}

/*--------------------------------------------------------------------------+
| setup_after_file_open														|
+--------------------------------------------------------------------------*/
static void
setup_after_file_open(
	const char *filename
)
{
	char dir[PATH_MAX];
	const char *name;	/* last part of file pathname */
	const char *p;
	char    window_title[PATH_MAX];

	assert(filename != NULL);

	/*
	 * Reset bar right now makes the player feel more
	 * reactive, because seek can actually take some time...
	 */
	fl_set_slider_value(Controls->frame_slider, 0.0);
	disable_audio_infos();
	disable_stream_infos();
	show_stream_controls();

	window_title[sizeof window_title - 1] = '\0';

	if (strncasecmp(filename, "file:", 5) == 0) {
		filename += 5;
	}

	if ((p = rindex(filename, (int)'/')) == NULL) {
		strcpy(dir, ".");
		name = filename;
		sprintf (window_title, " %s - %s, dir: .", MpegTV, filename);
	} else {
		strncpy(dir, filename, (size_t)(p - filename));
		dir[p - filename] = 0;
		name = p + 1;
		sprintf (window_title, " %s - %s, dir: %.*s", MpegTV, name, name - filename, filename);
	}
	
	if (window_title[sizeof window_title - 1] != '\0') {
		/* buffer overflow! */
		abort_now();
	}

	/* set the Player's window and icon name */
	G2A_SEND_STRING(COM_WINDOW_NAME, 0, window_title);
	G2A_SEND_STRING(COM_ICON_NAME, 0, name);
	
	fl_set_object_label(Infos->file_name, name);

	copy_str(Open_file.filename, name, sizeof Open_file.filename);
	copy_str(Open_file.directory, dir, sizeof Open_file.directory);
}


/*--------------------------------------------------------------------------+
| set_sync_sync																|
+--------------------------------------------------------------------------*/
static void
set_sync_sync()
{
	/* fast forward enabled */
	if (Controls->play_pause->visible) {
		fl_show_object(Controls->fast_forward);
	}
	G2A_SEND_INT(COM_SYNC_TYPE, 0, SYNC_FRAMERATE);
	Sync_type = SYNC_FRAMERATE;
	if (Specified_framerate_is_nominal) {
		G2A_SEND_DOUBLE(COM_SYNC_FRAMERATE, 0, 0.0);
		Framerate_is_nominal = True;
	} else {
		G2A_SEND_DOUBLE(COM_SYNC_FRAMERATE, 0, (double)fl_get_slider_value(Sync->sync_framerate_slider));
		Framerate_is_nominal = False;
	}
}

/*--------------------------------------------------------------------------+
| set_audio_channel_controls												|
+--------------------------------------------------------------------------*/
static void
set_audio_channel_controls(
	AUDIO_CHANNEL	audio_channel
)
{
	Audio_channel = audio_channel;
	switch (audio_channel) {
		case AUDIO_CHANNEL_LEFT: fl_set_button(Audio->audio_ch_left, 1); break;
		case AUDIO_CHANNEL_RIGHT: fl_set_button(Audio->audio_ch_right, 1); break;
		case AUDIO_CHANNEL_BOTH: fl_set_button(Audio->audio_ch_both, 1); break;
		default:	abort_now();
	}
}

/*--------------------------------------------------------------------------+
| set_audio_quality_controls												|
+--------------------------------------------------------------------------*/
static void
set_audio_quality_controls(
	AUDIO_QUALITY	audio_quality
)
{
	Audio_quality = audio_quality;
	switch (audio_quality) {
		case AUDIO_QUALITY_LOW: fl_set_button(Audio->audio_quality_low, 1); break;
		case AUDIO_QUALITY_MEDIUM: fl_set_button(Audio->audio_quality_medium, 1); break;
		case AUDIO_QUALITY_HIGH: fl_set_button(Audio->audio_quality_high, 1); break;
		default:	abort_now();
	}
}

/*--------------------------------------------------------------------------+
| save_preferences															|
+--------------------------------------------------------------------------*/
int
save_preferences()
{
	FILE	*file;
	
	file = fopen(User_options.pref_filename, "w");
	
	if (file == NULL) {
		show_alert_message("Cannot write preference file:", User_options.pref_filename);
#if 0 /* Tristan Savatier - Sun Dec 7 18:38:39 PST 1997 */
//		MESSAGE_ERROR(("cannot write preference file: %s", User_options.pref_filename));
#endif /* 0 - Tristan Savatier - Sun Dec 7 18:38:39 PST 1997 */
		return -1;
	}
	
	fprintf(file, "# MpegTV Player 1.0 Preferences File\n");
	fprintf(file, "# %s\n", VERSION_STR);
	fprintf(file, "# This file is generated automatically by mtv! Do not edit.\n");
	fprintf(file, "\n");
	fprintf(file, "%s: %s\n", PREF_LOOP, fl_get_button(Controls->loop) ? "True" : "False");
	fprintf(file, "%s: %u\n", PREF_ZOOM, fl_get_button(Controls->zoom));
#if 0 /* Tristan Savatier - Sat Jan 31 16:24:47 PST 1998 */
//	fprintf(file, "%s: %s\n", PREF_AUTO_PLAY, fl_get_button(Controls->auto_play) ? "True" : "False");
#endif /* 0 - Tristan Savatier - Sat Jan 31 16:24:47 PST 1998 */
	fprintf(file, "%s: %s\n", PREF_AUTO_PLAY, (fl_get_menu_item_mode(Controls->options_menu, OPTIONS_MENU_AUTO_PLAY) & FL_PUP_CHECK) ? "True" : "False");

#if 0 /* Tristan Savatier - Fri Jan 16 02:36:42 PST 1998 */
//	fprintf(file, "%s: %s\n", PREF_AUTO_EXIT, User_options.auto_exit ? "True" : "False");
#endif /* 0 - Tristan Savatier - Fri Jan 16 02:36:42 PST 1998 */
	fprintf(file, "%s: %6.4f\n", PREF_VIDEO_GAMMA, fl_get_slider_value(Video->video_gamma_slider));
	fprintf(file, "%s: %6.4f\n", PREF_VIDEO_LUMA_CONTRAST, fl_get_slider_value(Video->video_contrast_slider));
	fprintf(file, "%s: %6.4f\n", PREF_VIDEO_LUMA_BRIGHTNESS, fl_get_slider_value(Video->video_brightness_slider));
	fprintf(file, "%s: %6.4f\n", PREF_VIDEO_CHROMA_SATURATION, fl_get_slider_value(Video->video_chroma_saturation_slider));
	fprintf(file, "%s: %u\n", PREF_VIDEO_ASPECT_RATIO_MODE, Aspect_ratio_mode);
	fprintf(file, "%s: %s\n", PREF_VIDEO_PRIVATE_COLORMAP, fl_get_button(Video->video_dither_ordered_colormap_private) ? "True" : "False");
	fprintf(file, "%s: %d\n", PREF_MTVP_WINDOW_POSITION_X, User_options.mtvp_geometry_x);
	fprintf(file, "%s: %d\n", PREF_MTVP_WINDOW_POSITION_Y, User_options.mtvp_geometry_y);
	fprintf(file, "%s: %u\n", PREF_AUDIO_CHANNEL, Audio_channel);
	fprintf(file, "%s: %u\n", PREF_AUDIO_QUALITY, Audio_quality);
	fprintf(file, "%s: %u\n", PREF_SYNC_TYPE, Sync_type);
	fprintf(file, "%s: %6.4f\n", PREF_SYNC_FRAMERATE, Framerate_is_nominal ? 0 : fl_get_slider_value(Sync->sync_framerate_slider));

	if (Registered) {
		fprintf(file, "%s: %s\n", PREF_REGISTRATION_NAME, Name);
		fprintf(file, "%s: %s\n", PREF_REGISTRATION_ORDER_NUMBER, Order_number);
		fprintf(file, "%s: %s\n", PREF_REGISTRATION_KEY, Key);
	}

	fclose(file);
	
	return 0;
}

/*--------------------------------------------------------------------------+
| show_form_controls														|
+--------------------------------------------------------------------------*/
static void
show_form_controls()
{
	show_and_raise_form(Controls->controls, User_options.iconify ? FL_PLACE_ICONIC : FL_PLACE_POSITION, FL_FULLBORDER, "MpegTV Controls");
}

/*--------------------------------------------------------------------------+
| process_parameters_and_options											|
+--------------------------------------------------------------------------*/
static void
process_parameters_and_options()
{
	static Bool done;
		
	if (done) {
		if (Must_save_preferences) {
			save_preferences();
			Must_save_preferences = False;
		}
		return;
	}
	/*
	 * Ask to receive ButtonRelease events in Player's window.
	 * On some systems (Solaris, IRIX) it seems to be necessary
	 * to ask for ButtonPress events in order to get the ButtonRelease
	 * events. No idea why! Probably a bug in X.
	 */
	G2A_SEND_INT(COM_X_EVENT_MASK, 0, (int)(ButtonPressMask|ButtonReleaseMask));

	if (User_options.audio_mode == AUDIO_MODE_NONE) {
		fl_set_menu_item_mode(Controls->options_menu, OPTIONS_MENU_AUDIO, FL_PUP_GREY);
	}

	G2A_SEND_INT(COM_LOOP, 0, (int)User_options.loop);
	fl_set_button(Controls->loop, User_options.loop);

	G2A_SEND_INT(COM_ZOOM, 0, (int)User_options.zoom);
	fl_set_button(Controls->zoom, (int)User_options.zoom);
	
	G2A_SEND_INT(COM_VIDEO_DITHER_COLORMAP_PRIVATE, 0, (int)User_options.private_colormap);
	fl_set_button(Video->video_dither_ordered_colormap_private, (int)User_options.private_colormap);
	
#if 0 /* Tristan Savatier - Fri Jan 16 02:37:08 PST 1998 */
//	G2A_SEND_INT(COM_AUTO_EXIT, 0, (int)User_options.auto_exit);
#endif /* 0 - Tristan Savatier - Fri Jan 16 02:37:08 PST 1998 */
	/* fl_set_button(Controls->auto_exit, User_options.auto_exit); */

	G2A_SEND_INT(COM_AUTO_PLAY, 0, (int)User_options.auto_play);
#if 0 /* Tristan Savatier - Sat Jan 31 16:33:12 PST 1998 */
//	fl_set_button(Controls->auto_play, User_options.auto_play);
#endif /* 0 - Tristan Savatier - Sat Jan 31 16:33:12 PST 1998 */
	fl_set_menu_item_mode(Controls->options_menu, OPTIONS_MENU_AUTO_PLAY, User_options.auto_play ? (FL_PUP_BOX|FL_PUP_CHECK) : FL_PUP_BOX);

	if (User_options.dither_type != (DITHER_TYPE)-1) {
		G2A_SEND_INT(COM_VIDEO_DITHER_TYPE, 0, (int)User_options.dither_type);
	}

	if (User_options.sync_framerate == 0) {
		Framerate_is_nominal = True;
	} else {
		fl_set_slider_value(Sync->sync_framerate_slider, User_options.sync_framerate);
		Specified_framerate_is_nominal = False;
		Framerate_is_nominal = False;
		
		/*
		 * Maybe we should the audio is not played by setting the mute button ?
		 *
		 * Problem: maybe in a future version mtvp will be smart enough
		 * to play audio at other than the nominal framerate. For now on,
		 * we assume that if framerate is set to something
		 * other than nominal framerate, audio is not played, and to
		 * show that, we set the mute button.
		 * There should be a way to know whether audio is being played or
		 * not, but the current API does not have that... unfortunately.
		 */
	}
	G2A_SEND_DOUBLE(COM_SYNC_FRAMERATE, 0, User_options.sync_framerate);
	G2A_SEND_INT(COM_SYNC_TYPE, 0, (int)User_options.sync_type);
	set_sync_type_controls((SYNC_TYPE)User_options.sync_type);
	

	/*
	 * We dont need to use COM_MSG_SELECT anymore because
	 * the message selector mask
	 * is now passed through option -w when mtvp is started.
	 */
	/* G2A_SEND_INT(COM_MSG_SELECT, 0, (int)User_options.msg_select); */
	
	fl_set_slider_value(Video->video_gamma_slider, User_options.gamma);
	command(Video->video_gamma_slider, COM_VIDEO_GAMMA_SLIDER);
	
	fl_set_slider_value(Video->video_contrast_slider, User_options.contrast);
	command(Video->video_contrast_slider, COM_VIDEO_CONTRAST_SLIDER);
	
	fl_set_slider_value(Video->video_brightness_slider, User_options.brightness);
	command(Video->video_brightness_slider, COM_VIDEO_BRIGHTNESS_SLIDER);

	fl_set_slider_value(Video->video_chroma_saturation_slider, User_options.chroma_saturation);
	command(Video->video_chroma_saturation_slider, COM_VIDEO_CHROMA_SATURATION_SLIDER);

#ifdef USE_XIL
	switch (Aspect_ratio_mode) {
		case ASPECT_RATIO_SQUARE_PELS:
			G2A_SEND_INT(COM_VIDEO_ASPECT_RATIO, 0, ASPECT_RATIO_SQUARE_PELS);
			fl_set_button(Video->video_aspect_ratio_square_pels, 1);
			break;
		case ASPECT_RATIO_CODED:
			G2A_SEND_INT(COM_VIDEO_ASPECT_RATIO, 0, ASPECT_RATIO_CODED);
			fl_set_button(Video->video_aspect_ratio_coded, 1);
			break;
		case ASPECT_RATIO_FREE:
			G2A_SEND_INT(COM_VIDEO_ASPECT_RATIO, 0, ASPECT_RATIO_FREE);
			fl_set_button(Video->video_aspect_ratio_free, 1);
			break;
		default:	abort_now();			
	}
#else
	G2A_SEND_INT(COM_VIDEO_ASPECT_RATIO, 0, ASPECT_RATIO_SQUARE_PELS);
	fl_set_button(Video->video_aspect_ratio_square_pels, 1);
#endif

	switch (User_options.video_layout) {
		case VIDEO_LAYOUT_FILL_WINDOW:
			G2A_SEND_INT(COM_VIDEO_LAYOUT, 0, VIDEO_LAYOUT_FILL_WINDOW);
			fl_set_button(Video->video_layout_fill_window, 1);
			break;
		case VIDEO_LAYOUT_CENTERED:
			G2A_SEND_INT(COM_VIDEO_LAYOUT, 0, VIDEO_LAYOUT_CENTERED);
			fl_set_button(Video->video_layout_centered, 1);
			break;
		case VIDEO_LAYOUT_TOP_LEFT:
			G2A_SEND_INT(COM_VIDEO_LAYOUT, 0, VIDEO_LAYOUT_TOP_LEFT);
			fl_set_button(Video->video_layout_top_left, 1);
			break;
		default:	abort_now();			
	}
	
	/*
	 * We dont need to use COM_AUDIO_MSG_SELECT anymore because
	 * the audio message selector mask
	 * is now passed through option -aw when mtvp is started.
	 */
	/* G2A_SEND_INT(COM_AUDIO_MSG_SELECT, 0, (int)User_options.audio_msg_select); */

	G2A_SEND_INT(COM_AUDIO_QUALITY, 0, (int)User_options.audio_quality);
	set_audio_quality_controls(User_options.audio_quality);
	
	G2A_SEND_INT(COM_AUDIO_CHANNEL, 0, (int)User_options.audio_channel);
	set_audio_channel_controls(User_options.audio_channel);

	G2A_SEND_INT(COM_AUDIO_CONTROLS, 0, AUDIO_CONTROLS_PCM);

	if (User_options.audio_pcm_volume >= 0) {
		G2A_SEND_DOUBLE(COM_AUDIO_VOLUME, 0, User_options.audio_pcm_volume);
	}
	if (User_options.audio_pcm_balance >= 0) {
		G2A_SEND_DOUBLE(COM_AUDIO_BALANCE, 0, User_options.audio_pcm_balance);
	}
	
	G2A_SEND_INT(COM_AUDIO_CONTROLS, 0, AUDIO_CONTROLS_MASTER);

	if (User_options.audio_master_volume >= 0) {
		G2A_SEND_DOUBLE(COM_AUDIO_VOLUME, 0, User_options.audio_master_volume);
	}
	if (User_options.audio_master_balance >= 0) {
		G2A_SEND_DOUBLE(COM_AUDIO_BALANCE, 0, User_options.audio_master_balance);
	}

	switch (User_options.audio_controls) {
		case AUDIO_CONTROLS_MASTER:
			G2A_SEND_INT(COM_AUDIO_CONTROLS, 0, AUDIO_CONTROLS_MASTER);
			fl_set_button(Audio->audio_controls_master, 1);
			break;
		case AUDIO_CONTROLS_PCM:
			G2A_SEND_INT(COM_AUDIO_CONTROLS, 0, AUDIO_CONTROLS_PCM);
			fl_set_button(Audio->audio_controls_pcm, 1);
			break;
		default:	abort_now();			
	}
	if (User_options.audio_ports != -1) {
		G2A_SEND_INT(COM_AUDIO_PORTS_SELECT, 0, User_options.audio_ports);
	}

	if (User_options.loop_start_pos != 0) {
		G2A_SEND_INT(COM_AUTO_SEEK, 0, (int)(u_int)User_options.loop_start_pos);
		G2A_SEND_INT(COM_LOOP_START_POS, 0, (int)(u_int)User_options.loop_start_pos);
	}
	if (User_options.loop_end_pos != 0) {
		G2A_SEND_INT(COM_LOOP_END_POS, 0, (int)(u_int)User_options.loop_end_pos);
	}

	G2A_SEND_INT(COM_WINDOW_POSITION_X, 0, User_options.mtvp_geometry_x);
	G2A_SEND_INT(COM_WINDOW_POSITION_Y, 0, User_options.mtvp_geometry_y);
	
	if (User_options.existing_window == -1) {
		fl_set_menu_item_mode(Controls->window_menu, WINDOW_MENU_ROOT, FL_PUP_BOX|FL_PUP_CHECK);
	}
	if (User_options.existing_window != 0) {
		G2A_SEND_INT(COM_WINDOW_ID, 0, User_options.existing_window);
	}
	
#if 0 /* Tristan Savatier - Sun Mar 22 02:02:26 PST 1998 */
//	G2A_SEND_INT(COM_VIDEO_LAYOUT, 0, User_options.fill_window ? VIDEO_LAYOUT_FILL_WINDOW : VIDEO_LAYOUT_CENTERED);
#endif /* 0 - Tristan Savatier - Sun Mar 22 02:02:26 PST 1998 */
	
	if (Must_save_preferences) {
		save_preferences();
		Must_save_preferences = False;
	}

	if (User_options.url != NULL) {
		int		fd;
		struct stat stat_buf;
	
		fd = open(User_options.url, O_RDONLY|O_BINARY);
		if (fd != -1) {
		
			if (fstat(fd, &stat_buf) == 0) {

				if (S_ISDIR(stat_buf.st_mode)) {
		
					select_file_in_directory(User_options.url);
					goto ret;
				} else {
		
					set_file_selector_pattern(User_options.url);
				}
			}
			close(fd);
		}

		G2A_SEND_STRING(COM_OPEN_STREAM_URL, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, User_options.url);
	}

ret:
	done = True;
}

/*--------------------------------------------------------------------------+
| registration_order_number_filter											|
+--------------------------------------------------------------------------*/
static int
registration_order_number_filter(
	FL_OBJECT	*ob,
	const char *old,
	const char *cur,
	int		c
)
{
	if (isdigit(c)) {
		if (isdigit(*cur)) {
			return FL_INVALID | FL_RINGBELL;
		}
	} else {
#if 0 /* Tristan Savatier - Wed Mar 4 22:22:08 PST 1998 */
//		if (c != 'I') {
#endif /* 0 - Tristan Savatier - Wed Mar 4 22:22:09 PST 1998 */
		if (!isupper(c)) {
			return FL_INVALID | FL_RINGBELL;
		}
		if (cur[0] != c || cur[1] != 0) {
			return FL_INVALID | FL_RINGBELL;
		}
	}
	return FL_VALID;
}

/*--------------------------------------------------------------------------+
| registration_key_filter													|
+--------------------------------------------------------------------------*/
static int
registration_key_filter(
	FL_OBJECT	*ob,
	const char *old,
	const char *cur,
	int		c
)
{
#if 0 /* Tristan Savatier - Wed Oct 28 17:33:12 PST 1998 */
//	if (isdigit(c) || (c >= 'a' && c <= 'f') || (c >= 'A' && c <= 'F')) {
#endif /* 0 - Tristan Savatier - Wed Oct 28 17:33:12 PST 1998 */
	if (isalnum(c) || c == '-') {
		return FL_VALID;
	} else {
		return FL_INVALID | FL_RINGBELL;
	}
}

/*--------------------------------------------------------------------------+
| remove_splash_and_start_player											|
+--------------------------------------------------------------------------*/
static void
remove_splash_and_start_player()
{
	fl_hide_form(Splash->splash);
#if 0 /* Tristan Savatier - Sun Jan 11 22:58:19 PST 1998 */
//			fl_activate_form(Controls->controls);
#endif /* 0 - Tristan Savatier - Sun Jan 11 22:58:19 PST 1998 */
	show_form_controls();
	process_parameters_and_options();
}

#ifdef HANDLE_CHALLENGE
/*--------------------------------------------------------------------------+
| check_key																	|
+--------------------------------------------------------------------------*/
static Bool
check_key(
	const char *ascii_key,
	LicenseKey *key,
	Bool alert
)
{	
    int result;
    Bool	status = False;
    
    
    result  = keyformat_ascii_to_key(ascii_key, key);

#if 0 /* Tristan Savatier - Thu Nov 5 02:29:05 PST 1998 */
//    printf("checking key %s, result = %d\n", ascii_key, result);
#endif /* 0 - Tristan Savatier - Thu Nov 5 02:29:05 PST 1998 */

    switch(result) {
      case KEY_STATUS_SUCCESS:
        status = True;
      	/*
         * see if the key matches this target
         * we expect a licensee code & licensee mask to be non zero
         * we expect a product code == KEY_PRODUCT_CODE_MTV_PLAYER
         */
#ifdef KEY_LICENSEE_MASK
        if ((key->licensee & KEY_LICENSEE_MASK) == 0) break;
#endif
        if (key->product != KEY_PRODUCT_MTV_PLAYER) {
			if (alert) show_alert("Incorrect registration information!", ascii_key, "This key is for another product", 1);
        	status = False;
        }
#ifndef LINUX
        if (key->category == KEY_CATEGORY_SHAREWARE) {
			if (alert) show_alert("Incorrect registration information!", ascii_key, "This is a shareware key for Linux only. You need a commercial key", 1);
        	status = False;
        }
#endif
        break;
      case KEY_STATUS_INVALID_SIZE:
		if (alert) show_alert("Incorrect registration information!", ascii_key, "Key size is incorrect", 1);
		break;
      case KEY_STATUS_INVALID_FIELD_SEPARATOR:
		if (alert) show_alert("Incorrect registration information!", ascii_key, "Field separator is incorrect", 1);
		break;
      default:
		if (alert) show_alert("Incorrect registration information!", ascii_key, "Invalid key", 1);
        break;
    }
    if (status && key->category == KEY_CATEGORY_SHAREWARE) {
		if (alert) show_alert("You are using a Shareware license key", "You are restricted to non-commercial and non-governmental use of this product.", "For commercial or governmental use you need a Commercial license key", 1);
    }
    return status;
}

#endif

/*--------------------------------------------------------------------------+
| update_control_title														|
+--------------------------------------------------------------------------*/
static void
update_control_title()
{
	const char *reg_status;
	const char *state;
	char str[256];
	

	if (Registered) {
#ifdef HANDLE_CHALLENGE
		if (License_key.category == KEY_CATEGORY_SHAREWARE) {
			reg_status = "Shareware";
		} else if (License_key.category == KEY_CATEGORY_COMMERCIAL) {
			reg_status = "Commercial";
		} else {
			reg_status = "";
		}
#else
		reg_status = "";
#endif
	} else {
		reg_status = "UNREGISTERED";
	}
	
	switch (Gui_state) {
		case STATE_STOPPED:	state = "Stopped"; break;
		case STATE_PAUSED:	state = "Paused"; break;
		case STATE_PLAYING:	state = "Playing"; break;
		default:			abort_now();
	}
	str[sizeof str - 1] = 0;
	sprintf(str, "MpegTV %s (%s)", reg_status, state);
	if (str[sizeof str - 1] != 0) abort_now();
	
	fl_set_form_title(Controls->controls, str);
}


/*--------------------------------------------------------------------------+
| handle_registration														|
+--------------------------------------------------------------------------*/
static INLINE void
handle_registration()
{
	const char *order_number;
	const char *key;
	
	Name = fl_get_input(Registration->name);
	order_number = fl_get_input(Registration->order_number);
	key = fl_get_input(Registration->key);
	
	if (strlen(order_number) == 10 && strlen(key) == 8) {
		/* old style key */
	

		if (strcmp(order_number, "I000000000") == 0 && strcmp(key, "00000000") == 0) {
			if (fl_show_choice(
				"Do you have a valid mtvp key in mtvp.key ?",
				"Click Yes only if you have such a key and want to use it",
				"Click No if you have not purchased a license or don't have mtvp.key",
				2,
				"Yes",
				"No",
				"",
				1
				) == 1) {
				Registered = True;
				Use_mtvp_key = True;
			}
		} else {
		
#if 0 /* Tristan Savatier - Sat Oct 31 13:08:18 PST 1998 */
//#if defined(SOLARIS) && !defined(__i386__)
//
//			show_alert("Incorrect registration information!", key, "This is a Shareware key. You need a Commercial key for Solaris SPARC.", 1);
//
//#else
#endif /* 0 - Tristan Savatier - Sat Oct 31 13:08:18 PST 1998 */

#ifdef HANDLE_CHALLENGE
			Registered = check_registration_key(order_number, key, &Msg);

			if (Registered) {
#if 0 /* Tristan Savatier - Sat Oct 31 13:08:46 PST 1998 */
//				show_alert("You are using a Shareware license key", "You are restricted to non-commercial and non-governmental use of this product.", "For commercial or governmental use you need a Commercial license key", 1);
#endif /* 0 - Tristan Savatier - Sat Oct 31 13:08:46 PST 1998 */
				License_key.category = 0;	/* unknown */
			} else {
				show_alert("Incorrect registration information!", "Check registration information and try again", "", 1);
			}
#endif
#if 0 /* Tristan Savatier - Sat Oct 31 13:10:41 PST 1998 */
//#endif
#endif /* 0 - Tristan Savatier - Sat Oct 31 13:10:41 PST 1998 */

		}

	} else {
		/* new style key */
		
#ifdef HANDLE_CHALLENGE
		Registered = check_key(key, &License_key, True);
#endif
	}
	
	if (Registered) {
		Order_number = order_number;
		Key = key;
		Must_save_preferences = True;
		remove_splash_and_start_player();

		if (Registration->registration->visible) {
			fl_hide_form(Registration->registration);
		}

#if 0 /* Tristan Savatier - Wed Oct 28 17:04:37 PST 1998 */
//	} else {
//		fl_hide_form(Registration->registration);
#endif /* 0 - Tristan Savatier - Wed Oct 28 17:04:37 PST 1998 */
	}
	update_control_title();
}

#ifdef _REENTRANT
/*--------------------------------------------------------------------------+
| handle_udp																|
+--------------------------------------------------------------------------*/
static INLINE void
handle_udp()
{
	const char *udp_port_str;
	int udp_port;
	PMP_DATA_GENERIC data_generic;
	char *url = data_generic.data_string;
	
	udp_port_str = fl_get_input(Udp->port);
	udp_port = atoi(udp_port_str);
	if (udp_port <= 0) {
		show_alert("Cannot play from UDP socket", "Incorrect port number:", udp_port_str, 1);
		return;
	}

	url[sizeof data_generic.data_string - 1] = 0;
	sprintf(url, "udp:%u", udp_port);
	if (url[sizeof data_generic.data_string - 1] != 0) abort_now();
	
	G2A_SEND_STRING(COM_OPEN_STREAM_URL, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, url);
}
#endif

/*--------------------------------------------------------------------------+
| set_time_total															|
+--------------------------------------------------------------------------*/
static void
set_time_total()
{
	const char *time_str;
	int	sec;

#if 0 /* Tristan Savatier - Wed Feb 25 03:03:47 PST 1998 */
//	printf("set_time_total, Bits_filesize = %u\n", Bits_filesize);
#endif /* 0 - Tristan Savatier - Wed Feb 25 03:03:47 PST 1998 */

	sec = get_elapsed_seconds(Bits_filesize);
	
	time_str = get_time_string(sec);

	fl_set_object_label(Controls->time_total, time_str);
	
	fl_set_object_label(Infos->file_size, Bits_filesize == -1 ? "n/a" : u2str(UDIV_ROUNDED((u_int)Bits_filesize, 1024)));
	
#if 0 /* Tristan Savatier - Wed Feb 25 03:02:16 PST 1998 */
//	printf("Infos->file_size = %s\n", Bits_filesize == -1 ? "n/a" : u2str(UDIV_ROUNDED((u_int)Bits_filesize, 1024)));
#endif /* 0 - Tristan Savatier - Wed Feb 25 03:02:16 PST 1998 */


	if (Bits_filesize != -1) {
		fl_set_slider_bounds(Controls->frame_slider, 0.0, (double)Bits_filesize);
	}
}

/*--------------------------------------------------------------------------+
| set_time_elapsed															|
+--------------------------------------------------------------------------*/
static void
set_time_elapsed(
	int		pos
)
{
	const char *time_str;
	int	sec;

	sec = get_elapsed_seconds(pos);
	time_str = get_time_string(sec);
	
#if 0 /* Tristan Savatier - Mon Aug 18 00:42:25 PDT 1997 */
//	printf("set_time_elapsed: pos = %d, sec = %d, time_str = %s\n", pos, sec, time_str);
#endif /* 0 - Tristan Savatier - Mon Aug 18 00:42:25 PDT 1997 */
	
	fl_set_object_label(Controls->time_elapsed, time_str);
}

/*--------------------------------------------------------------------------+
| disable_frame_infos														|
+--------------------------------------------------------------------------*/
static void
disable_frame_infos()
{
	fl_set_object_label(Frame_infos->frame_type, "n/a");
	fl_set_object_label(Frame_infos->frame_temporal_reference, "n/a");
	fl_set_object_label(Frame_infos->frame_forward_f_code, "n/a");
	fl_set_object_label(Frame_infos->frame_forward_full_pel, "n/a");
	fl_set_object_label(Frame_infos->frame_backward_f_code, "n/a");
	fl_set_object_label(Frame_infos->frame_backward_full_pel, "n/a");
	fl_set_object_label(Frame_infos->frame_vbv_delay, "n/a");
}


#if 0 /* Tristan Savatier - Thu Oct 29 00:16:48 PST 1998 */
///*--------------------------------------------------------------------------+
//| update_control_title														|
//+--------------------------------------------------------------------------*/
//static void
//update_control_title()
//{
//	const char *str;
//
//	if (Registered) {
//		switch (Gui_state) {
//		case STATE_STOPPED:	str = "MpegTV Controls (Stopped)"; break;
//		case STATE_PAUSED:	str = "MpegTV Controls (Paused)"; break;
//		case STATE_PLAYING:	str = "MpegTV Controls (Playing)"; break;
//		default:			abort_now();
//		}
//	} else {
//		switch (Gui_state) {
//		case STATE_STOPPED:	str = "MpegTV [UNREGISTERED] Controls (Stopped)"; break;
//		case STATE_PAUSED:	str = "MpegTV [UNREGISTERED] Controls (Paused)"; break;
//		case STATE_PLAYING:	str = "MpegTV [UNREGISTERED] Controls (Playing)"; break;
//		default:			abort_now();
//		}
//	}
//	fl_set_form_title(Controls->controls, str);
//}
#endif /* 0 - Tristan Savatier - Thu Oct 29 00:16:48 PST 1998 */


/*--------------------------------------------------------------------------+
| update_state																|
+--------------------------------------------------------------------------*/
static void
update_state(
	STATE	state
)
{
#if 0 /* Tristan Savatier - Mon Jan 25 12:21:10 PST 1999 */
//	STATE	previous_state = Gui_state;
#endif /* 0 - Tristan Savatier - Mon Jan 25 12:21:10 PST 1999 */
	
	if (!Registered && Gui_state == STATE_PLAYING && state != STATE_PLAYING) {
		show_alert("Thanks for trying mtv. If you find it useful,", " please REGISTER your copy at www.mpegtv.com", "Registered copies do not display this notice", 1);
	}
	Gui_state = state;
	switch (state) {
	case STATE_STOPPED:
		get_position_parameters();
		fl_set_object_label(Controls->play_pause, "@>");
		fl_set_object_label(Controls->actual_framerate, "n/a");
		Gui_play_pause_set_state = STATE_PLAYING;
		break;
	case STATE_PAUSED:
		get_position_parameters();
		fl_set_object_label(Controls->play_pause, "@>");
		fl_set_object_label(Controls->actual_framerate, "n/a");
		Gui_play_pause_set_state = STATE_PLAYING;
		break;
	case STATE_PLAYING:
		fl_set_object_label(Controls->play_pause, "||");
		Gui_play_pause_set_state = STATE_PAUSED;
		disable_frame_infos();
		Play_start_time = get_time_now();
		break;
	default:
		abort_now();
	}
	update_control_title();
#if 0 /* Tristan Savatier - Mon Jan 25 12:21:48 PST 1999 */
//	if (!Registered && previous_state == STATE_PLAYING && state != STATE_PLAYING) {
//#if 0 /* Tristan Savatier - Mon Jan 25 12:10:20 PST 1999 */
////		show_alert("Thanks for trying MpegTV Player 1.0", "If you find it useful, please REGISTER your copy", "at www.mpegtv.com", 1);
//#endif /* 0 - Tristan Savatier - Mon Jan 25 12:10:20 PST 1999 */
//		fl_show_message("Thanks for trying mtv. If you find it useful,", " please REGISTER your copy at www.mpegtv.com", "Registered copies do not display this notice");
//	}
#endif /* 0 - Tristan Savatier - Mon Jan 25 12:21:48 PST 1999 */
}

/*--------------------------------------------------------------------------+
| send_seek																	|
+--------------------------------------------------------------------------*/
static void
send_seek(
	int		n
)
{
	reset_loop_if_pos_out_of_loop(n);
	G2A_SEND_INT(COM_SEEK, 0, n);
	G2A_SEND_VOID(COM_GET_NUMBER_OF_FRAMES_PLAYED, 0);
	Not_ready = True;
}

#ifdef VCD
/*--------------------------------------------------------------------------+
| close_vcd																	|
+--------------------------------------------------------------------------*/
static void
close_vcd()
{
#if 0 /* Tristan Savatier - Wed Aug 19 15:50:34 PDT 1998 */
//	VCD_CNTL vcd_cntl;
//
//	vcd_cntl.command = VCD_CNTL_COM_CLOSE_DEVICE;
//	G2A_SEND_CUSTOM(COM_VCD_CNTL, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, (const char *)&vcd_cntl, sizeof vcd_cntl);
#endif /* 0 - Tristan Savatier - Wed Aug 19 15:50:34 PDT 1998 */
	
	G2A_SEND_INT(COM_SIH_FREE, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, SIH_HANDLER_ID_VCD);

	fl_hide_object(Controls->vcd_track);
	Vcd_tracks = 0;
}
#endif


/*--------------------------------------------------------------------------+
| process_alert																|
+--------------------------------------------------------------------------*/
static void
process_alert(
	const PMP_DATA_ALERT *data
)
{
	const char *s1;
	const char *s2;
	const char *s3;

	s1 = "Warning!";
	
	if (data->error != 0) {
		s3 = strerror(data->error);
	} else {
		s3 = "";
	}
	switch (data->alert) {
		case ALERT_BAD_STREAM:
			hide_stream_controls();
#ifdef VCD
			if (Vcd_tracks != 0) {
				close_vcd();
				s2 = "Hum... This doesn't look like a VCD track! Check the CD.";
				s3 = "Check that your CDROM drive and kernel support VCD.";
			} else {
#endif
				s2 = "Hum... This doesn't look like an MPEG Systems or Video stream";
				s3 = "Use Xaudio to play MPEG Audio streams (http://www.xaudio.com)";
#ifdef VCD
			}
#endif
			break;
		case ALERT_NO_AUDIO:
			s2 = "Audio device busy or sampl. rate unsup. (muting)";
			break;
		/* Note: ALERT_DEMO has become obsolete */
		/* Note: ALERT_DEMO_MUTE has become obsolete */
		case ALERT_NO_SHARED_MEMORY:
			s2 = "Can't share video memory with X server: performance will be reduced!";
			break;
		case ALERT_USING_PRIVATE_COLORMAP:
			if (fl_get_button(Video->video_dither_ordered_colormap_private) == 1) {
				return;
			}
			fl_set_button(Video->video_dither_ordered_colormap_private, 1);
			s2 = "Not enough colors available in shared colormap";
			s3 = "Using private colormap!";
			break;
		case ALERT_MPEG2_VIDEO:
			hide_stream_controls();
			G2A_SEND_VOID(COM_CLOSE_STREAM, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL);
			s2 = "Hum... This looks like an MPEG-2 stream";
			s3 = "MpegTV Player 1.0 only plays MPEG-1 Systems or Video streams";
			break;
		default:
			s2 = "Unknown alert";
			break;
	}

	show_alert(s1, s2, s3, 1);
}


/*--------------------------------------------------------------------------+
| send_state																|
+--------------------------------------------------------------------------*/
static void
send_state(
	STATE	state
)
{
	G2A_SEND_INT(COM_STATE, 0, (int)state);
}

/*--------------------------------------------------------------------------+
| set_state																	|
+--------------------------------------------------------------------------*/
static void
set_state(
	STATE	state
)
{
	send_state(state);
	update_state(state);
}


/*--------------------------------------------------------------------------+
| read_mtvp_message															|
+--------------------------------------------------------------------------*/
static int
read_mtvp_message(
	int		fd	/* not used because we use the wrapper a2g_read_msg() to read messages */
)
{
	PMP_HEADER	msg;
	PMP_DATA_A2G	data;
	
	if (a2g_read_msg(&Player, &msg, &data) == -1) {
		if (Quitting) {
			clean_exit(EXIT_SUCCESS);
		} else {
			MESSAGE_ERROR(("a2g_read_msg: error reading msg"));
			clean_exit(EXIT_FAILURE);
		}
	}

#if 0 /* Tristan Savatier - Sun Feb 1 02:09:32 PST 1998 */
//	MESSAGE(("received msg: %s", pmp_print_msg(&msg, (const char *)&data)));
#endif /* 0 - Tristan Savatier - Sun Feb 1 02:09:32 PST 1998 */

	switch ((A2G_COM)msg.type) {
		case A2G_COM_READY:
			/*
			 * Any initial command here will be executed by the player with
			 * a smaller delay, since the player is in Ready state.
			 * The player sends the A2G_COM_READY message when it has completed
			 * all its initializations and is ready to work.
			 */
			if (Registered) {
				process_parameters_and_options();
			}
			break;
		case A2G_COM_DISPLAY_READY:
#if 0 /* Tristan Savatier - Tue Apr 28 14:37:06 PDT 1998 */
//			printf("A2G_COM_DISPLAY_READY\n");
#endif /* 0 - Tristan Savatier - Tue Apr 28 14:37:06 PDT 1998 */
			if (!(fl_get_menu_item_mode(Controls->window_menu, WINDOW_MENU_ROOT) & FL_PUP_CHECK)) {
				fl_set_menu_item_mode(Controls->window_menu, WINDOW_MENU_RAISE, FL_PUP_NONE);
			}
			if (Seeking) {
				G2A_SEND_VOID(COM_GET_POS, 0);
			}
			Not_ready = False;
			if (Pending_seek >= 0) {
				send_seek(Pending_seek);
				Pending_seek = -1;
			}
			break;
		case A2G_COM_QUIT:
			/*
			 * mtvp sends A2G_COM_QUIT after receiving COM_QUIT and before exiting.
			 * mtv will terminate when it receives the SIGCHLD just after mtvp has exited.
			 */
			
			/* remove this callback (g2a_cleanup will close this pipe) */
			fl_remove_io_callback(Player.receive_channel.msg_pipe[PMP_FD_READ], FL_READ, (void (*)(int, void *))read_mtvp_message);

			if (Timeout_id != -1) {
				fl_remove_timeout(Timeout_id);
			}

			/* g2a_cleanup() should be called to close pipes to the (now defunct) player */
			g2a_cleanup(&Player);

			break;
		case A2G_COM_STATE:
			if (Seeking) {
				State_before_seeking = (STATE)data.data_generic.data_int;
			} else {
				update_state((STATE)data.data_generic.data_int);
				if (Gui_state != STATE_PLAYING) {
					get_position_parameters();
				}
			}
			break;
		case A2G_COM_POS:
			if (data.data_generic.data_int != -1) {
				Stream_pos = data.data_generic.data_int;
				
#if 0 /* Tristan Savatier - Wed Feb 25 03:04:08 PST 1998 */
//				printf("Stream_pos = %lu\n", (u_long)Stream_pos);
#endif /* 0 - Tristan Savatier - Wed Feb 25 03:04:08 PST 1998 */
				
				set_time_elapsed(Stream_pos);
				if (Bits_filesize != -1 && Stream_pos > Bits_filesize) {
					Bits_filesize = Stream_pos;
					set_time_total();
				}
				if (!Seeking) {
					fl_set_slider_value(Controls->frame_slider, (double)Stream_pos);
				}
			}
			break;
		case A2G_COM_ACTUAL_FRAMERATE:
			{
				double framerate = data.data_generic.data_double;
				const char *str;
				
				if (framerate == 0) {
					str = "n/a";
				} else {
					str = d2str(framerate, 1);
				}
				fl_set_object_label(Controls->actual_framerate, str);
			}
			break;
		case A2G_COM_NUMBER_OF_FRAMES_PLAYED:
			fl_set_object_label(Controls->number_of_frames_played, u2str((u_int)data.data_generic.data_int));
			break;
		case A2G_COM_AUDIO_ENABLED:
			show_audio_controls();
			break;
		case A2G_COM_AUDIO_DISABLED:
			hide_audio_controls();
			break;
		case A2G_COM_ALERT:
			process_alert(&data.data_alert);
			break;
		case A2G_COM_INPUT_FILE_SIZE:
			Bits_filesize = data.data_generic.data_int;
			if (Bits_filesize == -1) {
				hide_seek_controls();
			} else {
#if 0 /* Tristan Savatier - Sun Aug 23 02:34:23 PDT 1998 */
//				printf("Bits_filesize = %d\n", Bits_filesize);
#endif /* 0 - Tristan Savatier - Sun Aug 23 02:34:23 PDT 1998 */
				show_seek_controls();
			}
			set_time_total();
			break;
		case A2G_COM_INPUT_FD:
			break;
		case A2G_COM_INPUT_PORT:
			break;
		case A2G_COM_AUDIO_VOLUME:
			fl_set_slider_value(Controls->audio_volume_slider, data.data_generic.data_double);
			break;
		case A2G_COM_AUDIO_BALANCE:
			fl_set_slider_value(Controls->audio_balance_slider, data.data_generic.data_double);
			break;
		case A2G_COM_AUDIO_MUTE:
			fl_set_button(Controls->audio_mute, 1);
			break;
		case A2G_COM_VIDEO_INFOS:
			{
				const PMP_DATA_VIDEO_INFOS *video_infos = &data.data_video_infos;
				
				fl_set_object_label(Infos->video_width, u2str(video_infos->width));
				fl_set_object_label(Infos->video_height, u2str(video_infos->height));
				
				Video_bitrate = video_infos->bitrate;
				fl_set_object_label(Infos->video_bitrate, Video_bitrate == 0 ? "n/a" : d2str(Video_bitrate*(1.0/1000000), 6));
				set_time_total();

				if (Specified_framerate_is_nominal && Nominal_framerate != video_infos->framerate) {
					Specified_framerate_is_nominal = False;
				}
				Nominal_framerate = video_infos->framerate;
				fl_set_object_label(Infos->video_framerate, d2str(Nominal_framerate, 2));

				/*
				 * Currently mtvp resets automatically the sync framerate to
				 * the nominal framerate when a new file is opened (actually,
				 * it does that when the first sequence header is found). To work
				 * around this problem, mtv resets the framerate to the
				 * currect value after the file has been opened.
				 * This is not very good and may not work in all cases,
				 * because mtvp resets the framerate only once it finds
				 * a sequence header. So if there is garbage at the beginning
				 * of the file, we have a race condition!
				 * The only proper fix to this problem would be to add a new
				 * SYNC mode in mtvp, so that mtvp would know whether or not
				 * it should reset the framerate to nominal.
				 */
				if (fl_get_button(Sync->sync_sync)) {
					if (Specified_framerate_is_nominal) {
						G2A_SEND_DOUBLE(COM_SYNC_FRAMERATE, 0, 0.0);
						Framerate_is_nominal = True;
					} else {
						G2A_SEND_DOUBLE(COM_SYNC_FRAMERATE, 0, (double)fl_get_slider_value(Sync->sync_framerate_slider));
						Framerate_is_nominal = False;
					}
				} else if (fl_get_button(Sync->sync_nominal)) {
					Framerate_is_nominal = True;
				}


				fl_set_object_label(Infos->video_vbv_buffer_size, video_infos->vbv_buffer_size == 0 ? "n/a" : u2str(UDIV_ROUNDED(video_infos->vbv_buffer_size, 1024)));
				fl_set_object_label(Infos->video_constrained, video_infos->constrained_parameter_flag ? "yes" : "no");
				fl_set_object_label(Infos->video_pel_aspect_ratio, d2str(video_infos->pel_aspect_ratio, 4));
				{
					int		mbwidth = (video_infos->width + 15) / 16;
					int		mbheight = (video_infos->height + 15) / 16;
					
					if (video_infos->constrained_parameter_flag
						&& video_infos->width <= CONSTRAINED_MAX_FRAME_WIDTH
						&& video_infos->height <= CONSTRAINED_MAX_FRAME_HEIGHT
						&& mbwidth * mbheight <= CONSTRAINED_MAX_MB_COUNT
						&& mbwidth * mbheight * video_infos->framerate <= CONSTRAINED_MAX_MB_RATE
						&& video_infos->framerate <= CONSTRAINED_MAX_FRAME_RATE
						&& video_infos->bitrate != 0
						&& video_infos->bitrate <= CONSTRAINED_MAX_BIT_RATE
						&& video_infos->vbv_buffer_size != 0
						&& video_infos->vbv_buffer_size <= CONSTRAINED_MAX_VBV_BUFFER_SIZE) {
						fl_set_object_label(Infos->video_compliant, "maybe");
					} else {
						fl_set_object_label(Infos->video_compliant, "no");
					}
				}
			}
			break;
		case A2G_COM_PACK_INFOS:
			Mux_bitrate = data.data_pack_infos.mux_rate;
			fl_set_object_label(Infos->mux_bitrate, Mux_bitrate == 0 ? "n/a" : d2str(Mux_bitrate*(1.0/1000000), 6));
			set_time_total();
			break;
		case A2G_COM_AUDIO_INFOS:
			{
				const PMP_DATA_AUDIO_INFOS	*audio_infos = &data.data_audio_infos;
				const char *str;
				
				fl_set_object_label(Infos->audio_bitrate, u2str(UDIV_ROUNDED(audio_infos->bitrate, 1000)));
				fl_set_object_label(Infos->audio_layer, u2str((~(audio_infos->layer - 1)) & 3));
				
				switch (audio_infos->mode) {
					case MAL_MODE_STEREO:			str = "stereo"; break;
					case MAL_MODE_JOINT_STEREO:		str = "joint stereo"; break;
					case MAL_MODE_DUAL_CHANNEL:		str = "dual channel"; break;
					case MAL_MODE_SINGLE_CHANNEL:	str = "single channel"; break;
					default: 						abort_now();
				}
				fl_set_object_label(Infos->audio_mode, str);

				fl_set_object_label(Infos->audio_frequency, d2str(audio_infos->sampling_frequency/1000.0, 1));
				fl_set_object_label(Infos->audio_copyright, audio_infos->copyright ? "yes" : "no");
				fl_set_object_label(Infos->audio_original, audio_infos->original ? "original" : "copy");
				fl_set_object_label(Infos->audio_protected, audio_infos->protected_flag ? "yes" : "no");
			}
			break;

		case A2G_COM_VIDEO_DITHER_ENABLE:
			switch (data.data_generic.data_int) {
			case DITHER_NONE:
				fl_show_object(Video->video_dither_none);
				break;
			case DITHER_GRAY:
				fl_show_object(Video->video_dither_gray);
				break;
			case DITHER_ORDERED:
				fl_show_object(Video->video_dither_ordered);
				fl_show_object(Video->video_dither_ordered_colormap_private);
				break;
			case DITHER_COLOR:
				fl_show_object(Video->video_dither_color);
				break;
			default:	abort_now();
			}
			break;
		case A2G_COM_VIDEO_DITHER_TYPE:
			set_dither_type_controls((DITHER_TYPE)data.data_generic.data_int);
			break;
		case A2G_COM_VIDEO_COLORS_ALLOCATED:
			/* printf("%d colors allocated\n", data.data_generic.data_int); */
			break;
		case A2G_COM_FRAME_INFOS:
			{
				const PMP_DATA_FRAME_INFOS *frame_infos = &data.data_frame_infos;
				const char *str;

				if (Gui_state != STATE_PLAYING) {
					get_position_parameters();
				}

				switch (frame_infos->mode) {
				case MODE_INTRA:
					str = "I";
					fl_set_object_label(Frame_infos->frame_forward_f_code, "n/a");
					fl_set_object_label(Frame_infos->frame_forward_full_pel, "n/a");
					fl_set_object_label(Frame_infos->frame_backward_f_code, "n/a");
					fl_set_object_label(Frame_infos->frame_backward_full_pel, "n/a");
					break;
				case MODE_PREDICTIVE:
					str = "P";
					fl_set_object_label(Frame_infos->frame_forward_f_code, u2str(frame_infos->f_code[F][HT]));
					fl_set_object_label(Frame_infos->frame_forward_full_pel, frame_infos->full_pel[F] ? "full-pel" : "half-pel");
					fl_set_object_label(Frame_infos->frame_backward_f_code, "n/a");
					fl_set_object_label(Frame_infos->frame_backward_full_pel, "n/a");
					break;
				case MODE_INTERPOLATIVE:
					str = "B";
					fl_set_object_label(Frame_infos->frame_forward_f_code, u2str(frame_infos->f_code[F][HT]));
					fl_set_object_label(Frame_infos->frame_forward_full_pel, frame_infos->full_pel[F] ? "full-pel" : "half-pel");
					fl_set_object_label(Frame_infos->frame_backward_f_code, u2str(frame_infos->f_code[B][HT]));
					fl_set_object_label(Frame_infos->frame_backward_full_pel, frame_infos->full_pel[B] ? "full-pel" : "half-pel");
					break;
				default:					str = "?"; break;
				}
				fl_set_object_label(Frame_infos->frame_type, str);
				fl_set_object_label(Frame_infos->frame_temporal_reference, u2str(frame_infos->temporal_reference));
				fl_set_object_label(Frame_infos->frame_vbv_delay, frame_infos->vbv_delay < 0 ? "vbr" : d2str(frame_infos->vbv_delay, 2));
#if 0 /* Tristan Savatier - Wed Oct 29 00:07:47 PST 1997 */
//				fl_set_object_label(Controls->number_of_frames_played, u2str(frame_infos->display_index));
#endif /* 0 - Tristan Savatier - Wed Oct 29 00:07:47 PST 1997 */
			}
			break;
		case A2G_COM_GOP_INFOS:
			/* GOP info not yet implemented in mtv! */
#if 0 /* Tristan Savatier - Tue Jun 17 22:49:12 PDT 1997 */
//			{
//				const PMP_DATA_GOP_INFOS *gop_infos = &data.data_gop_infos;
//
//				printf("time_code_drop = %d\n", gop_infos->time_code_drop);
//				printf("time_code_hours = %d\n", gop_infos->time_code_hours);
//				printf("time_code_minutes = %d\n", gop_infos->time_code_minutes);
//				printf("time_code_seconds = %d\n", gop_infos->time_code_seconds);
//				printf("time_code_pictures = %d\n", gop_infos->time_code_pictures);
//				printf("closed_gop = %d\n", gop_infos->closed_gop);
//				printf("broken_link = %d\n", gop_infos->broken_link);
//				printf("\n");
//			}
#endif /* 0 - Tristan Savatier - Tue Jun 17 22:49:12 PDT 1997 */
			break;
		case A2G_COM_ERROR_IN_STREAM:
			/* we ignore those messages */
			break;
		case A2G_COM_AUDIO_PORTS_AVAILABLE:
			Audio_ports = data.data_generic.data_int;
			{
				if (data.data_generic.data_int & AUDIO_PORT_LINE_OUT) {
					fl_show_object(Audio->audio_port_line_out);
				}
				if (data.data_generic.data_int & AUDIO_PORT_HEADPHONE) {
					fl_show_object(Audio->audio_port_headphone);
				}
				if (data.data_generic.data_int & AUDIO_PORT_SPEAKER) {
					fl_show_object(Audio->audio_port_speaker);
				}
			}
			break;
		case A2G_COM_AUDIO_PORTS_SELECTED:
			/*
			 * Ignore value -1:
			 * -1 can be received only at startup (i.e. before
			 * receiving A2G_COM_READY) and it means that mtvp
			 * has not changed the current port selection. The actual ports
			 * that are selected are received later.
			 */
			if (data.data_generic.data_int != -1) {
				if (data.data_generic.data_int & AUDIO_PORT_LINE_OUT) {
					fl_set_button(Audio->audio_port_line_out, 1);
				} else {
					fl_set_button(Audio->audio_port_line_out, 0);
				}
				if (data.data_generic.data_int & AUDIO_PORT_HEADPHONE) {
					fl_set_button(Audio->audio_port_headphone, 1);
				} else {
					fl_set_button(Audio->audio_port_headphone, 0);
				}
				if (data.data_generic.data_int & AUDIO_PORT_SPEAKER) {
					fl_set_button(Audio->audio_port_speaker, 1);
				} else {
					fl_set_button(Audio->audio_port_speaker, 0);
				}
			}
			break;
		case A2G_COM_SYSTEMS_END_CODE:
			break;
		case A2G_COM_VIDEO_END_CODE:
#if 0 /* Tristan Savatier - Wed Aug 19 15:57:29 PDT 1998 */
//#ifdef VCD
//			if (Vcd_tracks != 0 && Vcd_track != Vcd_tracks) {
//				Vcd_track++;
//				fl_set_counter_value(Controls->vcd_track, (float)Vcd_track);
//				G2A_SEND_INT(COM_VCD_OPEN_TRACK, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, Vcd_track);
//			}
//#endif
#endif /* 0 - Tristan Savatier - Wed Aug 19 15:57:29 PDT 1998 */
			break;
		case A2G_COM_AUDIO_PID:
			/* printf("audio pid = %d\n", data.data_generic.data_int); */
			break;
		case A2G_COM_AUDIO_NO_CONTROLS_SELECT:
			fl_hide_object(Audio->audio_controls_master);
			fl_set_button(Audio->audio_controls_pcm, 1);
			break;
		case A2G_COM_CHALLENGE:
			/*
			 * This is a reserved message of the API.
			 * This private authentification request is sent only if the
			 * player (mtvp) cannot find a valid license key.
			 * The player does not need a valid license key when it
			 * recognizes the bundled mtv front-end binary (precompiled).
			 *
			 * If the player (mtvp) finds a valid license key, it does not
			 * send this message.
			 *
			 * Front-ends (other than the bundled mtv) who receive this message
			 * should advise the user that the player has no license and will
			 * not be functionnal.
			 */
			if (Use_mtvp_key) {
				FATAL(("Could not find a valid key. Remove the file ~/.mtvrc and start mtv again. For licensing information contact MpegTV at info@mpegtv.com or visit us at www.mpegtv.com"));
			}
#ifdef HANDLE_CHALLENGE
			if (handle_challenge(&Player, &msg, &data) != 0) {
				show_alert("This is a limited evaluation version.", "Please contact MpegTV at info@mpegtv.com for licensing information.", "Or visit us at www.mpegtv.com", 1);
			}
#else
			show_alert("This front-end is not authentified by mtvp.", "You need a development license (in file mtvp.key).", "See www.mpegtv.com/sdk.html for details.", 1);
#endif
			break;
#ifndef NO_X
		case A2G_COM_X_EVENT:				
			switch (data.data_x_event.type) {
				case ButtonRelease:
					if (Controls->controls->visible) {
						fl_hide_form(Controls->controls);
					} else {
						show_and_raise_form(Controls->controls, FL_PLACE_POSITION, FL_FULLBORDER, "MpegTV Controls");
						update_control_title();
					}
					break;
				default:
					/* ignore other X events */
					break;
			}
			break;
#endif
		case A2G_COM_SIH_FEEDBACK:
#ifdef VCD
			if (data.data_sih_handler_id == SIH_HANDLER_ID_VCD) {
				const SIH_VCD_FEEDBACK *p = (const SIH_VCD_FEEDBACK *)((const char *)&data + sizeof(SIH_HANDLER_ID));
				
				if (p->type == VCD_FEEDBACK_TYPE_VIDEO_TRACK_COUNT) {

					Vcd_tracks = p->data.video_track_count.count;
					if (Vcd_tracks == 0) {
						show_alert("Cannot play VCD:", "There is no Video track on CD:", Vcd_device, 1);
						close_vcd();
					} else {
#if 0 /* Tristan Savatier - Fri Sep 18 01:51:39 PDT 1998 */
//						PMP_DATA_GENERIC data_generic;
//						char *url = data_generic.data_string;
#endif /* 0 - Tristan Savatier - Fri Sep 18 01:51:39 PDT 1998 */
						
						Vcd_track = 1;
						fl_set_counter_bounds(Controls->vcd_track, 1.0, (double)Vcd_tracks);
						fl_set_counter_value(Controls->vcd_track, (float)Vcd_track);
						fl_show_object(Controls->vcd_track);

#if 0 /* Tristan Savatier - Fri Sep 18 01:51:32 PDT 1998 */
//						url[sizeof data_generic.data_string - 1] = 0;
//						sprintf(url, "vcd:%u", Vcd_track);
//						if (url[sizeof data_generic.data_string - 1] != 0) abort_now();
//	
//						G2A_SEND_STRING(COM_OPEN_STREAM_URL, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, url);
#endif /* 0 - Tristan Savatier - Fri Sep 18 01:51:32 PDT 1998 */
					}
				}
			}
#endif
			break;			
		case A2G_COM_RETURN_SUCCESS:
			/*
			 * This message is received each time a command that was sent with
			 * the PMP_FLAG_RETURN_STATUS flag succeeded.
			 * This is convenient to test the return status of commands like
			 * COM_VCI_LINK or COM_OPEN_FILE.  Note that if PMP_FLAG_RETURN_ORIGINAL
			 * was set too, a copy of the original message is sent back right
			 * after the A2G_COM_RETURN_FAILURE message.
			 */
			{
				PMP_HEADER	original_msg;
				PMP_DATA_G2A	original_data;
				if (a2g_read_msg(&Player, &original_msg, (PMP_DATA_A2G *)&original_data) == -1) {
					FATAL_ERROR(("a2g_read_msg: error reading msg"));
				}
				{
					char buf[PMP_PRINT_MSG_BUF_SIZE];
					MESSAGE(("received msg: %s", pmp_print_msg(&original_msg, (const char *)&original_data, buf)));
				}
#ifdef VCD
				if (Vcd_tracks != 0) {
					if (original_msg.type == COM_DEMO || original_msg.type == COM_CLOSE_STREAM) {
						close_vcd();
					} else if (original_msg.type == COM_OPEN_STREAM_URL
						&& data.data_return_success.value != SIH_HANDLER_ID_VCD) {
					
						close_vcd();
					}
				}

#endif
				switch (original_msg.type) {
#if 0 /* Tristan Savatier - Mon Aug 31 15:20:40 PDT 1998 */
//					case COM_OPEN_FILE:
//						setup_after_file_open(original_data.data_generic.data_string);
//						break;
#endif /* 0 - Tristan Savatier - Mon Aug 31 15:20:40 PDT 1998 */
					case COM_SIH_INIT:
#ifdef VCD
						if (original_data.data_sih_handler_id == SIH_HANDLER_ID_VCD) {
							G2A_SEND_STRING(COM_OPEN_STREAM_URL, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, "vcd:1");
						}
#endif
						break;
					case COM_CLOSE_STREAM:
						break;
#ifdef MTV_EASTER_EGG
					case COM_DEMO:		/*FALLTHROUGH*/
#endif
					case COM_OPEN_STREAM_URL:	/*FALLTHROUGH*/
						/*
						 * Reset bar right now makes the player feel more
						 * reactive, because seek can actually take some time...
						 */
						fl_set_slider_value(Controls->frame_slider, 0.0);

						{
							char type[512];
							char buf[512];
							
							type[sizeof type - 1] = 0;
							
							if (original_msg.type == COM_OPEN_STREAM_URL) {
								strcpy(type, original_data.data_generic.data_string);
								
								switch (data.data_return_success.value) {
									case SIH_HANDLER_ID_FILE:
										setup_after_file_open(original_data.data_generic.data_string);
										goto done;
										break;
									case SIH_HANDLER_ID_FD:
										if (strcasecmp(original_data.data_generic.data_string, "fd:0") == 0 || strcmp(original_data.data_generic.data_string, "-") == 0) {
											sprintf(type, "<stdin>");
										}
										break;
#ifdef VCD
									case SIH_HANDLER_ID_VCD:
										sprintf(type, "%s of %u", original_data.data_generic.data_string, Vcd_tracks);
										break;
#endif
									default:
										break;
								} 
#ifdef MTV_EASTER_EGG
							} else if (original_msg.type == COM_DEMO) {
								sprintf(type, "c\'est magnifique!");
#endif
							} else {
								abort_now();
							}
							if (type[sizeof type - 1] != 0) abort_now();
							
							buf[sizeof buf - 1] = 0;
							sprintf(buf, "MpegTV - %s", type);
							if (buf[sizeof buf - 1] != 0) abort_now();
							
							G2A_SEND_STRING(COM_WINDOW_NAME, 0, buf);
							G2A_SEND_STRING(COM_ICON_NAME, 0, type);

							disable_audio_infos();
							disable_stream_infos();
							fl_set_object_label(Infos->file_name, type);
							show_stream_controls();
						}
done:
						break;
					case COM_VIDEO_GAMMA:
						if (Gamma.is_pending) {
							G2A_SEND_DOUBLE(COM_VIDEO_GAMMA, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, Gamma.pending);
							Gamma.is_pending = False;
						} else {							
							Gamma.not_ready = False;
						}
						break;
					case COM_VIDEO_LUMA_FLOOR:
						if (Contrast.is_pending) {
							G2A_SEND_DOUBLE(COM_VIDEO_LUMA_FLOOR, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, - Contrast.pending);
							Contrast.is_pending = False;
						} else {
							Contrast.not_ready = False;
						}
						break;
					case COM_VIDEO_LUMA_SHIFT:
						if (Brightness.is_pending) {
							G2A_SEND_DOUBLE(COM_VIDEO_LUMA_SHIFT, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, Brightness.pending);
							Brightness.is_pending = False;
						} else {
							Brightness.not_ready = False;
						}
						break;
					case COM_VIDEO_CHROMA_SATURATION:
						if (Chroma_saturation.is_pending) {
							G2A_SEND_DOUBLE(COM_VIDEO_CHROMA_SATURATION, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, Chroma_saturation.pending);
							Chroma_saturation.is_pending = False;
						} else {
							Chroma_saturation.not_ready = False;
						}
						break;
					default:
						break;
				}
			}
			break;
		case A2G_COM_RETURN_FAILURE:
			/*
			 * This message is received each time a command that was sent with
			 * the PMP_FLAG_RETURN_STATUS flag failed.
			 * This is convenient to test the return status of commands like
			 * COM_VCI_LINK or COM_OPEN_FILE.  Note that if PMP_FLAG_RETURN_ORIGINAL
			 * was set too, a copy of the original message is sent back right
			 * after the A2G_COM_RETURN_FAILURE message.
			 */
			{
				PMP_HEADER	original_msg;
				PMP_DATA_G2A	original_data;
				if (a2g_read_msg(&Player, &original_msg, (PMP_DATA_A2G *)&original_data) == -1) {
					FATAL_ERROR(("a2g_read_msg: error reading msg"));
				}
				{
					char buf[PMP_PRINT_MSG_BUF_SIZE];
					MESSAGE(("received msg: %s", pmp_print_msg(&original_msg, (const char *)&original_data, buf)));
				}

				switch (original_msg.type) {
				case COM_SAVE_STREAM:
				case COM_SAVE_IMAGE_PPM:
				case COM_SAVE_IMAGE_TGA:
					if (data.data_return_failure.error == EEXIST) {
#if FL_REVISION == 81
						if (fl_show_question("The file that you have specified:", original_data.data_generic.data_string, "already exists. Do you want to replace it ?")) {
#else /* >= 86 */
						char buf[1024];
						
						buf[sizeof buf - 1] = 0;
						sprintf(buf, "The file that you have specified:\n%s\nalready exists. Do you want to replace it ?", original_data.data_generic.data_string);
						if (buf[sizeof buf - 1] != 0) {
							abort_now();
						}
						
						if (fl_show_question(buf, -1)) {
#endif
							if (unlink(original_data.data_generic.data_string) == 0) {
								g2a_send(&Player, &original_msg, (const PMP_DATA_G2A *)original_data.data_generic.data_string);
							} else {
								show_alert("Cannot remove file:", original_data.data_generic.data_string, strerror(errno), 1);
							}
						}
					} else {
						show_alert("Cannot save in file:", original_data.data_generic.data_string, strerror(data.data_return_failure.error), 1);
					}
					break;
#if 0 /* Tristan Savatier - Mon Aug 31 15:15:34 PDT 1998 */
//				case COM_OPEN_FILE:
//					show_alert("Cannot open file:", original_data.data_generic.data_string, strerror(data.data_return_failure.error), 1);
//					break;
#endif /* 0 - Tristan Savatier - Mon Aug 31 15:15:35 PDT 1998 */

				case COM_SIH_INIT:
					{
						const char *param = (const char *)&original_data + sizeof(SIH_HANDLER_ID);
#ifdef VCD
						if (original_data.data_sih_handler_id == SIH_HANDLER_ID_VCD) {
							char buf[256];
							
							buf[sizeof buf - 1] = 0;
							sprintf(buf, "Cannot open CD-ROM device: %s", param);
							if (buf[sizeof buf - 1] != 0) abort_now();
					
							show_alert(buf, strerror(data.data_return_failure.error), "Check env variable MTV_VCD_DEVICE (see \"mtv -h\" for help)", 1);
							break;
						}
#endif
						show_alert("COM_SIH_INIT error:", param, strerror(data.data_return_failure.error), 1);
					}
					break;
				case COM_OPEN_STREAM_URL:
					if (strcasecmp(original_data.data_generic.data_string, "fd:0") == 0 || strcmp(original_data.data_generic.data_string, "-") == 0) {
						show_alert("Cannot play from Stdin:", "Stdin is not a pipe, file or socket", strerror(data.data_return_failure.error), 1);
#ifdef VCD
					} else if (strncasecmp(original_data.data_generic.data_string, "vcd:", 4) == 0) {
						show_alert("Cannot play this VCD track:", original_data.data_generic.data_string, strerror(data.data_return_failure.error), 1);
						close_vcd();
#endif
					} else {
						show_alert("Cannot open this file or URL:", original_data.data_generic.data_string, strerror(data.data_return_failure.error), 1);
					}
					break;
#ifdef MTV_EASTER_EGG
				case COM_DEMO:
					break;
#endif
				default:
					{
						char buf[256];
						
						buf[sizeof buf - 1] = 0;
						sprintf(buf, "Error: command %d failed", original_msg.type);
						if (buf[sizeof buf - 1] != 0) abort_now();
						show_alert(buf, strerror(data.data_return_failure.error), "", 1);
					}
				}
			}
			break;
		default:
			/* msgs not understood should be ignored */
			{
				char buf[PMP_PRINT_MSG_BUF_SIZE];
				printf("gui: received unknown msg (ignored): %s\n", pmp_print_msg(&msg, (const char *)&data, buf));
			}
			break;
	}

	return 0;
}


/*--------------------------------------------------------------------------+
| seek																		|
+--------------------------------------------------------------------------*/
static void
seek(
	int		n
)
{
	if (Not_ready) {
		Pending_seek = n;
	} else {
		Pending_seek = -1;
		send_seek(n);
	}
}

/*--------------------------------------------------------------------------+
| warn_if_not_sync															|
+--------------------------------------------------------------------------*/
static void
warn_if_not_sync()
{
	if (!Framerate_is_nominal) {
	
		show_alert("Warning: audio is currently disabled.",
				"Audio can be played only when video plays at nominal framerate.",
				"Check Sync Options (Options->Sync...).",
				1);
	}
}

/*--------------------------------------------------------------------------+
| enter_fast_forward_mode													|
+--------------------------------------------------------------------------*/
static void
enter_fast_forward_mode()
{
	double frame_rate;
		
	if (Framerate_is_nominal) {
		frame_rate = Nominal_framerate;
	} else {
		frame_rate = fl_get_slider_value(Sync->sync_framerate_slider);
	}
	
	Fast_forward = True;
	State_before_fast_forward = Gui_state;
	G2A_SEND_DOUBLE(COM_SYNC_FRAMERATE, 0, FAST_FORWARD_COEF * frame_rate);
	
	if (State_before_fast_forward != STATE_PLAYING) {
		/*
		 * STATE_FLAG_NO_ECHO is set because we do not want to
		 * receive the A2G_COM_STATE message caused by this
		 * COM_STATE command.
		 * Without STATE_FLAG_NO_ECHO, the A2G_COM_STATE would cause
		 * the play button icon to change if fast-forward is used on
		 * a stopped or paused player. We don't want that.
		 */
		send_state(STATE_PLAYING|STATE_FLAG_NO_ECHO);
	}
}

/*--------------------------------------------------------------------------+
| quit_fast_forward_mode													|
+--------------------------------------------------------------------------*/
static void
quit_fast_forward_mode()
{

	if (!Fast_forward) {
		return;
	}

	if (Framerate_is_nominal) {
		G2A_SEND_DOUBLE(COM_SYNC_FRAMERATE, 0, 0.0);
	} else {
		double frame_rate;
		
		frame_rate = fl_get_slider_value(Sync->sync_framerate_slider);
		G2A_SEND_DOUBLE(COM_SYNC_FRAMERATE, 0, frame_rate);
	}
	
	Fast_forward = False;
	
	if (State_before_fast_forward != STATE_PLAYING) {
		/*
		 * STATE_FLAG_NO_ECHO is set because we do not want to
		 * receive the A2G_COM_STATE message caused by this
		 * COM_STATE command.
		 * Without STATE_FLAG_NO_ECHO, the A2G_COM_STATE would cause
		 * the play button icon to change if fast-forward is used on
		 * a stopped or paused player. We don't want that.
		 */
		send_state(STATE_PAUSED|STATE_FLAG_NO_ECHO);
	}
}				

/*--------------------------------------------------------------------------+
| show_form_license															|
+--------------------------------------------------------------------------*/
static void
show_form_license()
{
	show_and_raise_form(License->license, FL_PLACE_CENTER | FL_FREE_SIZE, FL_FULLBORDER, "MpegTV - License");
}

/*--------------------------------------------------------------------------+
| show_form_registration													|
+--------------------------------------------------------------------------*/
static void
show_form_registration()
{
	show_and_raise_form(Registration->registration, FL_PLACE_CENTER, FL_FULLBORDER, "MpegTV - Registration");
}

/*--------------------------------------------------------------------------+
| show_splash_if_not_registered												|
+--------------------------------------------------------------------------*/
static void
show_splash_if_not_registered()
{
	if (!Registered) {
		show_and_raise_form(Splash->splash, FL_PLACE_CENTER, FL_FULLBORDER, "MpegTV - UNREGISTERED");
	}
}

#ifdef _REENTRANT
/*--------------------------------------------------------------------------+
| show_form_udp																|
+--------------------------------------------------------------------------*/
static void
show_form_udp()
{
	show_and_raise_form(Udp->udp, FL_PLACE_CENTER, FL_FULLBORDER, "MpegTV - Play from UDP Port");
}
#endif

#if 0 /* Tristan Savatier - Tue Sep 1 19:55:31 PDT 1998 */
///*--------------------------------------------------------------------------+
//| update_slider_timer_CB													|
//+--------------------------------------------------------------------------*/
//void
//update_slider_timer_CB(
//	FL_OBJECT *fl_object,
//	long param
//)
//{
//	update_slider();
//	
//	fl_set_timer(Controls->update_slider_timer, 1.0);
//}
#endif /* 0 - Tristan Savatier - Tue Sep 1 19:55:32 PDT 1998 */

/*--------------------------------------------------------------------------+
| update_slider_timeout_CB													|
+--------------------------------------------------------------------------*/
#if 0 /* Tristan Savatier - Thu Sep 17 18:17:02 PDT 1998 */
//typedef void (*FL_TIMEOUT_CALLBACK)(int, void *);
#endif /* 0 - Tristan Savatier - Thu Sep 17 18:17:02 PDT 1998 */

void
update_slider_timeout_CB(
	int		id,
	void *data
)
{
	update_slider();
	
	Timeout_id = fl_add_timeout(1000L, update_slider_timeout_CB, NULL);
}

#ifdef USE_SMC

static int
getTitleBitRate(char *title)
{
    int maxBitRate;
    MsmTitle *mtp;
    MsmReturn retCode;

    retCode = msmTitleGetStatus(Smc_ms, title, &mtp);
    if (retCode != MSMRETURN_SUCCESS) {
		FATAL(("msmTitleGetStatus returns %s", msmErrorMsg(Smc_ms, retCode)));
    }
    maxBitRate = mtp->maxBitRate;
    /*
     * msmTitleGetStatus() allocates memory for its return value --
     * don't forget to free it.  (And don't forget to copy any values
     * you need before you free it!)
     */
    (void) msmTitleStatusFree(mtp);
    return maxBitRate;
}

static void
setPlaylist(char *title)
{
    MsmPlaylist playlist;
    MsmItem anItem;
    MsmReturn retCode;

    /*
     * Build the playlist
     */
    playlist.modstamp = MSMTIME_CURRENT;
    playlist.isLoop = FALSE;
    playlist.editStartPosition = MSMTIME_ZERO;
    playlist.editDuration = MSMTIME_MAXTIME;
    playlist.listStartPosition = MSMTIME_ZERO;

    playlist.items.len = 1;
    playlist.items.val = &anItem;

    playlist.items.val[0].type = MSMITEM_TITLE;
    playlist.items.val[0].un.titleItem.titleName = title;
    playlist.items.val[0].un.titleItem.itemDuration = MSMTIME_CURRENT;
    playlist.items.val[0].un.titleItem.startOffset = MSMTIME_ZERO;
    playlist.items.val[0].un.titleItem.playDuration = MSMTIME_CURRENT;
    playlist.items.val[0].un.titleItem.joinInDuration = MSMTIME_ZERO;
    playlist.items.val[0].un.titleItem.isTimeLocked = FALSE;
    playlist.items.val[0].un.titleItem.playClosestSpeed = TRUE;

    retCode = msmPlayerSetPlaylist(Smc_ms, Smc_mp, &playlist);
    if (retCode != MSMRETURN_SUCCESS) {
		deletePlayer();
		FATAL(("msmPlayerSetPlaylist returns %s", msmErrorMsg(Smc_ms, retCode)));
	}
}

static void
playTilDone(void)
{
    MsmReturn retCode;
    int stopped = 0;

    retCode = msmPlayerPlay(Smc_ms, Smc_mp, MSMSPEED_FORWARD, MSMTIME_ZERO,
			    MSMTIME_MAXTIME, MSMTIME_CURRENT);
    if (retCode != MSMRETURN_SUCCESS) {
		deletePlayer();
		FATAL(("msmPlayerPlay returns %s", msmErrorMsg(Smc_ms, retCode)));
    }

#if 0 /* Tristan Savatier - Mon Mar 16 01:45:37 PST 1998 */
//    do {
//	MsmPlayStatus *mpsp;
//
//	retCode = msmPlayerGetPlayStatus(Smc_ms, Smc_mp, &mpsp);
//	if (retCode != MSMRETURN_SUCCESS) {
//	    deletePlayer();
//		FATAL(("msmPlayerGetPlayStatus returns %s", msmErrorMsg(Smc_ms, retCode)));
//	}
//
//	stopped = mpsp->currentState == MSMPLAYSTATE_STOP;
//	(void) msmPlayerPlayStatusFree(mpsp);
//
//	(void) sleep(statusCheckSecs);
//    } while (! stopped);
#endif /* 0 - Tristan Savatier - Mon Mar 16 01:45:37 PST 1998 */
}

static void
createPlayer(char *dest, int maxBitRate)
{
    MsmConnect mc;
    MsmTime terminateDate;
    MsmReturn retCode;

    /*
     * If not multithread env, just hope we can clean-up after ourselves...
     */
    terminateDate = MSMTIME_MAXTIME;

    /*
     * Use the destination as the player "tag".  Use MSMLOOKUP_EXCL,
     * so we don't create another player if we already have one pointed at
     * this destination (possibly because of an untimely death of
     * an earlier instance of this program).  If this is a multithreaded
     * environment, that player should soon self-destruct.
     */
    retCode = msmPlayerLookup(Smc_ms, dest, terminateDate, MSMLOOKUP_EXCL, &Smc_mp);
    if (retCode != MSMRETURN_SUCCESS) {
		MESSAGE(("msmPlayerLookup returns %s", msmErrorMsg(Smc_ms, retCode)));
		if (retCode == MSMRETURN_PLAYERACTIVE) {
		    retCode = msmPlayerLookup(Smc_ms, dest, terminateDate, MSMLOOKUP_FIND, &Smc_mp);
		    if (retCode != MSMRETURN_SUCCESS) {
				FATAL(("msmPlayerLookup returns %s", msmErrorMsg(Smc_ms, retCode)));
		    }
		}
    }


    mc.destTiAddr = dest;
    mc.encap = "MPEGTS";	/* just assume MPEG2 Transport Stream */
    mc.rate = maxBitRate;

    retCode = msmPlayerSetConnect(Smc_ms, Smc_mp, &mc);
    if (retCode != MSMRETURN_SUCCESS) {
		deletePlayer();
		FATAL(("msmPlayerSetConnect returns %s", msmErrorMsg(Smc_ms, retCode)));
    }
}

static void
deletePlayer(void)
{
    MsmReturn retCode;

    retCode = msmPlayerDelete(Smc_ms, Smc_mp);
    if (retCode != MSMRETURN_SUCCESS) {
		MESSAGE(("msmPlayerDelete returns %s", msmErrorMsg(Smc_ms, retCode)));
	}
}

/*--------------------------------------------------------------------------+
| smc_play																	|
+--------------------------------------------------------------------------*/
static int
smc_play()
{
    char *server = "milk";
    char *title = "SMCnews_1.24Mbs";
    char *dest = "host=milk,udpport=8080";
    MsmReturn retCode;
    int maxBitRate;

    /*
     * Open rpc connection to Sun MediaCenter's Media Stream Manager.
     */
    retCode = msmServerOpen(server, &Smc_ms);
    if (retCode != MSMRETURN_SUCCESS) {
		FATAL(("msmServerOpen returns %s", msmErrorMsg(Smc_ms, retCode)));
		return -1;
    }

    /*
     * Get bit rate of title.
     */
    maxBitRate = getTitleBitRate(title);

    /*
     * Create a new player that sends its stream to the given destination.
     * Set the maximum bit rate of the stream to the bit rate of the title.
     */
    createPlayer(dest, maxBitRate);

    /*
     * Make a playlist consisting of this single title.
     */
    setPlaylist(title);

    /*
     * Play title until completion.
     */
    playTilDone();

#if 0 /* Tristan Savatier - Mon Mar 16 02:16:09 PST 1998 */
//    /*
//     * Delete the player.
//     */
//    deletePlayer();
#endif /* 0 - Tristan Savatier - Mon Mar 16 02:16:09 PST 1998 */

    /*
     * Close rpc connection to SMC.
     */
    retCode = msmServerClose(Smc_ms);
    if (retCode != MSMRETURN_SUCCESS) {
		FATAL(("msmServerClose returns %s", msmErrorMsg(Smc_ms, retCode)));
	}
	return 0;
}

#endif /* USE_SMC */

/*--------------------------------------------------------------------------+
| open_url																	|
+--------------------------------------------------------------------------*/
int
open_url (const char *url)
{
#if 0 /* Tristan Savatier - Wed Feb 17 05:03:57 PST 1999 */
//    while (isspace (*url)) {
//		++url;
//	}
#endif /* 0 - Tristan Savatier - Wed Feb 17 05:03:57 PST 1999 */

    if (mozilla_open_url (url) == 0) {
		return 0;
	}

	if (!mozilla_can_start_browser()) {
#if 0 /* Tristan Savatier - Sun Feb 21 16:21:55 PST 1999 */
//		show_alert("This resource is located at the following URL:", url, "You need a Web Browser (e.g. Netscape) to access it.", 1);
#endif /* 0 - Tristan Savatier - Sun Feb 21 16:21:55 PST 1999 */
		fl_show_message("This resource is located at the following URL:\n", url, "\nYou need a Web Browser (e.g. Netscape) to access it.");
		return -1;
	}
	
	{
#if FL_REVISION == 81
		if (fl_show_question("This resource is located at the following URL:", url, "Do you want mtv to start Netscape to see it ?")) {
#else /* >= 86 */
			char buf[1024];
			
			buf[sizeof buf - 1] = 0;
			sprintf(buf, "This resource is located at the following URL:\n\n%s\n\nDo you want mtv to start Netscape to see it ?", url);
			if (buf[sizeof buf - 1] != 0) {
				abort_now();
			}
			
			if (fl_show_question(buf, -1)) {
#endif
	
#if 0 /* Tristan Savatier - Thu Feb 18 02:06:19 PST 1999 */
//			fl_show_message("Please wait...", "", "mtv is starting your Netscape Browser");
#endif /* 0 - Tristan Savatier - Thu Feb 18 02:06:19 PST 1999 */
			show_alert("Please wait...", "mtv is starting your Netscape Browser", "This may take a few seconds", 1);
	
	    	return mozilla_start_browser (url);
	    } else {
	    	return 0;
	    }
	}
}

/*--------------------------------------------------------------------------+
| register_now																|
+--------------------------------------------------------------------------*/
void
register_now ()
{
	open_url(URL_REGISTER);
}

/*--------------------------------------------------------------------------+
| command																	|
+--------------------------------------------------------------------------*/
void
command(
	FL_OBJECT *fl_object,
	long param
)
{
	COMMAND	com = (COMMAND)param;
	
	/* this is our Xforms general callback */
	
	switch (com) {
		case COM_QUIT:
			G2A_SEND_VOID(COM_QUIT, 0);
			break;
		case COM_FILE_MENU:
			
			switch (fl_get_menu(Controls->file_menu)) {
				case FILE_MENU_OPEN_FILE:
					file_open();
					break;
				case FILE_MENU_OPEN_URL:
					url_open();
					break;
				case FILE_MENU_STDIN:
					reset_auto_seek();
					reset_loop();
					G2A_SEND_STRING(COM_OPEN_STREAM_URL, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, "fd:0");
					break;
#ifdef _REENTRANT
				case FILE_MENU_UDP:
					show_form_udp();
					break;
#endif
#ifdef VCD
				case FILE_MENU_VCD:
					if (Vcd_tracks != 0) {
						break;
					}
					reset_auto_seek();
					reset_loop();

					{
						struct {
							SIH_HANDLER_ID	sih_handler_id;
							char device_name[256];
						} sih_cntl_data;
						
						sih_cntl_data.sih_handler_id = SIH_HANDLER_ID_VCD;
						sih_cntl_data.device_name[sizeof sih_cntl_data.device_name - 1] = 0;
						strcpy(sih_cntl_data.device_name, Vcd_device);
						if (sih_cntl_data.device_name[sizeof sih_cntl_data.device_name - 1] != 0) abort_now();
		
						G2A_SEND_CUSTOM(COM_SIH_INIT, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, (const void *)&sih_cntl_data, sizeof(SIH_HANDLER_ID) + strlen(sih_cntl_data.device_name) + 1);
					}

					break;
#endif
#ifdef USE_SMC
				case FILE_MENU_SMC:
					smc_play();
					break;
#endif
				case FILE_MENU_CLOSE:
					hide_stream_controls();
					G2A_SEND_VOID(COM_CLOSE_STREAM, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL);
					break;
				case FILE_MENU_SAVE_AS:
					file_save();
					break;
				case FILE_MENU_INFOS:
					show_and_raise_form(Infos->infos, FL_PLACE_SIZE, FL_FULLBORDER, "MpegTV Sequence Infos");
					break;
				case FILE_MENU_QUIT:
					quit(EXIT_SUCCESS);
					break;
				case FILE_MENU_NONE:
				case FILE_MENU_HANGING:
					break;
				default:
					abort_now();
			}
			break;
		case COM_OPTIONS_MENU:
			switch (fl_get_menu(Controls->options_menu)) {
				case OPTIONS_MENU_AUDIO:
					show_and_raise_form(Audio->audio, FL_PLACE_SIZE, FL_FULLBORDER, "MpegTV Audio");
					break;
				case OPTIONS_MENU_VIDEO:
					show_and_raise_form(Video->video, FL_PLACE_SIZE, FL_FULLBORDER, "MpegTV Video");
					break;
				case OPTIONS_MENU_SYNC:
					show_and_raise_form(Sync->sync, FL_PLACE_SIZE, FL_FULLBORDER, "MpegTV Sync");
					break;
				case OPTIONS_MENU_AUTO_PLAY:
					if (fl_get_menu_item_mode(Controls->options_menu, OPTIONS_MENU_AUTO_PLAY) & FL_PUP_CHECK) {
						G2A_SEND_INT(COM_AUTO_PLAY, 0, 1);
					} else {
						G2A_SEND_INT(COM_AUTO_PLAY, 0, 0);							
					}
					break;
				case OPTIONS_MENU_SAVE:
					save_preferences();
					break;
				case OPTIONS_MENU_NONE:
				case OPTIONS_MENU_HANGING:
					break;
				default:
					abort_now();
			}
			break;
		case COM_WINDOW_MENU:
			switch (fl_get_menu(Controls->window_menu)) {
#ifdef USE_SDL
				case WINDOW_MENU_FULLSCREEN:
					G2A_SEND_INT(COM_VIDEO_MODE, 0, VIDEO_MODE_FULLSCREEN);
					break;
#endif
				case WINDOW_MENU_RAISE:
					G2A_SEND_VOID(COM_WINDOW_RAISE, 0);
					break;
				case WINDOW_MENU_ROOT:
					if (fl_get_menu_item_mode(Controls->window_menu, WINDOW_MENU_ROOT) & FL_PUP_CHECK) {
						G2A_SEND_INT(COM_WINDOW_ID, 0, -1);
						fl_set_menu_item_mode(Controls->window_menu, WINDOW_MENU_RAISE, FL_PUP_GREY);
					} else {
						G2A_SEND_INT(COM_WINDOW_ID, 0, 0);							
					}
					break;
				case WINDOW_MENU_FRAME_INFOS:
					show_and_raise_form(Frame_infos->frame_infos, FL_PLACE_SIZE, FL_FULLBORDER, "MpegTV Frame Infos");
					
					break;
				case WINDOW_MENU_SAVE_PPM:
					image_save_ppm();
					break;
				case WINDOW_MENU_SAVE_TGA:
					image_save_tga();
					break;
				case WINDOW_MENU_NONE:
				case WINDOW_MENU_HANGING:
					break;
				default:
					abort_now();
			}
			break;
		case COM_HELP_MENU:
			switch (fl_get_menu(Controls->help_menu)) {
				case HELP_MENU_ABOUT_MTV:
					show_and_raise_form(About->about, FL_PLACE_SIZE, FL_FULLBORDER, "About MpegTV");
					break;
				case HELP_MENU_ABOUT_MPEG:
					open_url(URL_ABOUT_MPEG);
					break;
				case HELP_MENU_BITSTREAMS:
					open_url(URL_BITSTREAMS);
					break;
				case HELP_MENU_RELEASE_NOTES:
					open_url(URL_RELEASE_NOTES);
					break;
				case HELP_MENU_MANUAL:
					open_url(URL_MANUAL);
					break;
				case HELP_MENU_FAQ:
					open_url(URL_FAQ);
					break;
				case HELP_MENU_SUPPORT:
					open_url(URL_SUPPORT);
					break;
				case HELP_MENU_LICENSE:
					show_form_license();
					break;
				case HELP_MENU_REGISTER:
					register_now();
					break;
				case HELP_MENU_KEY:
					show_form_registration();
					break;
				case HELP_MENU_NONE:
				case HELP_MENU_HANGING:
					break;
				default:
					abort_now();
			}
			break;
		case COM_PLAY_PAUSE:
			reset_loop_if_pos_out_of_loop(Stream_pos);
			set_state(Gui_play_pause_set_state);
			break;
		case COM_STOP:
			reset_loop();
			set_state(STATE_STOPPED);
			break;
		case COM_STEP:
			G2A_SEND_VOID(COM_STEP, 0);
			update_state(STATE_PAUSED);			
			break;
		case COM_AUDIO_MUTE:
			if (Audio_no_more) {
				disable_audio();
			} else {
				int 	muted = fl_get_button(fl_object);
				
				G2A_SEND_INT(com, 0, muted);
				if (!muted) {
					warn_if_not_sync();
				}
			}
			break;
		case COM_VIDEO_DITHER_COLORMAP_PRIVATE:
			G2A_SEND_INT(com, 0, fl_get_button(fl_object));
			break;
		case COM_LOOP:
			G2A_SEND_INT(com, 0, fl_get_button(fl_object));
			if (!fl_get_button(fl_object)) {
				reset_loop();
			}
			break;
		case COM_ZOOM:
			G2A_SEND_INT(com, 0, fl_get_button(fl_object) ? ZOOM_2X2 : ZOOM_1X1);
			break;
		case COM_AUDIO_CH_BOTH:
			Audio_channel = AUDIO_CHANNEL_BOTH;
			G2A_SEND_INT(COM_AUDIO_CHANNEL, 0, AUDIO_CHANNEL_BOTH);
			break;
		case COM_AUDIO_CH_LEFT:
			Audio_channel = AUDIO_CHANNEL_LEFT;
			G2A_SEND_INT(COM_AUDIO_CHANNEL, 0, AUDIO_CHANNEL_LEFT);
			break;
		case COM_AUDIO_CH_RIGHT:
			Audio_channel = AUDIO_CHANNEL_RIGHT;
			G2A_SEND_INT(COM_AUDIO_CHANNEL, 0, AUDIO_CHANNEL_RIGHT);
			break;
		case COM_AUDIO_CONTROLS_MASTER:
			G2A_SEND_INT(COM_AUDIO_CONTROLS, 0, AUDIO_CONTROLS_MASTER);
			break;
		case COM_AUDIO_CONTROLS_PCM:
			G2A_SEND_INT(COM_AUDIO_CONTROLS, 0, AUDIO_CONTROLS_PCM);
			break;
		case COM_AUDIO_VOLUME:
			G2A_SEND_DOUBLE(com, 0, (double)fl_get_slider_value(fl_object));
			break;
		case COM_AUDIO_BALANCE:
			G2A_SEND_DOUBLE(com, 0, (double)fl_get_slider_value(fl_object));
			break;
		case COM_SEEK:
			assert (Bits_filesize != -1);
			seek((int)fl_get_slider_value(fl_object));
			break;
		case COM_REWIND:
			assert (Bits_filesize != -1);
			/*
			 * Reset bar right now makes the player feel more
			 * reactive, because seek can actually take some time...
			 */
			fl_set_slider_value(Controls->frame_slider, 0.0);
			seek(0);
			get_position_parameters();
			break;
		case COM_END:
			assert (Bits_filesize != -1);
			/*
			 * Reset bar right now makes the player feel more
			 * reactive, because seek can actually take some time...
			 */
			fl_set_slider_value(Controls->frame_slider, (double)Bits_filesize);
			reset_loop();
			G2A_SEND_VOID(COM_END, 0);
			get_position_parameters();
			break;
		case COM_FAST_FORWARD:
			{

				if (fl_get_button(fl_object)) {
					enter_fast_forward_mode();
				} else {
					quit_fast_forward_mode();
				}
			}
			break;
		case COM_ABOUT_CLOSE:
			fl_hide_form(About->about);
			break;
		case COM_LICENSE_AGREE:
			fl_hide_form(License->license);
			if (!Controls->controls->visible) {
				remove_splash_and_start_player();
				update_control_title();
			}
			break;
		case COM_LICENSE_DISAGREE:
			switch (fl_show_choice(
					"You cannot continue using this software if you disagree",
					"with the terms and conditions set forth in the license.",
					"",
					2,
					"I still don't\nagree",
					"Read License\nagain",
					"",
					1
					)) {
				case 1: clean_exit(EXIT_FAILURE); break;
				case 2: break;
				default: abort_now();
			}
			break;
		case COM_REGISTRATION_CLOSE:
			fl_hide_form(Registration->registration);
			break;
		case COM_REGISTRATION_OK:
			handle_registration();
#if 0 /* Tristan Savatier - Wed Oct 28 17:06:21 PST 1998 */
//			fl_hide_form(Registration->registration);
#endif /* 0 - Tristan Savatier - Wed Oct 28 17:06:21 PST 1998 */
			break;
		case COM_SPLASH_REGISTER:
			register_now();
			break;
#if 0 /* Tristan Savatier - Wed Feb 17 04:15:14 PST 1999 */
//		case COM_SPLASH_LICENSE:
//			show_form_license();
//			break;
#endif /* 0 - Tristan Savatier - Wed Feb 17 04:15:14 PST 1999 */
		case COM_SPLASH_KEY:
			show_form_registration();
			break;
		case COM_SPLASH_EVALUATE:
#if 0 /* Tristan Savatier - Mon Jan 25 11:54:56 PST 1999 */
//			if (Expired) {
//				quit(EXIT_FAILURE);
//			}
#endif /* 0 - Tristan Savatier - Mon Jan 25 11:54:56 PST 1999 */
#if 0 /* Tristan Savatier - Wed Feb 17 04:05:51 PST 1999 */
//			remove_splash_and_start_player();
//			update_control_title();
#endif /* 0 - Tristan Savatier - Wed Feb 17 04:05:51 PST 1999 */
			show_form_license();
			break;
#ifdef _REENTRANT
		case COM_UDP_CLOSE:
			fl_hide_form(Udp->udp);
			break;
		case COM_UDP_OK:
			handle_udp();
			fl_hide_form(Udp->udp);
			break;
#endif
		case COM_AUDIO_CLOSE:
			fl_hide_form(Audio->audio);
			break;
		case COM_VIDEO_CLOSE:
			fl_hide_form(Video->video);
			break;
		case COM_SYNC_CLOSE:
			fl_hide_form(Sync->sync);
			break;
		case COM_INFOS_CLOSE:
			fl_hide_form(Infos->infos);
			break;
		case COM_FRAME_INFOS_CLOSE:
			fl_hide_form(Frame_infos->frame_infos);
			break;
		case COM_ALERT_CLOSE:
			fl_hide_form(fl_object->form);
			fl_free_form(fl_object->form);
			break;
		case COM_VIDEO_DITHER_NONE:
			G2A_SEND_INT(COM_VIDEO_DITHER_TYPE, 0, DITHER_NONE);
#ifdef USE_SDL
			fl_set_menu_item_mode(Controls->window_menu, WINDOW_MENU_FULLSCREEN, FL_PUP_GREY);
#endif
			break;
		case COM_VIDEO_DITHER_GRAY:
			G2A_SEND_INT(COM_VIDEO_DITHER_TYPE, 0, DITHER_GRAY);
#ifdef USE_SDL
			fl_set_menu_item_mode(Controls->window_menu, WINDOW_MENU_FULLSCREEN, FL_PUP_GREY);
#endif
			break;
		case COM_VIDEO_DITHER_ORDERED:
			G2A_SEND_INT(COM_VIDEO_DITHER_TYPE, 0, DITHER_ORDERED);
#ifdef USE_SDL
			fl_set_menu_item_mode(Controls->window_menu, WINDOW_MENU_FULLSCREEN, FL_PUP_GREY);
#endif
			break;
		case COM_VIDEO_DITHER_COLOR:
			G2A_SEND_INT(COM_VIDEO_DITHER_TYPE, 0, DITHER_COLOR);
#ifdef USE_SDL
			fl_set_menu_item_mode(Controls->window_menu, WINDOW_MENU_FULLSCREEN, FL_PUP_NONE);
#endif
			break;
		case COM_VIDEO_GAMMA_SLIDER:
			if (Gamma.not_ready) {
				Gamma.pending = fl_get_slider_value(fl_object);
				Gamma.is_pending = True;
			} else {
				Gamma.not_ready = True;
				Gamma.is_pending = False;
				G2A_SEND_DOUBLE(COM_VIDEO_GAMMA, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, (double)fl_get_slider_value(fl_object));
			}
			break;
		case COM_VIDEO_GAMMA_SLIDER_RESET:
			fl_set_slider_value(Video->video_gamma_slider, GAMMA_DEFAULT);
			command(Video->video_gamma_slider, COM_VIDEO_GAMMA_SLIDER);
			break;
		case COM_VIDEO_CONTRAST_SLIDER:
			if (Contrast.not_ready) {
				Contrast.pending = fl_get_slider_value(fl_object);
				Contrast.is_pending = True;
			} else {
				Contrast.not_ready = True;
				Contrast.is_pending = False;
				G2A_SEND_DOUBLE(COM_VIDEO_LUMA_FLOOR, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, - (double)fl_get_slider_value(fl_object));
			}
			break;
		case COM_VIDEO_CONTRAST_SLIDER_RESET:
			fl_set_slider_value(Video->video_contrast_slider, CONTRAST_DEFAULT);
			command(Video->video_contrast_slider, COM_VIDEO_CONTRAST_SLIDER);
			break;
		case COM_VIDEO_BRIGHTNESS_SLIDER:
			if (Brightness.not_ready) {
				Brightness.pending = fl_get_slider_value(fl_object);
				Brightness.is_pending = True;
			} else {
				Brightness.not_ready = True;
				Brightness.is_pending = False;
				G2A_SEND_DOUBLE(COM_VIDEO_LUMA_SHIFT, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, (double)fl_get_slider_value(fl_object));
			}
			break;
		case COM_VIDEO_BRIGHTNESS_SLIDER_RESET:
			fl_set_slider_value(Video->video_brightness_slider, BRIGHTNESS_DEFAULT);
			command(Video->video_brightness_slider, COM_VIDEO_BRIGHTNESS_SLIDER);
			break;
		case COM_VIDEO_CHROMA_SATURATION_SLIDER:
			if (Chroma_saturation.not_ready) {
				Chroma_saturation.pending = fl_get_slider_value(fl_object);
				Chroma_saturation.is_pending = True;
			} else {
				Chroma_saturation.not_ready = True;
				Chroma_saturation.is_pending = False;
				G2A_SEND_DOUBLE(COM_VIDEO_CHROMA_SATURATION, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, (double)fl_get_slider_value(fl_object));
			}
			break;
		case COM_VIDEO_CHROMA_SATURATION_SLIDER_RESET:
			fl_set_slider_value(Video->video_chroma_saturation_slider, CHROMA_SATURATION_DEFAULT);
			command(Video->video_chroma_saturation_slider, COM_VIDEO_CHROMA_SATURATION_SLIDER);
			break;
		case COM_VIDEO_ASPECT_RATIO_SQUARE_PELS:
			Aspect_ratio_mode = ASPECT_RATIO_SQUARE_PELS;
			G2A_SEND_INT(COM_VIDEO_ASPECT_RATIO, 0, ASPECT_RATIO_SQUARE_PELS);
			break;
		case COM_VIDEO_ASPECT_RATIO_CODED:
			Aspect_ratio_mode = ASPECT_RATIO_CODED;
			G2A_SEND_INT(COM_VIDEO_ASPECT_RATIO, 0, ASPECT_RATIO_CODED);
			break;
		case COM_VIDEO_ASPECT_RATIO_FREE:
			Aspect_ratio_mode = ASPECT_RATIO_FREE;
			G2A_SEND_INT(COM_VIDEO_ASPECT_RATIO, 0, ASPECT_RATIO_FREE);
			break;
		case COM_VIDEO_LAYOUT_FILL_WINDOW:
			G2A_SEND_INT(COM_VIDEO_LAYOUT, 0, VIDEO_LAYOUT_FILL_WINDOW);
			break;
		case COM_VIDEO_LAYOUT_CENTERED:
			G2A_SEND_INT(COM_VIDEO_LAYOUT, 0, VIDEO_LAYOUT_CENTERED);
			break;
		case COM_VIDEO_LAYOUT_TOP_LEFT:
			G2A_SEND_INT(COM_VIDEO_LAYOUT, 0, VIDEO_LAYOUT_TOP_LEFT);
			break;
		case COM_SYNC_I:
			/* fast forward disabled */
			if (Controls->play_pause->visible) {
				fl_hide_object(Controls->fast_forward);
			}
			G2A_SEND_INT(COM_SYNC_TYPE, 0, SYNC_I);
			Framerate_is_nominal = False;
			Sync_type = SYNC_I;
			break;		
		case COM_SYNC_IP:
			/* fast forward disabled */
			if (Controls->play_pause->visible) {
				fl_hide_object(Controls->fast_forward);
			}
			G2A_SEND_INT(COM_SYNC_TYPE, 0, SYNC_IP);
			Framerate_is_nominal = False;
			Sync_type = SYNC_IP;
			break;		
		case COM_SYNC_IPB:
			/* fast forward disabled */
			if (Controls->play_pause->visible) {
				fl_hide_object(Controls->fast_forward);
			}
			G2A_SEND_INT(COM_SYNC_TYPE, 0, SYNC_IPB);
			Framerate_is_nominal = False;
			Sync_type = SYNC_IPB;
			break;		
		case COM_SYNC_NOMINAL:
			/* fast forward enabled */
			if (Controls->play_pause->visible) {
				fl_show_object(Controls->fast_forward);
			}
			G2A_SEND_DOUBLE(COM_SYNC_FRAMERATE, 0, 0.0);
			Framerate_is_nominal = True;
			G2A_SEND_INT(COM_SYNC_TYPE, 0, SYNC_FRAMERATE);
			Sync_type = SYNC_FRAMERATE;
			break;		
		case COM_SYNC_SYNC:
			set_sync_sync();
			break;		
		case COM_SYNC_FRAMERATE:
			Specified_framerate_is_nominal = False;
			if (fl_get_button(Sync->sync_sync)) {
				G2A_SEND_DOUBLE(com, 0, (double)fl_get_slider_value(fl_object));
				Framerate_is_nominal = False;
			}
			break;
		case COM_SYNC_FRAMERATE_RESET:
			fl_set_slider_value(Sync->sync_framerate_slider, Nominal_framerate);
			Specified_framerate_is_nominal = True;
			if (fl_get_button(Sync->sync_sync)) {
				G2A_SEND_DOUBLE(COM_SYNC_FRAMERATE, 0, 0.0);
				Framerate_is_nominal = True;
				G2A_SEND_INT(COM_SYNC_TYPE, 0, SYNC_FRAMERATE);
			}
			break;
		case COM_AUDIO_QUALITY_LOW:
			Audio_quality = AUDIO_QUALITY_LOW;
			G2A_SEND_INT(COM_AUDIO_QUALITY, 0, AUDIO_QUALITY_LOW);
			break;
		case COM_AUDIO_QUALITY_MEDIUM:
			Audio_quality = AUDIO_QUALITY_MEDIUM;
			G2A_SEND_INT(COM_AUDIO_QUALITY, 0, AUDIO_QUALITY_MEDIUM);
			break;
		case COM_AUDIO_QUALITY_HIGH:
			Audio_quality = AUDIO_QUALITY_HIGH;
			G2A_SEND_INT(COM_AUDIO_QUALITY, 0, AUDIO_QUALITY_HIGH);
			break;
		case COM_AUDIO_PORT_LINE_OUT:	/* FALLTHROUGH */
		case COM_AUDIO_PORT_HEADPHONE:	/* FALLTHROUGH */
		case COM_AUDIO_PORT_SPEAKER:
			{
				int		ports = 0;
				if (fl_get_button(Audio->audio_port_line_out)) {
					ports |= AUDIO_PORT_LINE_OUT;
				}
				if (fl_get_button(Audio->audio_port_headphone)) {
					ports |= AUDIO_PORT_HEADPHONE;
				}
				if (fl_get_button(Audio->audio_port_speaker)) {
					ports |= AUDIO_PORT_SPEAKER;
				}
				G2A_SEND_INT(COM_AUDIO_PORTS_SELECT, 0, ports);
			}			
			break;
#ifdef VCD
		case COM_VCD_TRACK:
			{
				u_int	new_track = fl_get_counter_value(Controls->vcd_track);
				
				if (new_track != Vcd_track) {
					PMP_DATA_GENERIC data_generic;
					char *url = data_generic.data_string;

					Vcd_track = new_track;

					url[sizeof data_generic.data_string - 1] = 0;
					sprintf(url, "vcd:%u", Vcd_track);
					if (url[sizeof data_generic.data_string - 1] != 0) abort_now();

					G2A_SEND_STRING(COM_OPEN_STREAM_URL, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, url);
				}
			}
			break;
#endif
#ifdef MTV_EASTER_EGG
		case COM_DEMO:
			G2A_SEND_INT(COM_DEMO, PMP_FLAG_RETURN_STATUS|PMP_FLAG_RETURN_ORIGINAL, 0);
			break;
#endif
		default:
			G2A_SEND_VOID(com, 0);
			break;
	}
}

/*--------------------------------------------------------------------------+
| get_default_preferences_filename											|
+--------------------------------------------------------------------------*/
static const char *
get_default_preferences_filename()
{
	const char *home;
	static char filename[PATH_MAX];
	
	if (filename[0] != 0) {
		return filename;
	}
	
	{
		const char *p = getenv("MTV_PREFERENCE_FILE");
		
		if (p != NULL) {
			return p;
		}
	}
	
	home = getenv("HOME");

	if (home == NULL) {
		MESSAGE(("pref: env variable HOME not defined, using /"));
		home = "";
	}

	filename[sizeof filename - 1] = 0;
	sprintf(filename, "%s/.mtvrc", home);
	if (filename[sizeof filename - 1] != 0) abort_now();

	return filename;
}


/*--------------------------------------------------------------------------+
| user_options_set_default													|
+--------------------------------------------------------------------------*/
static void
user_options_set_default()
{	
	User_options.usage = False;
	User_options.loop = False;
	User_options.zoom = ZOOM_1X1;
	User_options.auto_play = True;
	User_options.dither_type = (DITHER_TYPE)-1;
	User_options.sync_type = SYNC_FRAMERATE;
	User_options.msg_select = 0;
	User_options.audio_msg_select = 0;
#if 0 /* Tristan Savatier - Sun Aug 23 02:08:36 PDT 1998 */
//	User_options.port = 0;
//	User_options.port_udp = 0;
//	User_options.stdin_mode = False;
#endif /* 0 - Tristan Savatier - Sun Aug 23 02:08:36 PDT 1998 */
	User_options.gamma = GAMMA_DEFAULT;
	User_options.contrast = CONTRAST_DEFAULT;
	User_options.brightness = BRIGHTNESS_DEFAULT;
	User_options.chroma_saturation = CHROMA_SATURATION_DEFAULT;
	User_options.audio_mode = AUDIO_MODE_PLAY;
	User_options.audio_quality = AUDIO_QUALITY_MEDIUM;
	User_options.audio_channel = AUDIO_CHANNEL_LEFT;
	User_options.audio_controls = AUDIO_CONTROLS_PCM;
	User_options.audio_pcm_volume = INITIAL_PCM_VOLUME;
	User_options.audio_pcm_balance = INITIAL_PCM_BALANCE;
	User_options.audio_master_volume = INITIAL_MASTER_VOLUME;
	User_options.audio_master_balance = INITIAL_MASTER_BALANCE;
	User_options.audio_ports = -1;
	User_options.xforms_argv = NULL;
	User_options.pmp_base_fd = 3;
	User_options.existing_window = 0;
	User_options.video_layout = VIDEO_LAYOUT_FILL_WINDOW;
	User_options.loop_start_pos = 0;
	User_options.loop_end_pos = 0;
	User_options.mtvp_geometry_x = 200;
	User_options.mtvp_geometry_y = 300;
	User_options.geometry_x = 200;
	User_options.geometry_y = 192+80-95;
	User_options.geometry_w = 490;
	User_options.geometry_h = 95;
	User_options.pref_filename = get_default_preferences_filename();
}


/*--------------------------------------------------------------------------+
| oneliner_timer_CB															|
+--------------------------------------------------------------------------*/
void
oneliner_timer_CB(
	FL_OBJECT *fl_object,
	long param
)
{
	FL_OBJECT *ob = Oneliner_object;
	
	if (ob != NULL && ob->u_vdata != NULL && ob->visible) {
		fl_show_oneliner((const char *)ob->u_vdata, ob->form->x + ob->x + ob->w / 2,
                        ob->form->y + ob->y + ob->h + 5);
	}
}

/*--------------------------------------------------------------------------+
| post_handler																|
+--------------------------------------------------------------------------*/
static int
post_handler(
	FL_OBJECT *ob,
	int event,
    FL_Coord mx,
	FL_Coord my,
	int key,
	void *xev
)
{
	if(event == FL_ENTER) {
     	Oneliner_object = ob;
		fl_set_timer(Controls->oneliner_timer, 1.0);
	} else if (Oneliner_object == ob && (event == FL_PUSH
										|| event == FL_RELEASE
										|| event == FL_LEAVE
										|| event == FL_MOUSE
										|| event == FL_UNFOCUS
										|| event == FL_KEYBOARD
										|| event == FL_STEP
										|| event == FL_SHORTCUT
										|| event == FL_FREEMEM
										|| event == FL_OTHER
										|| event == FL_DBLCLICK
										|| event == FL_TRPLCLICK
										|| event == FL_PS)) {
		fl_set_timer(Controls->oneliner_timer, 0.0);
		Oneliner_object = NULL;
		fl_hide_oneliner();
	}
	
	/*
	 * Work around a bug of the IN-OUT buttons in Xforms!
	 * When the mouse pointer gets out of an IN-OUT button
	 * while depressed, the IN-OUT button goes in the OUT state
	 * (depressed, value 0) without calling the callback!
	 * To work around this bug, we use the post-handler.
	 */
	if (Fast_forward && ob == Controls->fast_forward && event == FL_LEAVE) {
		quit_fast_forward_mode();
	}
	
	/*
	 * Handle the main slide-bar. We need special handling
	 * because if the slide-bar is moved
	 * while playing, we want the player to pause while seeking
	 * and to restart playing when the slide-bar is released.
	 */
	if (ob == Controls->frame_slider) {		
		if (event == FL_PUSH) {
			Seeking = True;
			State_before_seeking = Gui_state;
			/*
			 * STATE_FLAG_NO_ECHO is necessary to avoid a race condition that
			 * may cause the player to stop when seeking while playing.
			 * The race condition would occur if the A2G_COM_STATE corresponding
			 * to this STATE_PAUSED was received AFTER the FL_RELEASE and
			 * if the A2G_COM_STATE corresponding to the send_state() done
			 * at release is received when Seeking is set again later.
			 */
			send_state(STATE_PAUSED|STATE_FLAG_NO_ECHO);
			fl_set_object_label(Controls->actual_framerate, "n/a");
			if (State_before_seeking == STATE_STOPPED) {
				update_state(STATE_PAUSED);
			}
			
		} else if (event == FL_RELEASE) {
			Seeking = False;
			/*
			 * Be careful: State_before_seeking may have changed
			 * since the push, because a STATE_STOP was received
			 * while seeking. This can happen if the used catches the
			 * slide bar right before the player reach the end of a stream.
			 * This is a race condition!
			 */
			if (State_before_seeking == STATE_PLAYING) {
				/*
				 * STATE_FLAG_NO_ECHO is set because we do not want to
				 * receive the A2G_COM_STATE message caused by this
				 * COM_STATE command.
				 */
				send_state(STATE_PLAYING|STATE_FLAG_NO_ECHO);
			} else if (State_before_seeking == STATE_STOPPED) {
				/*
				 * This is necessary to avoid a race condition that
				 * may cause the player get in the stop state after seeking.
				 * The race condition would occur if the A2G_COM_STATE corresponding
				 * to a STATE_STOP was received AFTER the FL_PUSH.
				 */
				send_state(STATE_PAUSED);
			}
		}
	}

	return 0;
}

/*--------------------------------------------------------------------------+
| set_oneliners																|
+--------------------------------------------------------------------------*/
static void
set_oneliners()
{
	fl_mapcolor(FL_FREE_COL1, 255, 250, 205);
	fl_set_oneliner_color(FL_BLACK, FL_FREE_COL1);
	
	SET_ONELINER(Controls->play_pause, "Play/Pause");
	SET_ONELINER(Controls->step, "Step to next frame");
	SET_ONELINER(Controls->stop, "Stop");
	SET_ONELINER(Controls->time_total, "Total play time HH:MM:SS");
	SET_ONELINER(Controls->time_elapsed, "Elapsed play time HH:MM:SS");
	SET_ONELINER(Controls->end, "Go to end");
	SET_ONELINER(Controls->actual_framerate, "Actual framerate\nin frames per second");
	SET_ONELINER(Controls->number_of_frames_played, "Number of frames played\nsince last random access");
	SET_ONELINER(Controls->loop, "Auto-repeat on/off");
	SET_ONELINER(Controls->audio_mute, "Audio on/off");
	SET_ONELINER(Controls->zoom, "Zoom (2x2) on/off");
	SET_ONELINER(Controls->fast_forward, "Fast forward");
	SET_ONELINER(Controls->rewind, "Rewind");
	SET_ONELINER(Controls->frame_slider, "Move to any position in the sequence");

	SET_ONELINER(Controls->audio_volume_slider, "If audio mixer is available,\nVolume and Balance controls\ncan be attached to PCM or Master levels\n(see Audio options)");
	SET_ONELINER(Controls->audio_balance_slider, "If audio mixer is available,\nVolume and Balance controls\ncan be attached to PCM or Master levels\n(see Audio options)");
	SET_ONELINER(Controls->vcd_track, "Select VCD track");
	
	SET_ONELINER(Audio->audio_controls_pcm, "Attach Volume and Balance controls\nto PCM levels");
	SET_ONELINER(Audio->audio_controls_master, "Attach Volume and Balance controls\nto Master levels (if audio mixer is available)");
	SET_ONELINER(Audio->audio_quality_low, "ulaw, if supported (8 KHz)");
	SET_ONELINER(Audio->audio_quality_medium, "Half sampling freq. (22 KHz)");
	SET_ONELINER(Audio->audio_quality_high, "Nominal sampling freq. (44 KHz)");
	SET_ONELINER(Audio->audio_port_line_out, "Line-out jack");
	SET_ONELINER(Audio->audio_port_headphone, "Headphone jack");
	SET_ONELINER(Audio->audio_port_speaker, "Built-in speaker");
	SET_ONELINER(Audio->audio_ch_both, "Play both channels\nwhen audio is coded in stereo\nor in dual-channel mode");
	SET_ONELINER(Audio->audio_ch_right, "Play right channel only\nwhen audio is coded in stereo\nor in dual-channel mode");
	SET_ONELINER(Audio->audio_ch_left, "Play left channel only\nwhen audio is coded in stereo\nor in dual-channel mode");

	SET_ONELINER(Infos->mux_bitrate, "Total bitrate of multiplexed stream\n(Video + Audio + Systems)");
	SET_ONELINER(Infos->video_framerate, "Nominal framerate");
	SET_ONELINER(Infos->audio_frequency, "Nominal sampling frequency");
	SET_ONELINER(Infos->video_constrained, "Indicates if the constrained-parameter\nflag is set");
	SET_ONELINER(Infos->video_pel_aspect_ratio, "Pixel aspect ratio\nheight / width of pel");
	SET_ONELINER(Infos->video_compliant, "Indicates if this is a compliant\nconstrained-parameter MPEG-1 video bitstream");

	SET_ONELINER(Sync->sync_i, "Play only I-frames\nas fast as possible (no audio)");	
	SET_ONELINER(Sync->sync_ip, "Play only I- and P-frames\nas fast as possible (no audio)");	
	SET_ONELINER(Sync->sync_ipb, "Play all video frames\nas fast as possible (no audio)");	
	SET_ONELINER(Sync->sync_nominal, "Play video at the nominal framerate\n(audio is played, if any");
	SET_ONELINER(Sync->sync_sync, "Play video at the specified framerate\nNote: Audio can be played only at Nominal framerate");
	SET_ONELINER(Sync->sync_framerate_slider, "Select video framerate");	
	SET_ONELINER(Sync->sync_reset_framerate_slider, "Reset video framerate to Nominal framerate\nAudio can be played only at Nominal framerate");
	SET_ONELINER(Video->video_dither_none, "video off");
	SET_ONELINER(Video->video_dither_gray, "gray levels");
	SET_ONELINER(Video->video_dither_ordered, "dithered color");
	SET_ONELINER(Video->video_dither_color, "full color");
	SET_ONELINER(Video->video_aspect_ratio_square_pels, "use square pels");
	SET_ONELINER(Video->video_aspect_ratio_coded, "use coded pel aspect ratio\n(when supported)");
	SET_ONELINER(Video->video_aspect_ratio_free, "fill-up window\n(when supported)");
	
}


/*--------------------------------------------------------------------------+
| do_gui_fork																|
+--------------------------------------------------------------------------*/
static void
do_gui_fork(
	const char *argv0,
	const char **argv
)
{
	int		argc;
	static const char	*fl_argv[] = {
		/* app name updated later */
		NULL,
		
		/* border width -1 looks better! */
		"-bw",
		"-1",
		
		/* try to work-around the famous Xforms clipped-menus bug: */
#ifdef SGI		
		"-visual",
		"Default",
		"-depth",
		"8",
#else
		"-visual",
		"PseudoColor",
#endif
		/* probably a good idea */
		"-shared",

		NULL,
	};
	
	if (argv == NULL) {
		argv = fl_argv;
	}
	argv[0] = argv0;
	{
		const char **p = argv;
		
		for (argc = 0; p[argc] != NULL; argc++);
	}
	
	/*
	 * Check that an X server is available.
	 * We do that because there is a bug in Xforms causing fl_initialize to
	 * kill us if it cannot open the X server.
	 */
	{
		Display *display;
		if ((display = XOpenDisplay (NULL)) == NULL) {
			FATAL_ERROR (("cannot open X11 display (check env variable DISPLAY)"));
		}
		XCloseDisplay(display);
	}

	if (fl_initialize (&argc, (char **)argv, "FormDemo", 0, 0) == NULL) {
		FATAL_ERROR(("cannot initialize xform"));
	}
	
	/* check version */
	{
		int		version, revision;
		
		if (fl_library_version(&version, &revision) != FL_VERSION*1000 + FL_REVISION) {
			fprintf(stderr, "mtv: warning: xforms dynamic library (libforms.s%c) version mismatch: need version %d.%d, using version %d.%d. Proceed at your own risks!\n",
#ifdef HP
				'l',
#else
				'o',
#endif
				FL_VERSION, FL_REVISION, version, revision);
		}
	}
	 
	{
#include "mpegtv.xbm"
		Icon_pixmap = fl_create_from_bitmapdata(fl_default_win(), mpegtv_bits, mpegtv_width, mpegtv_height);
	}

#if 0 /* Tristan Savatier - Mon Mar 16 00:15:19 PST 1998 */
//	fl_set_idle_callback((FL_APPEVENT_CB)my_idle, NULL);
#endif /* 0 - Tristan Savatier - Mon Mar 16 00:15:19 PST 1998 */

	/* if I don't do that, I get killed when closing goodies (with the WM) */
	fl_set_atclose(at_close_ignore, NULL);
	
	if ((Controls = create_form_controls()) == NULL) {
		FATAL_ERROR(("cannot create Controls form"));
	}
	fl_set_form_atclose(Controls->controls, at_close_controls, NULL);
	fl_set_form_icon(Controls->controls, Icon_pixmap, 0);
#if 0 /* Tristan Savatier - Sun Jan 11 20:29:21 PST 1998 */
//	fl_set_form_atactivate(Controls->controls, at_activate_controls, NULL);
//	fl_set_form_atdeactivate(Controls->controls, at_deactivate_controls, NULL);
#endif /* 0 - Tristan Savatier - Sun Jan 11 20:29:21 PST 1998 */
	
	if ((About = create_form_about()) == NULL) {
		FATAL_ERROR(("cannot create About form"));
	}
	fl_set_form_atclose(About->about, at_close_hide, NULL);
	fl_set_form_icon(About->about, Icon_pixmap, 0);

	if ((License = create_form_license()) == NULL) {
		FATAL_ERROR(("cannot create License form"));
	}
	fl_set_form_atclose(License->license, at_close_ignore, NULL);
	fl_set_form_icon(License->license, Icon_pixmap, 0);

	if ((Splash = create_form_splash()) == NULL) {
		FATAL_ERROR(("cannot create Splash form"));
	}
	fl_set_form_atclose(Splash->splash, at_close_quit, NULL);
	fl_set_form_icon(Splash->splash, Icon_pixmap, 0);

	if ((Registration = create_form_registration()) == NULL) {
		FATAL_ERROR(("cannot create Registration form"));
	}
	fl_set_form_atclose(Registration->registration, at_close_hide, NULL);
	fl_set_form_icon(Registration->registration, Icon_pixmap, 0);
	fl_set_input_maxchars(Registration->order_number, 10);
	fl_set_input_filter(Registration->order_number, registration_order_number_filter);
#ifdef KEY_ASCII_LENGTH
	fl_set_input_maxchars(Registration->key, KEY_ASCII_LENGTH);
#else
	fl_set_input_maxchars(Registration->key, 8);
#endif
	fl_set_input_filter(Registration->key, registration_key_filter);
	fl_set_focus_object(Registration->registration, Registration->name);
	if (Registered) {
		fl_set_input(Registration->name, Name);
		fl_set_input(Registration->order_number, Order_number);
		fl_set_input(Registration->key, Key);
	}


#ifdef _REENTRANT
	if ((Udp = create_form_udp()) == NULL) {
		FATAL_ERROR(("cannot create Udp form"));
	}
	fl_set_form_atclose(Udp->udp, at_close_hide, NULL);
	fl_set_form_icon(Udp->udp, Icon_pixmap, 0);
	fl_set_input_maxchars(Udp->port, 5);
#endif

	if ((Audio = create_form_audio()) == NULL) {
		FATAL_ERROR(("cannot create Audio form"));
	}
	fl_set_form_atclose(Audio->audio, at_close_hide, NULL);
	fl_set_form_icon(Audio->audio, Icon_pixmap, 0);
	
	if ((Video = create_form_video()) == NULL) {
		FATAL_ERROR(("cannot create Video form"));
	}
	fl_set_form_atclose(Video->video, at_close_hide, NULL);
	fl_set_form_icon(Video->video, Icon_pixmap, 0);
	
	if ((Sync = create_form_sync()) == NULL) {
		FATAL_ERROR(("cannot create Sync form"));
	}
	fl_set_form_atclose(Sync->sync, at_close_hide, NULL);
	fl_set_form_icon(Sync->sync, Icon_pixmap, 0);
	
	if ((Infos = create_form_infos()) == NULL) {
		FATAL_ERROR(("cannot create Infos form"));
	}
	fl_set_form_atclose(Infos->infos, at_close_hide, NULL);
	fl_set_form_icon(Infos->infos, Icon_pixmap, 0);
	
	if ((Frame_infos = create_form_frame_infos()) == NULL) {
		FATAL_ERROR(("cannot create Frame_infos form"));
	}
	fl_set_form_atclose(Frame_infos->frame_infos, at_close_hide, NULL);
	fl_set_form_icon(Frame_infos->frame_infos, Icon_pixmap, 0);
	
	{
		u_int		i;
		for (i = 0; i < sizeof License_lines / sizeof *License_lines; i++) {
			fl_add_browser_line(License->browser, License_lines[i]);
		}
		fl_set_browser_fontstyle(License->browser, FL_FIXED_STYLE);
	}

	{
		char buf[256];
		
		buf[0] = buf[sizeof buf - 1] = 0;
		strcat(buf, "Open File...|Play from URL...|Play from Stdin|");
#ifdef _REENTRANT
		strcat(buf, "Play from UDP Socket...|");
#endif
#ifdef VCD
		strcat(buf, "Play VCD|");
#endif
#ifdef USE_SMC
		strcat(buf, "Play SMC|");
#endif
		strcat(buf, "Close|Save as...|Sequence infos%l|Quit");
		if (buf[sizeof buf - 1] != 0) {
			abort_now();
		}
		fl_set_menu(Controls->file_menu, buf);
	}
#if 0 /* Tristan Savatier - Mon Mar 16 01:19:37 PST 1998 */
//#ifdef VCD
//	fl_set_menu(Controls->file_menu, "Open File...|Play from Stdin|Play VCD|Close|Save as...|Sequence infos%l|Quit");
//#else
//	fl_set_menu(Controls->file_menu, "Open File...|Play from Stdin|Close|Save as...|Sequence infos%l|Quit");
//#endif
#endif /* 0 - Tristan Savatier - Mon Mar 16 01:19:37 PST 1998 */

	fl_set_menu(Controls->options_menu, "Audio...|Video...|Framerate & Sync...|Auto play|Save Options");

#ifdef USE_SDL
	fl_set_menu(Controls->window_menu, "Full screen|Raise window|Play in root window|Frame infos|Capture Frame (PPM format)...|Capture Frame (TGA format)...");
#else
	fl_set_menu(Controls->window_menu, "Raise window|Play in root window|Frame infos|Capture Frame (PPM format)...|Capture Frame (TGA format)...");
#endif
	fl_set_menu_item_mode(Controls->window_menu, WINDOW_MENU_ROOT, FL_PUP_BOX);

	fl_set_menu(Controls->help_menu, "About mtv...|Release Notes...|User's Manual...|FAQ...|Support...|License...%l|Register...|Enter Key...%l|About MPEG...|MPEG Bitstreams...");

#ifndef USE_XIL
	fl_hide_object(Video->video_aspect_ratio_coded);
	fl_hide_object(Video->video_aspect_ratio_free);
#endif

	fl_set_slider_bounds(Sync->sync_framerate_slider, FRAMERATE_MIN, FRAMERATE_MAX);
	fl_set_slider_bounds(Controls->audio_volume_slider, 0.0, 1.0);

	fl_set_slider_bounds(Controls->audio_balance_slider, 0.0, 1.0);
	fl_set_slider_size(Controls->frame_slider, 0.04);

	fl_set_slider_bounds(Video->video_gamma_slider, GAMMA_MIN, GAMMA_MAX);
	fl_set_slider_size(Video->video_gamma_slider, 0.04);
	
	fl_set_slider_bounds(Video->video_contrast_slider, CONTRAST_MIN, CONTRAST_MAX);
	fl_set_slider_size(Video->video_contrast_slider, 0.04);

	fl_set_slider_bounds(Video->video_brightness_slider, BRIGHTNESS_MIN, BRIGHTNESS_MAX);
	fl_set_slider_size(Video->video_brightness_slider, 0.04);
	
	fl_set_slider_bounds(Video->video_chroma_saturation_slider, CHROMA_SATURATION_MIN, CHROMA_SATURATION_MAX);
	fl_set_slider_size(Video->video_chroma_saturation_slider, 0.04);
	
	/*fl_set_form_geometry(Controls->controls, 200, 192+80-95, 490, 95);*/
	fl_set_form_geometry(Controls->controls,
							(FL_Coord)User_options.geometry_x,
							(FL_Coord)User_options.geometry_y,
							(FL_Coord)User_options.geometry_w,
							(FL_Coord)User_options.geometry_h);

	hide_stream_controls();
	fl_hide_object(Video->video_dither_none);
	fl_hide_object(Video->video_dither_gray);
	fl_hide_object(Video->video_dither_ordered);
	fl_hide_object(Video->video_dither_ordered_colormap_private);
	fl_hide_object(Video->video_dither_color);

	fl_hide_object(Audio->audio_port_line_out);
	fl_hide_object(Audio->audio_port_headphone);
	fl_hide_object(Audio->audio_port_speaker);

	
	fl_add_io_callback(Player.receive_channel.msg_pipe[PMP_FD_READ], FL_READ, (void (*)(int, void *))read_mtvp_message, NULL);

	fl_set_object_label(About->version, VERSION_STR);
	
	set_oneliners();

	if (Registered) {
		show_form_controls();
	} else {
		MpegTV = "MpegTV [UNREGISTERED]";
		{
			time_t	now;
			time_t	expire;
		
			now = time((time_t *)NULL);
			
			expire = mktime(&Expiration_time);
			
			if (now > expire) {
				show_alert("This UNREGISTERED copy of mtv is very old", "You should download a more recent version or register this copy", "Registered copies do not display this notice", 1);
				
#if 0 /* Tristan Savatier - Mon Jan 25 11:54:33 PST 1999 */
//				fl_set_object_label(Splash->splash_greetings, "This UNREGISTERED copy of MpegTV Player 1.0\nhas expired.\n\nYou can download a more recent version from:\nwww.mpegtv.com/download.html\n\nor REGISTER your copy at:\nwww.mpegtv.com\n\nRegistered copies do not expire.\n\nThe Shareware license does not authorize commercial\nor govermental use of this product.\n\nFor commercial\nor govermental use you must purchase a commercial license");
//				fl_set_object_label(Splash->button_try, "Quit");
//				Expired = True;
#endif /* 0 - Tristan Savatier - Mon Jan 25 11:54:33 PST 1999 */
			}
		}
		show_splash_if_not_registered();
	}

	Timeout_id = fl_add_timeout(1000L, update_slider_timeout_CB, NULL);
#if 0 /* Tristan Savatier - Tue Sep 1 19:59:03 PDT 1998 */
//	fl_set_timer(Controls->update_slider_timer, 1.0);
#endif /* 0 - Tristan Savatier - Tue Sep 1 19:59:03 PDT 1998 */

	while(True) {
#ifdef MSG_DEBUG
		printf("calling fl_do_forms\n");
#endif
		fl_do_forms();
	}
}


/*--------------------------------------------------------------------------+
| my_own_error_handler														|
+--------------------------------------------------------------------------*/
static void
my_own_error_handler(const char *func, const char *fmt,...)
{
	/* our dummy Xforms warning message routine does nothing */
}

#if !defined(__osf__)
/*--------------------------------------------------------------------------+
| whereError																|
+--------------------------------------------------------------------------*/
void
(*
#if defined(LINUX) && defined(HANDLE_CHALLENGE) && defined(__GNUC__)
__wrap_whereError
#else
whereError
#endif
(int something,
                  int level,
                  const char *string,
                  int some_int))(const char *func, const char *fmt,...)
{
	/* whereError() shadows Xforms warning message routine */

    return my_own_error_handler;
}
#endif


/*--------------------------------------------------------------------------+
| handler_exit																|
+--------------------------------------------------------------------------*/
/*ARGSUSED*/
static void
handler_exit(
	int     sig
)
{
	MESSAGE(("(pid=%u): received signal %d (aborting)", (u_int)Pid, sig));

	clean_exit(EXIT_FAILURE);
}


#if 0 /* Tristan Savatier - Wed Feb 17 04:44:32 PST 1999 */
///*--------------------------------------------------------------------------+
//| handler_sigchld															|
//+--------------------------------------------------------------------------*/
///*ARGSUSED*/
//static void
//handler_sigchld(
//	int     sig
//)
//{
//	/*
//	 * Note: If you want to call any Xlib or Xforms routine from
//	 * a signal handler, you must register it using
//	 * fl_add_signal_callback() so that
//	 * the signal handler will be executed in the proper context
//	 * in the Xforms event loop.  Please refer to the Xforms
//	 * manual, Signals section.
//	 *
//	 * This particular signal handler does not
//	 * call any Xlib or Xforms routine, so it is safe
//	 * to install it with signal() or sigaction().
//	 */
//
//	MESSAGE(("(pid=%u): received signal %d, child process %u died, exiting...", (u_int)Pid, sig, (u_int)Player.pid));
//	clean_exit(EXIT_FAILURE);
//}
#endif /* 0 - Tristan Savatier - Wed Feb 17 04:44:32 PST 1999 */


/*--------------------------------------------------------------------------+
| init_signals																|
+--------------------------------------------------------------------------*/
static void
init_signals()
{
	struct sigaction sigact;

	/* terminate with handler_exit if any unexpected signal is received */ 
	portable_catch_termination_signals(handler_exit);

#if 0 /* Tristan Savatier - Wed Feb 17 04:44:25 PST 1999 */
//	sigact.sa_handler = handler_sigchld;
//	sigemptyset(&sigact.sa_mask);
//	sigact.sa_flags = SA_NOCLDSTOP;
//	if (sigaction(SIGCHLD, &sigact, NULL) != 0) {
//		FATAL_ERROR(("can't set handler for SIGCHLD"));
//	}
#endif /* 0 - Tristan Savatier - Wed Feb 17 04:44:25 PST 1999 */
}


/*--------------------------------------------------------------------------+
| get_int0																	|
+--------------------------------------------------------------------------*/
static int
get_int0(
	char	o,
	const char	**p,	/* side effect on purpose: *p will be updated */
	int	vmin,
	int	vmax
)
{
	const char	*s = *p;
	int		n;

	n = strtol(s, (char **)p, 10);
	if (*p == s) {
		FATAL(("bad option %c%s (number expected)", o, s));
	}
	if (vmin <= vmax && (n < vmin || n > vmax)) {
		FATAL(("bad option %c%d (%d out of range [%d, %d])", o, n, n, vmin, vmax));
	}
	return n;
}


/*--------------------------------------------------------------------------+
| get_int																	|
+--------------------------------------------------------------------------*/
static int
get_int(
	char	o,
	const char	*s,
	int		vmin,
	int		vmax
)
{
	return get_int0(o, &s, vmin, vmax);
}

/*--------------------------------------------------------------------------+
| get_double																|
+--------------------------------------------------------------------------*/
static double
get_double(
	char	o,
	const char	*s,
	double		vmin,
	double		vmax
)
{
	char	*p;
	double	d;

	d = strtod(s, &p);
	if (p == s) {
		FATAL(("bad option %c%s (floating number expected)", o, s));
	}

	if (vmin <= vmax && (d < vmin || d > vmax)) {
		FATAL(("bad option %c%4.2f (%4.2f out of range [%4.2f, %4.2f])", o, d, d, vmin, vmax));
	}

	return d;
}


/*--------------------------------------------------------------------------+
| usage - print the help.													|
+--------------------------------------------------------------------------*/
static void
usage()
{
	static const char *helptab[] = {
"Copyright (c) 1997-1998, MpegTV LLC\n",
"ALL RIGHTS RESERVED\n",
"\n",
"Portions of this software:\n",
"Copyright (c) 1995-1998 by T.C. Zhao and Mark Overmars\n",
"ALL RIGHTS RESERVED\n",
"\n",
"Shareware license is for personal and non-profit use only.\n",
"\n",
"Shareware license does not allow commercial or governmental\n",
"use of this product. Contact sales@mpegtv.com for commercial\n",
"license information.\n",
"\n",
"Please read the LICENSE file. Use at your own risks!\n",
"MpegTV Home page: http://www.mpegtv.com\n",
"\n",
"Function:  Graphic front-end for the MpegTV run-time Player.\n",
"           The MpegTV run-time Player (mtvp) is a realtime\n",
"           software-only MPEG Video player with audio/sync\n",
"Note:      mtv requires mtvp, the MpegTV run-time Player\n",
"Usage:     mtv {options} [URL]\n",
"           URL must be the location of an MPEG-1 Video or Systems stream.\n",
"           URL may be:\n",
"           - <directory> (will open a file selector in this directory)\n",
"           - file:<filename> or just <filename>\n",
"           - fd:<filedescriptor> (e.g. fd:0 for stdin), or '-' for stdin\n",
"           - http://<hostname>/{<dir>/}<filename>\n",
"           - ftp://<hostname>/{<dir>/}<filename>\n",
#ifdef _REENTRANT
"           - udp:<port-number> (e.g. udp:8888)\n",
#endif
#ifdef VCD
"           - vcd:<track-number> (e.g. vcd:1)\n",
#endif
"Current limitations:\n",
"           Decode only MPEG-1 Systems or Video streams\n",
"           GUI does not allow to control all the parameters\n",
"Options:   Single letter options with no parameters can be combined\n",
"           Example: mtv -lz\n",
"           One option with parameters can be appended to a list of\n",
"           options with no parameters. Example: -wxd2\n",
"           No space is allowed between option and parameter.\n",
" -d<n>     display mode:\n",
"           n=0: no display (but frames are decoded and audio is played)\n",
"           n=1: gray\n",
"           n=2: 8-bit pseudo-color\n",
"           n=3: 16-bit true color\n",
"           By default mtv picks the best display mode for your display.\n",
" -s<n>     synchronization type:\n",
"           Note that audio is played only with -s0 (default)\n",
"           n=0: synchronize to match sequence nominal frame-rate (default)\n",
"           n=1: display all frames as fast as possible\n",
"           n=2: display all I and P frames as fast as possible\n",
"           n=3: display all I frames as fast as possible\n",
" -sf<n>    play video at specified framerate [1.0 to 240.0]\n",
"           By default video plays at the nominal framerate (coded in the stream).\n",
"           Note that audio is not played when framerate is specified.\n",
" -g<n>     gamma correction, range [0.1 to 2.0]\n",
"           (default: irix: 0.8, solaris: 1.4, other: 1.0)\n",
"           n large produces lighter images, n small produces darker images\n",
" -c<n>     contrast, range [-0.50 to 0.50] (default 0.0)\n",
"           n > 0 for more contrast, n < 0 for less contrast\n",
" -b<n>     brightness, range [-0.50 to 0.50] (default 0.0)\n",
"           n > 0 for brighter, n < 0 for darker\n",
" -w[<n>]   print informative messages and warnings.\n",
"           Several values can be specified with several -w<n> options.\n",
"           If <n> is not specified all the messages are printed.\n",
"           n=0: general messages\n",
"           n=1: non-fatal error messages\n",
"           n=2: warnings (e.g. when errors are found in MPEG stream)\n",
"           n=3: frame sync messages\n",
"           n=4: audio PTS messages\n",
"           n=5: interprocess communication messages\n",
"           n=6: input stream read messages\n",
"           n=7: systems parsing messages\n",
"           n=8: Xlib messages\n",
"           n=9: indicate systems parser position in messages\n",
"           n=10: indicate video parser position in messages\n",
"           n=11: indicate time in messages\n",
"           n=12: indicate source file name and line in messages\n",
"           n=13: read thread synchronisation messages\n",
"           n=14: Stream Input Handler (SIH) messages\n",
"           n<0: set message mask to -n - 1\n",
" -l        loop (only works when reading from a file)\n",
" -z        zoom (2x2)\n",
" -i        iconify controls\n",
" -G[=]{+-}<x>{+-}<y>  set initial position of player video window\n",
" -F[=][<w>{xX}<h>]{+-}<x>{+-}<y>  set initial position and size\n",
"           of control window\n",
" -a<n>     audio mode\n",
"           n=0: no audio\n",
"           n=1: play audio (default)\n",
"           n=2: output audio stream on stdout\n",
" -aw[<n>]  audio process informative messages and warnings.\n",
"           Several values can be specified with several -aw<n> options.\n",
"           If <n> is not specified all the messages are printed.\n",
"           n=0: general messages\n",
"           n=1: non-fatal error messages\n",
"           n=2: warnings (e.g. when errors are found in MPEG stream)\n",
"           n=4: audio PTS messages\n",
"           n=5: interprocess communication messages\n",
"           n=8: audio sync messages\n",
"           n=9: audio ioctl messages\n",
"           n=10: audio write sample messages\n",
"           n=11: indicate time in messages\n",
"           n=12: indicate source file name and line in messages\n",
"           n<0: set message mask to -n - 1\n",
" -aq<n>    audio quality\n",
"           n=0: low\n",
"           n=1: medium (default)\n",
"           n=2: high\n",
" -ac<n>    audio channel\n",
"           n=0: both (stereo)\n",
"           n=1: left (default)\n",
"           n=2: right\n",
" -ao<n>    set audio output port(s)\n",
"           This option can be used several times to select multiple ports.\n",
"           By default the ports currently selected are used.\n",
"           n=0: internal speaker (if available)\n",
"           n=1: headphone jack (if available)\n",
"           n=2: line-out (if available)\n",
" -aM       link volume and balance controls to Master (mixer) volume if available.\n",
"           by default volume and balance controls are linked to PCM volume\n",
" -a<c><x><n> set initial audio levels\n",
"           n is a double in the range [0.0, 1.0] or -1 to keep the setting unchanged\n",
"           c=p: PCM levels\n",
"           c=m: Master levels (only on systems with a /dev/mixer device)\n",
"           x=v: volume\n",
"           x=b: balance (n=0.0 means left, n=0.5 means centered, n=1.0 means right)\n",
"           By default, options -amv-1 -amb-1 -apv-1 -apb-1 are assumed,\n",
"           i.e. master and PCM volume keep their current setting (possibly 0!).\n",
" -p        auto-play off (default auto-play on)\n",
" -Wr       play in root window (default create new window)\n",
" -W<n>     specify window id of an existing window\n",
"           n=0: create new window (default)\n",
"           n>0: play in existing window with id <n>.\n",
" -Wf<n>    video layout; indicates the position of the video in the window.\n",
"           n=0: centered\n",
"           n=1: fill-up (default)\n",
"           n=2: top-left\n",
" -S<n>     auto-seek byte offset position (default 0). Player will seek to this\n",
"           position right after opening the specified file. Can be used with -p.\n",
" -E<n>     auto-end byte offset position (default end-of-file).\n",
"           When the auto-end position is reached:\n",
"           If Loop on, Player will seek back to auto-seek position and loop.\n",
"           If Loop off, Player will pause after passing this byte offset,\n",
"           or it will exit if auto-exit is on.\n",
"           Auto-seek and auto-end are only valid for the file specified in the\n",
"           command-line and are reset to defaults when user seeks out of the\n",
"           interval [auto-seek, auto-end] or turns Loop off.\n",
" -x<name>  pathname of MpegTV run-time Player executable file (default mtvp)\n",
" +         all the options and parameters following a single + option\n",
"           are passed to the Xforms library.  See Xforms manual for infos.\n",
"Environment variables:\n",
"   MTV_PREFERENCE_FILE = name of mtv preference file (default $HOME/.mtvrc)\n",
#ifdef VCD
#ifdef LINUX
"   MTV_VCD_DEVICE = name of cdrom device for playing VCD (default /dev/cdrom)\n",
#endif
#ifdef SOLARIS
"   MTV_VCD_DEVICE = name of cdrom device for playing VCD\n",
"           Default: /vol/dev/aliases/cdrom0 if Volume management is running\n",
"                    /dev/rdsk/c0t6d0s2 if Volume management is not running\n",
#endif
#endif
"\n",

	NULL,
	};
	const char	**p = helptab;

	printf("MpegTV Player 1.0 - %s\n", VERSION_STR);
	while (*p != NULL) {
		printf(*p++);
	}
	if (Registered) {
		printf("This copy is registered to: %s\n", Name);
	} else {
		time_t expire = mktime(&Expiration_time);
		
		printf("This UNREGISTERED copy expires on: %s", ctime(&expire));
		printf("REGISTERED copies do not expire - Register at www.mpegtv.com\n");
	}
	printf("\nPlease send bug reports and comments to support@mpegtv.com\n");
}

/*--------------------------------------------------------------------------+
| parse_options - parse the option string "p".								|
| The first "-" character is assumed to be skipped before this function		|
| is called.																|
| This function is called when an unknown option is encountered.			|
+--------------------------------------------------------------------------*/
static void
parse_options(
	const char *p
)
{
	int		o;

	for (; ; ) {
		switch (o = *p++) {
		case 'h':	User_options.usage = True; continue;
		case 'w':	if (!isdigit(*p)) {
						User_options.msg_select = (u_int)0x7fffffff;
						MSG_PTR->select = (u_int)0x7fffffff;
						continue;
					} else {
						User_options.msg_select |= (1 << get_int(o, p, 0, 31));
						MSG_PTR->select |= (1 << get_int(o, p, 0, 31));
					}
					return;
		case 'l':	User_options.loop = True;
					User_options.loop_defined = True;
					continue;
		case 'z':	User_options.zoom = ZOOM_2X2;
					User_options.zoom_defined = True;
					continue;
		case 'i':	User_options.iconify = True; continue;
		case 'G':	{
					u_int	w, h;
					XParseGeometry(p,
						&User_options.mtvp_geometry_x,
						&User_options.mtvp_geometry_y,
						&w,
						&h);
					User_options.mtvp_geometry_defined = True;
				}
				return;
		case 'F':	XParseGeometry(p,
						&User_options.geometry_x,
						&User_options.geometry_y,
						&User_options.geometry_w,
						&User_options.geometry_h);
				return;
		case 'a':
			if (isdigit(*p)) {
				User_options.audio_mode = (AUDIO_MODE)get_int(o, p, 0, (int)AUDIO_MODE_COUNT - 1);
				return;
			} else {
				switch (o = *p++) {
				case 'w':
					if (!isdigit(*p) && !(*p == '-' && isdigit(*(p + 1)))) {
						User_options.audio_msg_select = (u_int)0x7fffffff;
						continue;
					} else {
						int		n = get_int(o, p, INT_MIN, 31);
						
						if (n >= 0) {
							User_options.audio_msg_select |= (1 << n);
						} else {
							User_options.audio_msg_select = - (n + 1);
						}
					}
					break;
				case 'q':
					User_options.audio_quality = (AUDIO_QUALITY)get_int(o, p, 0, (int)AUDIO_QUALITY_COUNT - 1);
					User_options.audio_quality_defined = True;
					break;
				case 'c':
					User_options.audio_channel = (AUDIO_CHANNEL)get_int(o, p, 0, (int)AUDIO_CHANNEL_COUNT - 1);
					User_options.audio_channel_defined = True;
					break;
				case 'M':
					User_options.audio_controls = AUDIO_CONTROLS_MASTER; continue;
				case 'p':
					switch (o = *p++) {
					case 'v':
						User_options.audio_pcm_volume = get_double(o, p, -1.0, 1.0);
						break;
					case 'b':
						User_options.audio_pcm_balance = get_double(o, p, -1.0, 1.0);
						break;
					default:
						FATAL(("illegal option: -ap%c", *(p - 1)));
						break;
					}
					break;
				case 'm':
					switch (o = *p++) {
					case 'v':
						User_options.audio_master_volume = get_double(o, p, -1.0, 1.0);
						break;
					case 'b':
						User_options.audio_master_balance = get_double(o, p, -1.0, 1.0);
						break;
					default:
						FATAL(("illegal option: -am%c", *(p - 1)));
						break;
					}
					break;
				case 'o':
					if (User_options.audio_ports == -1) {
						User_options.audio_ports = 0;
					}
					{
						int x = get_int(o, p, 0, (int)AUDIO_PORT_BIT_COUNT - 1);
						
						User_options.audio_ports |= (1 << x); break;
					}
					break;
				default:
					FATAL(("illegal option: -a%c", *(p - 1)));
					break;
				}
			}
			return;
		case 'p':
			User_options.auto_play = False;
			User_options.auto_play_defined = True;
			continue;
#if 0 /* Tristan Savatier - Sun Aug 23 02:19:06 PDT 1998 */
//		case 'P':
//			User_options.port = get_int(o, p, 1, (1 << BITS(short)) - 1);
//			return;
//#ifdef _REENTRANT
//		case 'U':
//			/*
//			 * UDP implementation is experimental at this point, and it does
//			 * not work well.  It is not officially supported.
//			 */ 
//			User_options.port_udp = get_int(o, p, 1, (1 << BITS(short)) - 1);
//			return;
//#endif
#endif /* 0 - Tristan Savatier - Sun Aug 23 02:19:06 PDT 1998 */
		case 'd':
			User_options.dither_type = (DITHER_TYPE)get_int(o, p, 0, (int)DITHER_COUNT - 1);
			return;
		case 's':
			switch (o = *p) {
				case 'f':
					p++;
					User_options.sync_framerate = get_double(o, p, FRAMERATE_MIN, FRAMERATE_MAX);
					User_options.sync_framerate_defined = True;
					User_options.sync_type_defined = True;
					User_options.sync_type = SYNC_FRAMERATE;
					break;
				default:
					User_options.sync_type = (SYNC_TYPE)get_int(o, p, 0, (int)SYNC_COUNT - 1);
					User_options.sync_type_defined = True;
					break;
			}
			return;

		case 'g':
			User_options.gamma = get_double(o, p, GAMMA_MIN, GAMMA_MAX);
			User_options.gamma_defined = True;
			return;
		case 'c':
			User_options.contrast = get_double(o, p, CONTRAST_MIN, CONTRAST_MAX);
			User_options.contrast_defined = True;
			return;
		case 'b':
			User_options.brightness = get_double(o, p, BRIGHTNESS_MIN, BRIGHTNESS_MAX);
			User_options.brightness_defined = True;
			return;
		case 'x':
			Player_name = p;
			return;
		case 'W':	if (isdigit(*p)) {
						User_options.existing_window = get_int(o, p, 0, MAXINT);
						return;
					}
					switch (o = *p++) {
					case 'r':
						/* play in root window */
						User_options.existing_window = -1;
						break;
					case 'f':
						User_options.video_layout = atoi(p);
						break;
					default:
						FATAL(("illegal option: -W%c", *(p - 1)));
						break;
					}
					return;
		case 'S':	User_options.loop_start_pos = get_int(o, p, 0, MAXINT);
					return;
		case 'E':	User_options.loop_end_pos = get_int(o, p, 0, MAXINT);
					return;
		case '\0':	return;
		default:	FATAL(("unknown option -%c; use -h for usage", o));
					clean_exit(EXIT_FAILURE);
		}
	}
}

/*--------------------------------------------------------------------------+
| init_defaults																|
+--------------------------------------------------------------------------*/
static void
init_defaults()
{
#ifdef VCD
	{
		const char *p = getenv("MTV_VCD_DEVICE");
		
		if (p == NULL) {
#ifdef LINUX
			Vcd_device = "/dev/cdrom";
#endif
#ifdef SOLARIS
	    if (volmgt_running()) {
	        Vcd_device     = "/vol/dev/aliases/cdrom0";
	    } else {
	        Vcd_device     = "/dev/rdsk/c0t6d0s2";
	    }
#endif
   		} else {
			Vcd_device = p;
		}
		MESSAGE(("using VCD device = \"%s\"", Vcd_device));
	}
#endif
	
	user_options_set_default();
}


/*--------------------------------------------------------------------------+
| parse_preferences															|
+--------------------------------------------------------------------------*/
static void
parse_preferences()
{	
	/*
	 * Init default preferences.
	 */
	pref_init_keys(Pref_keys, sizeofarray(Pref_keys));

	/*
	 * Load preferences from preference file.
	 */
	if (pref_load(User_options.pref_filename) != 0) {
		Must_save_preferences = True;
	}
	
	/*
	 * Set user options from preferences.
	 * Preferences values can be overridden by using command-line options,
	 * so do not set values already defined by command-line options.
	 */
	 
	if (!User_options.loop_defined) {
		pref_get_value_bool(PREF_LOOP, &User_options.loop);
	}
	if (!User_options.zoom_defined) {
		pref_get_value_int(PREF_ZOOM, (int *)&User_options.zoom, ZOOM_1X1, ZOOM_2X2);
	}
	if (!User_options.auto_play_defined) {
		pref_get_value_bool(PREF_AUTO_PLAY, &User_options.auto_play);
	}
	if (!User_options.gamma_defined) {
		pref_get_value_double(PREF_VIDEO_GAMMA, &User_options.gamma, GAMMA_MIN, GAMMA_MAX);
	}
	
	if (!User_options.contrast_defined) {
		pref_get_value_double(PREF_VIDEO_LUMA_CONTRAST, &User_options.contrast, CONTRAST_MIN, CONTRAST_MAX);
	}
	if (!User_options.brightness_defined) {
		pref_get_value_double(PREF_VIDEO_LUMA_BRIGHTNESS, &User_options.brightness, BRIGHTNESS_MIN, BRIGHTNESS_MAX);
	}
	if (!User_options.chroma_saturation_defined) {
		pref_get_value_double(PREF_VIDEO_CHROMA_SATURATION, &User_options.chroma_saturation, CHROMA_SATURATION_MIN, CHROMA_SATURATION_MAX);
	}
	if (!User_options.mtvp_geometry_defined) {
		pref_get_value_int(PREF_MTVP_WINDOW_POSITION_X, &User_options.mtvp_geometry_x, MININT, MAXINT);
		pref_get_value_int(PREF_MTVP_WINDOW_POSITION_Y, &User_options.mtvp_geometry_y, MININT, MAXINT);
	}
	if (!User_options.audio_channel_defined) {
		pref_get_value_int(PREF_AUDIO_CHANNEL, (int *)&User_options.audio_channel, 0, AUDIO_CHANNEL_COUNT - 1);
	}
	if (!User_options.audio_quality_defined) {
		pref_get_value_int(PREF_AUDIO_QUALITY, (int *)&User_options.audio_quality, 0, AUDIO_QUALITY_COUNT - 1);
	}
	if (!User_options.sync_type_defined) {
		pref_get_value_int(PREF_SYNC_TYPE, (int *)&User_options.sync_type, 0, SYNC_COUNT - 1);	
	}
	if (!User_options.sync_framerate_defined) {
		pref_get_value_double(PREF_SYNC_FRAMERATE, &User_options.sync_framerate, 0.0, SYNC_FRAMERATE_MAX);
	}
	pref_get_value_int(PREF_VIDEO_ASPECT_RATIO_MODE, (int *)&Aspect_ratio_mode, 0, MAXINT);
	pref_get_value_bool(PREF_VIDEO_PRIVATE_COLORMAP, (int *)&User_options.private_colormap);
	
	Name = pref_get_value(PREF_REGISTRATION_NAME);
	Order_number = pref_get_value(PREF_REGISTRATION_ORDER_NUMBER);
	Key = pref_get_value(PREF_REGISTRATION_KEY);
		
	if (strlen(Order_number) == 10 && strlen(Key) == 8) {
	
		if (strcmp(Order_number, "I000000000") == 0 && strcmp(Key, "00000000") == 0) {
			Registered = True;
			Use_mtvp_key = True;
		} else {
#ifdef HANDLE_CHALLENGE
			Registered = check_registration_key(Order_number, Key, &Msg);
#endif
		}
	}
#ifdef HANDLE_CHALLENGE
	else if (strlen(Order_number) == 0 && strlen(Key) == KEY_ASCII_LENGTH) {
		Registered = check_key(Key, &License_key, False);
	}
#endif
}

/*--------------------------------------------------------------------------+
| start_mtvp																|
+--------------------------------------------------------------------------*/
static void
start_mtvp()
{
#if 0 /* Tristan Savatier - Wed Mar 11 01:42:59 PST 1998 */
//	char window_flag[12];
#endif /* 0 - Tristan Savatier - Wed Mar 11 01:42:59 PST 1998 */
	char msg_flag[14];
	char audio_msg_flag[15];
	char audio_flag[4];
	const char *mtvp_argv[6];

	/*
	 * Start the player
	 */
	
	/*
	 * Set player options. be careful not to overrun any array!
	 */
	
#if 0 /* Tristan Savatier - Wed Mar 11 01:41:56 PST 1998 */
//	window_flag[sizeof window_flag - 1] = 0;
//	sprintf(window_flag, "-W%d", /* User_options.existing_window == -1 ? 0 : User_options.existing_window */ 0);
//	if (window_flag[sizeof window_flag - 1] != 0) {
//		abort_now();
//	}
#endif /* 0 - Tristan Savatier - Wed Mar 11 01:41:57 PST 1998 */
	msg_flag[sizeof msg_flag - 1] = 0;
	sprintf(msg_flag, "-w%d", - (User_options.msg_select + 1));
	if (msg_flag[sizeof msg_flag - 1] != 0) {
		abort_now();
	}
	audio_msg_flag[sizeof audio_msg_flag - 1] = 0;
	sprintf(audio_msg_flag, "-aw%d", - (User_options.audio_msg_select + 1));
	if (audio_msg_flag[sizeof audio_msg_flag - 1] != 0) {
		abort_now();
	}
	audio_flag[sizeof audio_flag - 1] = 0;
	sprintf(audio_flag, "-a%d", User_options.audio_mode);
	if (audio_flag[sizeof audio_flag - 1] != 0) {
		abort_now();
	}
	
	mtvp_argv[0] = Player_name;		/* default "mtvp" */
#if 0 /* Tristan Savatier - Wed Mar 11 01:42:08 PST 1998 */
//	mtvp_argv[1] = window_flag;
#endif /* 0 - Tristan Savatier - Wed Mar 11 01:42:08 PST 1998 */
	mtvp_argv[1] = msg_flag;
	mtvp_argv[2] = audio_msg_flag;
	mtvp_argv[3] = audio_flag;
#ifdef HANDLE_CHALLENGE
	if (Use_mtvp_key) {
		mtvp_argv[4] = NULL;
	} else {
		mtvp_argv[4] = "-k";	/* use challenge, don't look for keys in mtvp.key */
		mtvp_argv[5] = NULL;
	}
#else
	mtvp_argv[4] = NULL;
#endif
	/*
	 * First parameter of g2a_init() is a ptr to a NULL-terminated array
	 * of strings.
	 * The first string is the name of the Player (e.g. "mtvp").
	 * The following strings are possible parameters.
	 * If the first string is a full pathname (starting with "/")
	 * g2a_init() will run this program as "the Player".
	 *
	 * If the first string is not a full pathname (starting with "/")
	 * g2a_init() will look for the Player in the following places:
	 * - in the directory where argv[0] is located (if argv[0] contains a '/')
	 * - in the current directory (".")
	 * - in the PATH
	 *
	 * Typically first string is "mtvp", if the player was installed
	 * under this name.
	 *
	 * Second parameter of g2a_init() is a string that should be
	 * the argv[0] from the calling program.
	 */
	 	
	if (g2a_init(&Player, mtvp_argv, Argv0) == -1) {
		if (errno == 0) {
			FATAL_ERROR(("cannot start player (%s) - probably mtvp API version mismatch", mtvp_argv[0]));
		} else if (errno == EPIPE) {
			FATAL_ERROR(("cannot start player (%s) - check that player is in your PATH!", mtvp_argv[0]));
		} else {
			FATAL_ERROR(("cannot start player (%s) - could not fork", mtvp_argv[0]));
		}
	}
	MESSAGE(("player (%s) started, pid = %u", mtvp_argv[0], (u_int)Player.pid));
}




/*--------------------------------------------------------------------------+
| main																		|
+--------------------------------------------------------------------------*/
/*ARGSUSED*/
int
main(
	int argc,
	const char **argv,
	const char **envp
)
{
	const char	**p, **q;
	int		n;

	MSG_PTR->progname = "mtv";	/* for messages, see msg.h */
	Argv0 = argv[0];
	
	init_signals();

	init_defaults();

	/* parse options */
	for (p = argv + 1, q = argv + argc, n = 0; p != q; p++) {
		if (**p == '-') {
			if ((*p)[1] == 0) {
				/* parameter '-' means read from stdin */
#if 0 /* Tristan Savatier - Sun Aug 23 02:08:55 PDT 1998 */
//				User_options.stdin_mode = True;
//				User_options.filename = NULL;
#endif /* 0 - Tristan Savatier - Sun Aug 23 02:08:55 PDT 1998 */
				switch (++n) {
					case 1:		User_options.url = *p; break;
					default:	FATAL(("too many parameters"));
				}
			}
		} else if (**p == '+') {
			/* all the options after a "mtv" are for xforms! */
			User_options.xforms_argv = p;
			break;
		} else {
			switch (++n) {
				case 1:		User_options.url = *p; break;
				default:	FATAL(("too many parameters"));
			}
			continue;
		}
		parse_options(*p + 1);
	}

	parse_preferences();
	
	if (User_options.usage) {
		usage();
		clean_exit(EXIT_FAILURE);
	}

#if 0 /* Tristan Savatier - Sun Aug 23 02:17:44 PDT 1998 */
//	if (User_options.filename != NULL || User_options.port != 0 || User_options.port_udp != 0 || User_options.stdin_mode) {
//		if ((User_options.filename != NULL)+(User_options.port != 0)+(User_options.port_udp != 0)+(User_options.stdin_mode == True) != 1) {
//			FATAL(("only one input should be specified"));
//		}
//	}
#endif /* 0 - Tristan Savatier - Sun Aug 23 02:17:44 PDT 1998 */

	Pid = getpid();

	MESSAGE(("mtv started, pid = %u", (u_int)Pid));

	start_mtvp();
	
	/* do the work here */
	do_gui_fork(argv[0], User_options.xforms_argv);

	/* terminated successfully */
	clean_exit(EXIT_SUCCESS);

	/* so that lint does not complain! */
	return 0;
}


/* eof */
